#include <iostream>
#include "Symmetry.h"
#include <Qt/qstring.h>
#include <QApplication>

#define DEFAULT_SRC_DIR "../../../data/"
#define DEFAULT_SETTINGS_DIR "../../scripts/settings/"

void PrintUsage(char * programName)
{
	std::cout<<"Usage:"<<std::endl;
	std::cout<<programName<<" srcImg inSettings outputDir [-trgSym g] [alternative settings]"<<std::endl;
}

int main(int argc, char** argv)
{
	//Q_INIT_RESOURCE(application);
	
	QApplication app(argc, argv);
	
	TimeProfiler profile;
	
	if (argc<4)
	{
		PrintUsage(argv[0]);
		exit(0);
	}
	
	std::cout<<"Loading images... "<<std::flush;
	QGLWidget * widget = new QGLWidget ();
	
	QString settingsFile(argv[2]);
	if (argv[2][0]=='-')
		settingsFile = QString(DEFAULT_SETTINGS_DIR) + settingsFile.right(settingsFile.length()-1);
	
	QString outDir(argv[3]);
	ExampleWithAnswer * example = NULL;	
	QImage img = SymPipelineParams::LoadImage(QString(argv[1]), DEFAULT_SRC_DIR, &example);

	QImage * targetSymmetry = NULL;
	if (argc>=5 && strcmp(argv[4], "-trgSym")==0)
	{
		QString trgSym(argv[5]);
		targetSymmetry = new QImage();
		ExampleWithAnswer * exampleTrg;
		*targetSymmetry = SymPipelineParams::LoadImage(trgSym, DEFAULT_SRC_DIR, &exampleTrg);
	}
	
	std::cout<<"Loading parameters... "<<std::endl;	
	int argID = targetSymmetry==NULL ? 4 : 6;
	SymParamsText paramsText(settingsFile, &(argv[argID]), argc-argID);
	SymPipelineParams * params;

	std::cout<<"parameter file: "<<settingsFile.toStdString().c_str()<<std::endl;
	
	if (targetSymmetry==NULL)	// use filtered f as a target
		params = new SymPipelineParams(img, paramsText, example);	
	else	// given target symmetry
		params = new SymPipelineParams(img, *targetSymmetry, paramsText, example);
	
	std::cout<<"Initializing symmetrizer..."<<std::endl;
	SymPipeline symmetrizer(params);
	profile.startedProcess("Symmetrizing (optimization)");
	symmetrizer.FindOptimalSolution();
	profile.finishedProcess("Symmetrizing (optimization)", true);
	//QImage & imOut = ((ImageWarp*)params->m_pDeformationModel)->GetDeformedImage();

	if (targetSymmetry==NULL && params->m_pSg==NULL)
		symmetrizer.OutputResultsImg(outDir, &img, NULL, NULL, NULL, "png", true);
	else if (targetSymmetry==NULL)
		symmetrizer.OutputResultsImg(outDir, &img, NULL, NULL, (SymMatrix2DReal *)params->m_pSg, "png", true);
	else
		symmetrizer.OutputResultsImg(outDir, &img, NULL, targetSymmetry, (SymMatrix2DReal *)params->m_pSg, "png", true);
	
	//SymPipeline::TestSimple();
	//	std::cout<<"============= Testing fft ======"<<std::endl;
	//	FFTManager::TestFFT2D();
	//	std::cout<<"============= Testing toroidal symmetry ======"<<std::endl;
	//	SymSpaceTransToroidal2D::TestSymmetryTransform();
	//	SymEnergySumPow::TestSymEnergySumPow();
	//	
}

