#include <iostream>
#include "Symmetry.h"
#include <Qt/qstring.h>

#include "opencv/cv.h"
#include "opencv/highgui.h"

#define DEFAULT_IMG_DIR "../../../data/imageSymmetrization/"
#define DEFAULT_SETTINGS_DIR "../../scripts/settings/"

void TestImageWarpHierarchiesRealImg(QString, QString);
void TestIterativeS();
void TestHierarchicalSymmetries(QString img);
void TestPerfectInput();
void TestColorIndependentEnergy();
void TestPntAssymmetrization();
void TestTextureSynth();
void TestOpenCVCanny();
void CreateAPerfectStarfish();
//void TestPointDistributionSmooth();

void PrintUsage(char * programName)
{
	std::cout<<"Usage:"<<std::endl;
	std::cout<<programName<<" curveFile inSettings outputDir [-trgSym g] [alternative settings]"<<std::endl;
}

int high_switch_value = 0;
int highInt = 0;
int low_switch_value = 0;
int lowInt = 0;

void switch_callback_h( int position ){
	highInt = position;
}
void switch_callback_l( int position ){
	lowInt = position;
}






int main(int , char** )
{
//	SymSpaceTransToroidal2D::TestSymmetryTransform();	
	//ImageWarpLinearMesh::TestImageWarpMeshing();
//	ImageWarpLinearMesh::TestImageWarpHierarchies();
//	ImageWarpLinearMesh::TestImageWarpHierarchiesUsingSwitch();
//	ImageWarpLinearMesh::TestImageWarpFoldover();	
//	TestImageWarpHierarchiesRealImg(QString(DEFAULT_IMG_DIR)+"1.161_w17x17.jpg", QString(DEFAULT_SETTINGS_DIR)+"symmetrizerENormSHierarch.txt");
//	TestImageWarpHierarchiesRealImg(QString(DEFAULT_IMG_DIR)+"1.161_w17x17.jpg", QString(DEFAULT_SETTINGS_DIR)+"symmetrizerENormS.txt");
//	TestIterativeS();
//	TestHierarchicalSymmetries("-3.bricks_sm7x11.jpg");
//	ImageWarpLinearMesh::TestAsRigidAsPossibleWarp();
//	ImageWarpLinearMesh::TestAsRigidAsPossibleWarpHierarchySwitchConsistency();
//	ImageWarpLinearMesh::TestFeatureBasedMeshing();
//	TestCheckerboardPerfectSolution();
//	TestPerfectInput();
//	TestColorIndependentEnergy();
//	TestPntAssymmetrization();
//	TestTextureSynth();
//	TestPointDistributionSmooth();
	
//	DeformationLensDistortion::TestLensDistortion();
//	SymSpacePointwisePadded::TestPointwise();
//	TestOpenCVCanny();
	CreateAPerfectStarfish();
}


void TestImageWarpHierarchiesRealImg(QString imgSrc, QString setup)
{
	QImage img(imgSrc);
	SymParamsText paramsText(setup, NULL, 0);
	SymPipelineParams * params;
	params = new SymPipelineParams(img, paramsText);	// use filtered f as a target
	
	params->m_pOptimizer->Iterate();
		
	SymPipeline symmetrizer(params);
	symmetrizer.FindOptimalSolution();
}

void TestIterativeS()
{	
	for (int i=0; i<4; i++)
	{
		QImage srcRegImg;
		QString name;
		if (i==0)
		{
			name = "rectangles1";
			srcRegImg = ImageWarp::TestCreateRegularSpacedGlyph(5, 5, 20, 20, 10, 10, "rectangle");
		}
		else if (i==1)
		{
			name = "rectangles2";
			srcRegImg = ImageWarp::TestCreateRegularSpacedGlyph(5, 5, 20, 20, 10, 5, "rectangle");
		}
		else if (i==2)
		{
			name = "ellipsoids1";
			srcRegImg = ImageWarp::TestCreateRegularSpacedGlyph(5, 5, 20, 20, 10, 10, "ellipse");
		}
		else if (i==3)
		{
			name = "crosses1";
			srcRegImg = ImageWarp::TestCreateRegularSpacedGlyph(5, 5, 20, 20, 10, 10, "cross");
		}
		
		SymSpace * S = new SymSpaceTransToroidal2D(srcRegImg.width(), srcRegImg.height(), -1, -1, SymSpace::NORM_SYMMETRY_SPACE, 1, 1);
		
		for (int numS = 0; numS<4; numS++)
		{
			srcRegImg.save(name+"_S"+QString::number(numS)+".png");
			SymMatrix2DReal imgMatrix(srcRegImg);	
			srcRegImg = ((SymMatrix2DReal&)S->CalculateSymmetryAndProcess(imgMatrix)).ToImage();
		}
	}
}

void TestHierarchicalSymmetries(QString imgPath)
{
	ExampleWithAnswer * example;
	QImage img = SymPipelineParams::LoadImage(imgPath, DEFAULT_IMG_DIR, &example);
	SymSpaceTransToroidal2D * S = new SymSpaceTransToroidal2D(img.width(), img.height(), -1, -1, SymSpace::NORM_SYMMETRY_SPACE_MAX_TO_1, 1, 1);	
	std::vector<SymMatrix2DReal*> symmetries;
	SymMatrix2DReal f(img);
	SymMatrix2DReal matrixOfSums(S->symm_m, S->symm_n);
	symmetries.push_back(new SymMatrix2DReal((SymMatrix2DReal&)S->CalculateSymmetryAndProcess(f)));	
	std::cout<<"Symmetries[0]="<<(*(symmetries[0]))(0,0)<<", "<<(*(symmetries[0]))(0,1)<<std::endl;
	double p=10;
	for (int i=0; i<5; i++)
	{
		SymMatrix2DReal * newMatrix = new SymMatrix2DReal(S->symm_m, S->symm_n);
		newMatrix->Add(*(symmetries[0]));
		newMatrix->Subtract(matrixOfSums);
		newMatrix->PointwisePower(p, true);
		S->Normalize(*newMatrix, 0);
		std::cout<<"Symmetries["<<i+1<<"]="<<(*newMatrix)(0,0)<<", "<<(*newMatrix)(0,1)<<std::endl;		
		
		symmetries.push_back(newMatrix);
		matrixOfSums.Add(*newMatrix);
		
		newMatrix->ToImage().save("S_"+QString::number(i+1)+".png");
	}
}

void TestPerfectInput()
{
	std::cout<<"Running Tester"<<std::endl;
	ImageWarpTestExample newExample("simpleCircleExample");
	
	//lattice
	newExample.m_NumX = 8;
	newExample.m_NumY = 8;
	newExample.m_Blur = 0;
	newExample.m_V1 = R2Vector(0, 30);
	newExample.m_V2 = R2Vector(30, 0);
	newExample.m_BackgroundColor = 1;
	
	//glyphs
	newExample.m_GlyphName = "ellipse";
	//newExample.m_GlyphName = "rectangle";
	//newExample.m_GlyphName = "cross";
	newExample.m_GlyphX = 10;
	newExample.m_GlyphY = 10;
	newExample.m_GlyphMeanColor = 0;
	newExample.m_OddShift = 0;
	newExample.m_GlyphQuadVariations = 0;
	
	newExample.m_FractionMissing = 0;
	newExample.m_FractionExtra = 0;
	newExample.m_ExtraBlobs = 0;
	newExample.m_ExtraBlobsFractionArea = 0;
	newExample.m_GlyphLocationVariationHighFreq = 0;
	newExample.m_GlyphLocationVariationLowFreq = 0;
	newExample.m_GlyphDimensionsVariation = 0;
	newExample.m_GlyphColorVariations = 0;
	newExample.m_bPutCtrlAtGlyphCorners = false;
	
	std::cout<<"Generating example"<<std::endl;	
	
	newExample.GenerateExample(1234);
	newExample.m_ImageInputExample.save("InputExample.png");
	newExample.m_ImagePerfectOutput.save("PerfectOutput.png");
	
	std::cout<<"Deforming"<<std::endl;		
	newExample.m_pValidSolutionSpace->DeformAtCurrentHierarchyLevel(newExample.m_pValidSolutionSpace->GetParameterVector());
	newExample.m_pValidSolutionSpace->GetDeformedImage().save("InputExample_DeformationImage.png");
	newExample.m_pValidSolutionSpace->DeformAtCurrentHierarchyLevel(*(newExample.m_pPerfectSolution));	
	newExample.m_pValidSolutionSpace->GetDeformedImage().save("InputExample_PerfectDeformation.png");	
	std::cout<<"Done Running Tester"<<std::endl;	
	
	QImage img = newExample.m_pValidSolutionSpace->GetDeformedImage();
	
	SymSpace * symSpaceToroidal = new SymSpaceTransToroidal2D(img.width(), img.height(), img.width(), img.height());
	SymSpace * symSpace0Padding = new SymSpaceTrans2D(img.width(), img.height());
	SymMatrix2DReal & matrixToroidal = (SymMatrix2DReal&)symSpaceToroidal->CalculateSymmetryAndProcess(newExample.m_pValidSolutionSpace->DeformAtCurrentHierarchyLevel(newExample.m_pValidSolutionSpace->GetParameterVector()));
	SymMatrix2DReal & matrix0Padding = (SymMatrix2DReal&)symSpace0Padding->CalculateSymmetryAndProcess(newExample.m_pValidSolutionSpace->DeformAtCurrentHierarchyLevel(newExample.m_pValidSolutionSpace->GetParameterVector()));
	matrixToroidal.ToImage().save("SymSpace_Toroidal.png");
	matrix0Padding.ToImage().save("SymSpace_0Padding.png");
}

void TestColorIndependentEnergy()
{
	ExampleWithAnswer * exampleBlack;
	ExampleWithAnswer * exampleWhite;	
	QImage blackBack = SymPipelineParams::LoadImage("~regGlyph_background_0_latticeDim_8_8_blur_0_v1_15_0_v2_0_15_oddShift_0_glyphType_ellipse_glyphDim_5_5_color_1_corners_false_miss_0_extra_0_blobs_0_area_0_highFreq_8_lowFreq_5_dimVar_0_colorVar_0_quadVar_0_seed_1243.png", "./", &exampleBlack);
	std::cout<<"================================================"<<std::endl;	
	QImage whiteBack = SymPipelineParams::LoadImage("~regGlyph_background_1_latticeDim_8_8_blur_0_v1_15_0_v2_0_15_oddShift_0_glyphType_ellipse_glyphDim_5_5_color_0_corners_false_miss_0_extra_0_blobs_0_area_0_highFreq_8_lowFreq_5_dimVar_0_colorVar_0_quadVar_0_seed_1243.png", "./", &exampleWhite);	

	std::cout<<"================================================"<<std::endl;
	std::cout<<"================================================"<<std::endl;
	SymSpaceTransToroidal2D * symSpaceToroidal = new SymSpaceTransToroidal2D(blackBack.width(), blackBack.height(), blackBack.width(), blackBack.height());
	SymEnergySymmetrization * energySymm = new SymEnergySymmetrization(symSpaceToroidal, "ColorConsistentSymmetrization");
	SymMatrix2DReal & blackMatrix =(SymMatrix2DReal & ) exampleBlack->m_pValidSolutionSpace->DeformAtCurrentHierarchyLevel(*(exampleBlack->m_pPerfectSolution));
	SymMatrix2DReal & whiteMatrix =(SymMatrix2DReal & ) exampleWhite->m_pValidSolutionSpace->DeformAtCurrentHierarchyLevel(*(exampleBlack->m_pPerfectSolution));
	blackMatrix.ToImage().save("blackBack.png");
	whiteMatrix.ToImage().save("whiteBack.png");	
	
	double diff1=0;
	double diff2=0;
	for (int i=0; i<blackMatrix.m; i++)
	{
		for (int j=0; j<blackMatrix.n; j++)
		{
			diff1+=qAbs(blackMatrix(i,j) - whiteMatrix(i,j));
			diff2+=qAbs(blackMatrix(i,j) + whiteMatrix(i,j) - 1);
		}
	}

	std::cout<<"diff1="<<diff1<<" diff2="<<diff2<<std::endl;
	std::cout<<"Black background, white foreground = "<<energySymm->GetEnergy(blackMatrix)<<std::endl;
	std::cout<<"White background, black foreground = "<<energySymm->GetEnergy(whiteMatrix)<<std::endl;	
}

void TestPntAssymmetrization()
{
	double * params = new double[8];
	params[0] = 15;
	params[1] = 15;	// lattice
	params[2] = 10;
	params[3] = 10;	// glyph dims
	params[4] = 0;
	params[5] = 0;	// noise high
	params[6] = 0;  // noise low
	
	DeformCurve * deformation = new DeformCurve("points", params, 1, -1, -1, 16, true);
	int numParams = deformation->GetParameterVector().m;
	SymMatrix2DReal & deformMatrix = (SymMatrix2DReal &)deformation->DeformAtCurrentHierarchyLevel(deformation->GetParameterVector());
	SymSpaceTransToroidal2D * symSpaceToroidal = new SymSpaceTransToroidal2D(deformMatrix.m, deformMatrix.n, deformMatrix.m, deformMatrix.n);
	SymEnergySymmetrization * energyFn = new SymEnergySymmetrization(symSpaceToroidal, "RatioOfNorms", true);	
	
	SymVectorReal bestSolution(deformation->GetParameterVector().m);
	int step = 1;
	double currEnergy = SYM_INFINITY;
	int lastUpdate = -1000000;
	SymMatrix2DComplex trg(deformMatrix.m, deformMatrix.n);
	FFTManager manager;
	for (int i=0; i<4000; i++)
	{
		SymMatrix2DReal & deformMatrix = (SymMatrix2DReal &)deformation->DeformAtCurrentHierarchyLevel(deformation->GetParameterVector());		
		double energy = energyFn->GetEnergy(deformMatrix);
		if (energy < currEnergy)
		{
			bestSolution.CopyValuesFrom(deformation->GetParameterVector().m_pData);
			currEnergy = energy;
			if (i - lastUpdate > 100)
			{
				deformation->GetDeformedImage().save("fp_"+QString::number(i)+".png");			
				((SymMatrix2DReal&) symSpaceToroidal->CalculateSymmetryAndProcess(deformMatrix)).ToImage().save("Sf_"+QString::number(i)+".png");
				manager.FFT2D(deformMatrix, trg);
				((SymMatrix2DReal&)trg).ToImage().save("FourierF_"+QString::number(i)+".png");
				
				std::cout<<"updated at "<<i<<" E="<<energy<<std::endl;
				lastUpdate = i;
			}
			
		}
		else
		{
			deformation->GetParameterVector().CopyValuesFrom(bestSolution.m_pData);
		}
		
		//std::cout<<"Energy["<<i<<"] = "<<energyFn->GetEnergy(deformMatrix)<<std::endl;
		int rndParam = rand() % (numParams/2);
		double rndAngle = (rand() % 360) * (2*3.14159265) / 360.;
		double dx = step * cos(rndAngle);
		double dy = step * sin(rndAngle);		
		deformation->GetParameterVector()(2*rndParam+0) += dx;
		deformation->GetParameterVector()(2*rndParam+1) += dy;		
	}
	
	deformation->DeformAtCurrentHierarchyLevel(bestSolution);
	deformation->GetDeformedImage().save("points_assymmetric.png");
}

void TestTextureSynth()
{
	QImage img = QImage(QString(DEFAULT_IMG_DIR)+"../textureSynthesis/src2.jpg");
	int numCandidatesExamined = 1000;
	int numBestSelected = 1;
	DeformTextureSynth textureSynth(img, 200, 200, 16, 16, 8, 8, numCandidatesExamined, numBestSelected, true);
	((SymMatrix2DReal&)textureSynth.DeformAtCurrentHierarchyLevel(textureSynth.GetParameterVector())).ToImage().save("f0.png");;
	textureSynth.GetDeformedImage().save("f0_img.png");
	int numPatches = textureSynth.m_iPatchNumX * textureSynth.m_iPatchNumY;
	int numIter = 1;
	std::cout<<"Iterating: "<<std::endl;
	for (int i=0; i<numPatches * numBestSelected * numIter; i++)
	{
		if (i%10==0)
			std::cout<<"."<<std::flush;
		bool force=false;
		textureSynth.TakeRandomStep(force);
	}
	std::cout<<std::endl;
	((SymMatrix2DReal&)textureSynth.DeformAtCurrentHierarchyLevel(textureSynth.GetParameterVector())).ToImage().save("f1.png");;
	textureSynth.GetDeformedImage().save("f1_img.png");
}

//void TestPointDistributionSmooth()
//{
//	QString curveFilename;	
//											//dx, dy, numX, numY, numAdditional, highFreq, lowFreq, probMissing, seed, 
//	QString curveFileFast = DeformCurve::GetPointsDescriptor(15, 15, 10, 10, 0, 0, 0, 0, 10248,
//											//fastInt, numSmpls1, numSmpls2, approxPoly, approxType
//												true, 64, 64, 1, DeformCurve::CARTESIAN_COORDS);	
//	double * paramsFast = SymPipelineParams::DivideCurveParams(curveFileFast, &curveFilename);	
//
//	//dx, dy, numX, numY, numAdditional, highFreq, lowFreq, probMissing, seed, 
//	QString curveFileSlow = DeformCurve::GetPointsDescriptor(15, 15, 10, 10, 0, 0, 0, 0, 10248,
//												//fastInt, numSmpls1, numSmpls2, approxPoly, approxType
//												false, 64, 64, 1, DeformCurve::CARTESIAN_COORDS);	
//	double * paramsSlow = SymPipelineParams::DivideCurveParams(curveFileSlow, &curveFilename);	
//	
//	DeformCurve pnts1("points", paramsFast, 1, -1, -1, 2, true);
//	DeformCurve pnts2("points", paramsSlow, 1, -1, -1, 2, true);
//	int numVertices = pnts1.GetParameterVector().m/2;
//	int vertex = numVertices/2 + 5;
//	
//	double paramX = pnts1.GetParameterVector()(2*vertex+0);
//	double dx = 1;
//	SymMatrix2DReal & deformMatrix = (SymMatrix2DReal & ) pnts1.DeformAtCurrentHierarchyLevel(pnts1.GetParameterVector());
//	//SymSpaceTransToroidal2D symSpace(deformMatrix.m, deformMatrix.n);
//	SymSpaceTrans2D symSpace(deformMatrix.m, deformMatrix.n);
//
//	SymEnergySymmetrization energyFn(&symSpace, "LpNorm", true, 2);	
//
//	TimeProfiler profiler;
//	std::vector<double> energies1;
//	std::vector<double> energies2;
//	std::vector<double> paramsVals1;
//	std::vector<double> paramsVals2;	
//	double minParamX = paramX - 20;
//	double maxParamX = paramX + 20;
////	double minParamX = paramX;
////	double maxParamX = paramX;
//	profiler.startedProcess("Original Method");
//	SymMatrix2DComplex tempFourier(deformMatrix.m, deformMatrix.n);
//	FFTManager manager;
//	for (double currParamX = minParamX; currParamX <= maxParamX; currParamX+=dx)
//	{
//		pnts1.GetParameterVector()(2*vertex+0) = currParamX;
//		SymMatrix2DReal & deformMatrix1 = (SymMatrix2DReal&) pnts1.DeformAtCurrentHierarchyLevel(pnts1.GetParameterVector());
////		manager.FFT2D(deformMatrix1, tempFourier);
////		double L2FT = 0;
////		for (int i=0; i<tempFourier.m; i++)
////			for (int j=0; j<tempFourier.n; j++)
////				L2FT += tempFourier(i,j).dat[0]*tempFourier(i,j).dat[0] + tempFourier(i,j).dat[1]*tempFourier(i,j).dat[1];		
////		std::cout<<"L2(f(x,y)) = "<<deformMatrix1.SumElements(2)<<std::endl;
////		std::cout<<"L1(f(x,y)) = "<<deformMatrix1.SumElements(1)<<std::endl;
////		std::cout<<"L2(FT(f(x,y))="<<L2FT / (tempFourier.m*tempFourier.n)<<std::endl;
//// 		deformMatrix1.ToImage().save("pnts1_img.png");
//		//pnts1.GetDeformedImage().save("img_"+QString::number(currParamX)+".png");
//
//		energies1.push_back(energyFn.GetEnergy(deformMatrix1));
//		//energies1.push_back(deformMatrix1.SumElements(2));
//		paramsVals1.push_back(currParamX - paramX);
//		std::cout<<"."<<std::flush;
////		std::cout<<"E1(0) = "<<energies1[energies1.size()-1]<<std::endl;
//	}
//	std::cout<<std::endl;
//	profiler.finishedProcess("Original Method", false);	
//	profiler.startedProcess("New Method");		
//	for (double currParamX = minParamX; currParamX <= maxParamX; currParamX+=dx)
//	{
//		pnts2.GetParameterVector()(2*vertex+0) = currParamX;
//		SymMatrix2DReal & deformMatrix2 = (SymMatrix2DReal&) pnts2.DeformAtCurrentHierarchyLevel(pnts2.GetParameterVector());
//		double L2S = pnts2.m_fL2S;
//		double L2f = pnts2.m_fL2f;
//		
//		energies2.push_back(deformMatrix2(0,0));
//		//energies2.push_back(L2f);
//		paramsVals2.push_back(currParamX - paramX);
//		std::cout<<"."<<std::flush;
////		std::cout<<"E2(0) = "<<energies2[energies2.size()-1]<<std::endl;
//	}
//	std::cout<<std::endl;
//		
//	profiler.finishedProcess("New Method", false);		
//	
//	std::cout<<"plot(["<<std::flush;
//	for (int i=0; i<(int)paramsVals1.size(); i++)
//		std::cout<<paramsVals1[i]<<" "<<std::flush;
//	std::cout<<"], ["<<std::flush;
//	for (int i=0; i<(int)energies1.size(); i++)
//		std::cout<<energies1[i]<<" "<<std::flush;
//	std::cout<<"]); title('Energy Orignal L2');"<<std::endl;
//	
//	std::cout<<"figure, plot(["<<std::flush;
//	for (int i=0; i<(int)paramsVals2.size(); i++)
//		std::cout<<paramsVals2[i]<<" "<<std::flush;
//	std::cout<<"], ["<<std::flush;
//	for (int i=0; i<(int)energies2.size(); i++)
//		std::cout<<energies2[i]<<" "<<std::flush;
//	std::cout<<"]); title('Energy New L2');"<<std::endl;
//	
//	std::cout<<"\n"<<profiler.profile().toStdString().c_str()<<std::endl;
//}

void TestOpenCVCanny()
{
	const char* name = "Edge Detection Window";
	
	// Kernel size
	int N = 7;
	
	// Set up images
	IplImage* img = cvLoadImage( "MGC.jpg", 0 );
	std::cout<<"depth="<<img->depth<<" chan="<<img->nChannels<<std::endl;
	IplImage* img_b = cvCreateImage( cvSize(img->width+N-1,img->height+N-1), img->depth, img->nChannels );
	IplImage* out = cvCreateImage( cvGetSize(img_b), IPL_DEPTH_8U, img_b->nChannels );
	
	// Add convolution boarders
	CvPoint offset = cvPoint((N-1)/2,(N-1)/2);
	cvCopyMakeBorder(img, img_b, offset, IPL_BORDER_REPLICATE, cvScalarAll(0));
	
	// Make window
	cvNamedWindow( name, 1 );
	
	// Edge Detection Variables
	int aperature_size = N;
	double lowThresh = 20;
	double highThresh = 40;
	
	// Create trackbars
	cvCreateTrackbar( "High", name, &high_switch_value, 200, switch_callback_h );
	cvCreateTrackbar( "Low", name, &low_switch_value, 200, switch_callback_l );
	
	while( 1 ) {
		highThresh = 200+ highInt * 20;
		lowThresh =  lowInt * 10;
		
		// Edge Detection
		cvCanny( img_b, out, lowThresh*N*N, highThresh*N*N, aperature_size );		
		cvShowImage(name, out);
		
		if( cvWaitKey( 15 ) == 27 ) 
			break;
	}
	std::cout<<"Thresh=["<<lowThresh<<", "<<highThresh<<"]"<<std::endl;	
	// Release
	cvReleaseImage( &img );
	cvReleaseImage( &img_b );
	cvReleaseImage( &out );
	cvDestroyWindow( name );
}	

void CreateAPerfectStarfish()
{
	int symmetryNum = 4;
	std::vector<double> angles;
	angles.push_back(0.    * 2. * 3.14159265 / 360.);
	angles.push_back(65.   * 2. * 3.14159265 / 360.);
	angles.push_back(170.  * 2. * 3.14159265 / 360.);	
	angles.push_back(210.  * 2. * 3.14159265 / 360.);	
	angles.push_back(260.  * 2. * 3.14159265 / 360.);		
	
	if (angles.size()>0)
		symmetryNum = (int)angles.size();
	
	int width=7;
	int rad = 256;	
	QImage perfect(rad, rad, QImage::Format_RGB32);
	double px=width/(double)rad;
	double cx = .5;
	double cy = .5;
	
	for (int aID=0; aID<symmetryNum; aID++)
	{
		double angle = ((double)aID) * 2. * 3.14159265 / symmetryNum;
		if (angles.size()>0)
			angle=angles[aID];
		double locX = cx+cos(angle)/2.5;
		double locY = cx+sin(angle)/2.5;		
		std::cout<<"drawing: "<<"[.5 .5] -> ["<<locX<<", "<<locY<<"] w/ rad="<<2./rad<<std::endl;

		DrawingRoutines::DrawLine(perfect, cx, cy, locX, locY, px, qRgb(255, 255, 255));
	}
	perfect.save("PerfectRot"+QString::number(symmetryNum)+"_"+QString::number(width)+".png");
}



