/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim, Princeton University			   *
 *   Author: Vladimir Kim (kim_vladimir@yahoo.com, vk@princeton.edu)       *
 *   Research Supervisor: Tom Funkhouser (funk@cs.princeton.edu)           *
 *                                  ****                                   *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "MVCImage.h"
#include "CVLib.h"

MVCMeshVertex::MVCMeshVertex(double origX, double origY)
{
	
	x = origX;
	y = origY;
}

MVCImage::MVCImage()
{
	NUM_GRID_PNTS = 250;
	NUM_BND_PNTS = 250;
}


QImage MVCImage::StitchTileToItself(QImage tile)
{
	std::cout<<"Stitching To Itself: 1"<<std::endl;
	m_SrcImg = tile;
	m_TrgImg = tile;
		
	// initialize boundary
	std::vector<int> boundary = CreateRectangleBoundary(NUM_BND_PNTS);
	
	std::cout<<"Stitching To Itself: 2"<<std::endl;	

		// initialize mesh
	OverlayRectangleMesh(boundary, NUM_GRID_PNTS, NUM_GRID_PNTS);
	std::cout<<"Stitching To Itself: 3"<<std::endl;	
	CalculateMVC();
	std::cout<<"Stitching To Itself: 4"<<std::endl;	
	
	// for boundaries find diff
	for (int i=0; i<(int)m_Boundary.size(); i++)
	{
		double bndX = m_MeshVertices[m_Boundary[i]].x;
		double bndY = m_MeshVertices[m_Boundary[i]].y;
		QRgb g, fStar;
		if (bndX==0)
		{			
//			g = qRgb(0,0,0);
//			fStar = qRgb(0,0,0);
			int croppedY = (int)bndY<m_SrcImg.height() ? bndY : (bndY-1);
			fStar = m_SrcImg.pixel(m_SrcImg.width()-1, croppedY);
			g = m_SrcImg.pixel(0, croppedY);
		}
		else if (bndY==0)
		{
//			g = qRgb(0,0,0);
//			fStar = qRgb(0,0,0);
			int croppedX = (int)bndX<m_SrcImg.width() ? bndX : (bndX-1);
			fStar = m_SrcImg.pixel(croppedX, m_SrcImg.height()-1);
			g = m_SrcImg.pixel(croppedX, 0);
		}
		else if (bndX==m_SrcImg.width())
		{
			int croppedY = (int)bndY<m_SrcImg.height() ? bndY : (bndY-1);
			fStar = m_SrcImg.pixel(0, croppedY);
			g = m_SrcImg.pixel(m_SrcImg.width()-1, croppedY);
		}
		else if (bndY==m_SrcImg.height())
		{
			int croppedX = (int)bndX<m_SrcImg.width() ? bndX : (bndX-1);
			fStar = m_SrcImg.pixel(croppedX, 0);
			g = m_SrcImg.pixel(croppedX, (int)m_SrcImg.height()-1);
		}
		else
			assert(false);
		
		int redAve = (qRed(g)+qRed(fStar))/2;
		int greenAve = (qGreen(g)+qGreen(fStar))/2;
		int blueAve = (qBlue(g)+qBlue(fStar))/2;
		fStar = qRgb(redAve, greenAve, blueAve);
		//g = qRgb(redAve, greenAve, blueAve);
		
//		std::cout<<"m_MeshVertices["<<m_Boundary[i]<<"] = ["<<qRed(g)-qRed(fStar)<<", ";
//		std::cout<<qGreen(g)-qGreen(fStar)<<", "<<qBlue(g)-qBlue(fStar)<<"]"<<std::endl;
		m_MeshVertices[m_Boundary[i]].m_BoundaryDiffR = qRed(fStar)-qRed(g);
		m_MeshVertices[m_Boundary[i]].m_BoundaryDiffG = qGreen(fStar)-qGreen(g);
		m_MeshVertices[m_Boundary[i]].m_BoundaryDiffB = qBlue(fStar)-qBlue(g);
	}
	
	std::cout<<"Stitching To Itself: 5"<<std::endl;	
	// for vertices find MVC interpolant
	for (int i=(int)m_Boundary.size(); i<(int)m_MeshVertices.size(); i++)
	{
		double rR = 0;
		double rG = 0;
		double rB = 0;
		for (int j=0; j<(int)m_Boundary.size(); j++)
		{
			assert(j < (int)m_MeshVertices[i].m_MVCoords.size());
			double lambda = m_MeshVertices[i].m_MVCoords[j];
			double dR = m_MeshVertices[m_Boundary[j]].m_BoundaryDiffR;
			double dG = m_MeshVertices[m_Boundary[j]].m_BoundaryDiffG;
			double dB = m_MeshVertices[m_Boundary[j]].m_BoundaryDiffB;
			rR += lambda * dR;
			rG += lambda * dG;
			rB += lambda * dB;			
		}
		m_MeshVertices[i].m_BoundaryDiffR = rR;
		m_MeshVertices[i].m_BoundaryDiffG = rG;
		m_MeshVertices[i].m_BoundaryDiffB = rB;		
	}
	
	std::cout<<"Stitching To Itself: 6"<<std::endl;	
	// render r(x) 
//	m_iTempCoord = 498;
	QImage interpolant = RenderMeshMVCInterpolant();
	interpolant.save("Interpolant.png");
//	interpolant.save("MVC_498_Interpolant.png");
//	m_iTempCoord = 747;
//	 interpolant = RenderMeshMVCInterpolant();
//	interpolant.save("MVC_747_Interpolant.png");
//	m_iTempCoord = 128;
//	 interpolant = RenderMeshMVCInterpolant();
//	interpolant.save("MVC_128_Interpolant.png");
	
	std::cout<<"Stitching To Itself: 7"<<std::endl;	
	for (int x=0; x<m_SrcImg.width(); x++)
	{
		for (int y=0; y<m_SrcImg.height(); y++)		
		{
			int red = qRed(interpolant.pixel(x, y));
			red = (((double)red/255.)-0.5)*255.*2.;
			red += qRed(m_SrcImg.pixel(x, y));
//			if (x==0 && y==0)
//				std::cout<<"red="<<red<<" -> dR="<<(((double)red/255.)-0.5)*255.*2.<<std::endl;
			
			if (red<0)
				red = 0;
			else if (red>255)
				red = 255;
			
			int green = qGreen(interpolant.pixel(x, y));
			green = (((double)green/255.)-0.5)*255.*2.;
			green += qGreen(m_SrcImg.pixel(x, y));
			if (green<0)
				green = 0;
			else if (green>255)
				green = 255;
			
			int blue = qBlue(interpolant.pixel(x, y));
			blue = (((double)blue/255.)-0.5)*255.*2.;
			blue += qBlue(m_SrcImg.pixel(x, y));
			if (blue<0)
				blue = 0;
			else if (blue>255)
				blue = 255;
			m_TrgImg.setPixel(x, y, qRgb(red, green, blue));
		}
	}
	std::cout<<"Stitching To Itself: DONE!"<<std::endl;	
	return m_TrgImg;
}

std::vector<int> MVCImage::CreateRectangleBoundary(int numBnd)
{
	std::vector<int> boundary;
	
	int w = m_SrcImg.width();
	int h = m_SrcImg.height();	

	int numBndVX = qMin(numBnd, w-1);
	int numBndVY = qMin(numBnd, h-1);	
	
//	std::cout<<"ImgDims = "<<w<<"x"<<h<<std::endl;
	assert(numBndVX<w && numBndVY<h);
	
	// right
	for (int i=0; i<numBndVY; i++)
	{
		boundary.push_back(w);
		boundary.push_back(i*h/numBndVY);
	}
	
	// bottom
	for (int i=0; i<numBndVX; i++)
	{
		boundary.push_back(w-i*w/numBndVX);
		boundary.push_back(h);
	}
	
	// left
	for (int i=0; i<numBndVY; i++)
	{
		boundary.push_back(0);
		boundary.push_back(h-i*h/numBndVY);
	}
	
	// top
	for (int i=0; i<numBndVX; i++)
	{
		boundary.push_back(i*w/numBndVX);
		boundary.push_back(0);
	}
	return boundary;
}

void MVCImage::OverlayRectangleMesh(std::vector<int> & boundary, int numX, int numY)
{
	for (int i=0; i<(int)boundary.size()/2; i++)
	{		
		m_Boundary.push_back(i);
		m_MeshVertices.push_back(MVCMeshVertex((double)boundary[2*i+0], (double)boundary[2*i+1]));
	}

	int w = qMin(numX, m_SrcImg.width()-1);
	int h = qMin(numY, m_SrcImg.height()-1);
//	std::cout<<"wh="<<w<<"x"<<h<<std::endl;
	// simple (improve to adaptive?): throw n vertices on a grid
	for (int gx=1; gx<w+1; gx++)
	{
		for (int gy=1; gy<h+1; gy++)
		{
			int x = gx * m_SrcImg.width() / (w+1);
			int y = gy * m_SrcImg.height() / (h+1);
			m_MeshVertices.push_back(MVCMeshVertex(x, y));
		}
	}
	
	std::vector<double> verticesCV;
	
	for (int i=0; i<(int)m_MeshVertices.size(); i++)
	{
		verticesCV.push_back(m_MeshVertices[i].x);
		verticesCV.push_back(m_MeshVertices[i].y);
	}
	
	CVLibRoutines::GetDelaunayTriangulation(verticesCV, m_Faces);	

	QImage copyImg = m_SrcImg;
	CVLibRoutines::DrawMesh(copyImg, verticesCV, m_Faces);
	copyImg.save("Mesh.png");
	
	InitializeBuffer();
}

void MVCImage::CalculateMVC()
{
		//for all mesh coords
	for (int i=(int)m_Boundary.size(); i<(int)m_MeshVertices.size(); i++)	
	{	
		double x = m_MeshVertices[i].x;
		double y = m_MeshVertices[i].y;		
		double totalW = 0;
			//for all boundary pnts - find MVC
		for (int j=0; j<(int)m_Boundary.size(); j++)
		{
			double currBndX = m_MeshVertices[m_Boundary[j]].x;
			double currBndY = m_MeshVertices[m_Boundary[j]].y;			
			double prevBndX = m_MeshVertices[m_Boundary[(j-1+m_Boundary.size())%m_Boundary.size()]].x;
			double prevBndY = m_MeshVertices[m_Boundary[(j-1+m_Boundary.size())%m_Boundary.size()]].y;
			double nextBndX = m_MeshVertices[m_Boundary[(j+1)%m_Boundary.size()]].x;
			double nextBndY = m_MeshVertices[m_Boundary[(j+1)%m_Boundary.size()]].y;
			
			double vCurrX	= (currBndX-x);
			double vCurrY	= (currBndY-y);
			double normCurr = sqrt(vCurrX*vCurrX + vCurrY*vCurrY);
			double vNextX	= (nextBndX-x);
			double vNextY	= (nextBndY-y);
			double normNext = sqrt(vNextX*vNextX + vNextY*vNextY);			
			double vPrevX	= (prevBndX-x);
			double vPrevY	= (prevBndY-y);
			double normPrev = sqrt(vPrevX*vPrevX + vPrevY*vPrevY);			
			
			double alphaPrev = acos((vCurrX/normCurr)*(vPrevX/normPrev) + (vCurrY/normCurr)*(vPrevY/normPrev));
			double alphaNext = acos((vCurrX/normCurr)*(vNextX/normNext) + (vCurrY/normCurr)*(vNextY/normNext));			
			double w = (tan(alphaPrev/2.) + tan(alphaNext/2.)) / normCurr;
			
			m_MeshVertices[i].m_MVCoords.push_back(w);
			assert(w<0||w>=0);
//			if (!(w<0||w>=0))
//			{
//				std::cout<<"===============\n"<<std::endl;
//				std::cout<<"\tw="<<w<<" alphaPrev="<<alphaPrev<<" alphaNext="<<alphaNext<<std::endl;
//				std::cout<<"\tnorms="<<normPrev<<", "<<normCurr<<", "<<normNext<<"]"<<std::endl;
//				std::cout<<"\tPrev=["<<vPrevX<<", "<<vPrevY<<"] Curr=["<<vCurrX<<", "<<vCurrY<<"] Prev=["<<vPrevX<<", "<<vPrevY<<"]"<<std::endl;
//				std::cout<<"\tacos("<<(vCurrX/normCurr)*(vPrevX/normPrev) + (vCurrY/normCurr)*(vPrevY/normPrev)<<")"<<" and acos("<<(vCurrX/normCurr)*(vNextX/normNext) + (vCurrY/normCurr)*(vNextY/normNext)<<")"<<std::endl;
//			}
			totalW+=w;
		}
		
		for (int j=0; j<(int)m_Boundary.size(); j++)
			m_MeshVertices[i].m_MVCoords[j] /= totalW;
		if (!(totalW<0 || totalW>=0))
			std::cout<<"Vert["<<m_MeshVertices[i].x<<", "<<m_MeshVertices[i].y<<"] MVC="<<m_MeshVertices[i].m_MVCoords[0]<<" total="<<totalW<<std::endl;		
	}
}

void MVCImage::InitializeBuffer()
{
	QGLFormat format;
	format.setSampleBuffers(true);
	format.setSamples(4);
	QGLPixelBuffer * newBuffer = new QGLPixelBuffer(m_SrcImg.width(), m_SrcImg.height(), format);
	newBuffer->makeCurrent();	
	
	glDisable(GL_TEXTURE_2D);
	glShadeModel(GL_SMOOTH);
	glClearColor(0.0f, 0.0f, 0.0f, 1.);				
	glClearDepth(1.0f);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);
	
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glViewport(0,0, m_SrcImg.width(), m_SrcImg.height());
	gluOrtho2D(0, m_SrcImg.width(), 0, m_SrcImg.height());
	
	glScalef(1, -1, 1);
	glTranslatef(0, -m_SrcImg.height(), 0);
	glMatrixMode(GL_MODELVIEW);		
	m_pFramebuffer = newBuffer;
}

QImage MVCImage::RenderMeshMVCInterpolant()
{
	m_pFramebuffer->makeCurrent();
	
	glBegin(GL_TRIANGLES);
	double scalar = 255.*2.;
	
	for (int i=0; i<(int)m_Faces.size()/3; i++)
	{
		MVCMeshVertex & v1 = m_MeshVertices[m_Faces[3*i+0]];
		MVCMeshVertex & v2 = m_MeshVertices[m_Faces[3*i+1]];
		MVCMeshVertex & v3 = m_MeshVertices[m_Faces[3*i+2]];		
		//m_MeshVertices[i].m_MVCoords[j]
		double scale=249;
//		m_iTempCoord = 0;
		int coord = m_iTempCoord;
//		glColor3d(v1.m_MVCoords[coord]*scale, v1.m_MVCoords[coord]*scale, v1.m_MVCoords[coord]*scale);
		glColor3d(v1.m_BoundaryDiffR/scalar+0.5, v1.m_BoundaryDiffG/scalar+0.5, v1.m_BoundaryDiffB/scalar+0.5);
		glVertex2d(v1.x, v1.y);
//		glColor3d(v2.m_MVCoords[coord]*scale, v2.m_MVCoords[coord]*scale, v2.m_MVCoords[coord]*scale);		
		glColor3d(v2.m_BoundaryDiffR/scalar+0.5, v2.m_BoundaryDiffG/scalar+0.5, v2.m_BoundaryDiffB/scalar+0.5);			
		glVertex2d(v2.x, v2.y);
//		glColor3d(v3.m_MVCoords[coord]*scale, v3.m_MVCoords[coord]*scale, v3.m_MVCoords[coord]*scale);		
		glColor3d(v3.m_BoundaryDiffR/scalar+0.5, v3.m_BoundaryDiffG/scalar+0.5, v3.m_BoundaryDiffB/scalar+0.5);	
		glVertex2d(v3.x, v3.y);			
	}
	glEnd();
	return m_pFramebuffer->toImage();
}


