/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim, Princeton University			   *
 *   Author: Vladimir Kim (kim_vladimir@yahoo.com, vk@princeton.edu)       *
 *   Research Supervisor: Tom Funkhouser (funk@cs.princeton.edu)           *
 *                                  ****                                   *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "TimeProfiler.h"
#ifdef CALCULATE_TIMING
	TimeProfiler * TimeProfiler::m_pGlobalSuperProfiler = new TimeProfiler();;
#endif


TimeProfiler::Process::Process()
{
	m_iCurrStartTime = -1;
	m_iTotalRunningTime = 0;	
}

TimeProfiler::TimeProfiler()
{
}

void TimeProfiler::startedProcess(QString processID)
{
	if (m_mIDToProcessMap[processID].m_iCurrStartTime!=-1)
		finishedProcess(processID);			//if process is currently running - finish it first
	m_mIDToProcessMap[processID].m_iCurrStartTime = clock();
	if (m_mIDToProcessMap[processID].m_sProcessName.isEmpty())
		m_mIDToProcessMap[processID].m_sProcessName = processID;
}

void TimeProfiler::finishedProcess(QString processID, bool printTime)
{
	Process & process = m_mIDToProcessMap[processID];
	long diff = clock() -process.m_iCurrStartTime;
	process.m_iRunningTimes.push_back(diff);
	process.m_iTotalRunningTime += diff;
	process.m_iCurrStartTime = -1;
	
	if (printTime)
		std::cout<<">>>>>>> Finished: "<<process.m_sProcessName.toStdString()<<" - "<<getTotalProcessLengthInSeconds(process.m_sProcessName).toStdString()<<std::endl;
}

long TimeProfiler::getTotalProcessLengthInCycles(QString processID)
{
	return m_mIDToProcessMap[processID].m_iTotalRunningTime;
}

QString TimeProfiler::getTotalProcessLengthInSeconds(QString processID)
{
	return QString::number(m_mIDToProcessMap[processID].m_iTotalRunningTime / CLOCKS_PER_SEC) + "." + QString::number(m_mIDToProcessMap[processID].m_iTotalRunningTime % CLOCKS_PER_SEC) + "s";	
}

int TimeProfiler::getTotalProcessLengthInSecondsInt(QString processID)
{
	return m_mIDToProcessMap[processID].m_iTotalRunningTime / CLOCKS_PER_SEC;
}

QString TimeProfiler::profile()
{
	if (m_mIDToProcessMap.size()==0)
		return "Empty Time Profile";
	QString profile = "Time Profile:\n";
	std::map<QString, Process>::iterator iter = m_mIDToProcessMap.begin();
	for (; iter!=m_mIDToProcessMap.end(); iter++)
	{
		profile += QString("    ") + iter->second.m_sProcessName + " - " + getTotalProcessLengthInSeconds(iter->second.m_sProcessName) + "\n";
	}
	return profile;
}







