#include "Numerical.h"
#include "DeformImageWarping.h"
#include "R2Shapes/R2Shapes.h"
#include "R3Shapes/R3Shapes.h"
#include "R3Graphics/R3Graphics.h"
#include <QtOpenGL/QGLContext>
#include <QtOpenGL/QGLFramebufferObject>
#include <QtOpenGL/QGLPixelBuffer>
#include <QtOpenGL/QGLFormat>
#include <Qt/qimage.h>


#ifndef __DEFORM_IMAGE_WARPING_LINEAR_MESH_H
#define __DEFORM_IMAGE_WARPING_LINEAR_MESH_H

#define USE_QT_FRAMEBUFFERS

// the off-screen buffers
#ifndef USE_QT_FRAMEBUFFERS
struct BufferID
{
	GLuint m_iFrameBufferObject;
	GLuint m_iDepthBuffer;
	GLuint m_iColorBuffer;
	GLuint m_iTextureBuffer;
	unsigned char * m_aColorBuffer;
	GLuint m_iTextureBufferGrayscale;
	int m_iWidth;
	int m_iHeight;
};
#endif


class AsRigidAsPossibleLevelDescriptor
{
	public:
		AsRigidAsPossibleLevelDescriptor(int constrainedVars, int freeVars,
										 std::vector<DeformationHierarchyNode*> & breadthFirstSorted);
		void InitializeMatrices(std::vector<ImageWarp::ImgVertex> & vertices, std::vector<ImageWarp::ImgFace> & faces);
		void FillFreeParams(bool preserveScale);
		void FillConstrainedParams(SymVectorReal & inParamVector);
		void FillOutputParams(SymVectorReal & outParamVector);
		void DrawControlPoints(QImage & trg, bool inverse, std::vector<ImageWarp::ImgVertex> & vertices);
		
	protected:
		struct FittedTriangle;
	
		void AddValueG(SymMatrix2DReal & G, ImageWarp::ImgVertex & v1, ImageWarp::ImgVertex & v2, int v1_xory, int v2_xory, double coeff);
		void AddCoeffsForTriangleStep1(SymMatrix2DReal & G, ImageWarp::ImgVertex & v0, ImageWarp::ImgVertex & v1, ImageWarp::ImgVertex & v2);
		void AddCoeffsForTriangleStep2(SymMatrix2DReal & H, FittedTriangle & triangle);
		void AddCoeffsForTriangleStep2(SymVectorReal & f0, FittedTriangle & triangle);
	
		int * m_aParamMapConstrainted;
		int * m_aParamMapFree;	
		int * m_aParamMapToG;
		int m_iFreeVars;
		int m_iConstrainedVars;
		int m_iFreeVertices;
		int m_iConstrainedVertices;
	// Step 1
		SymMatrix2DReal * m_pGInverseB;
	
	// Step 2.1
		struct FittedTriangle
		{
			public:
				FittedTriangle(ImageWarp::ImgVertex v0, ImageWarp::ImgVertex v1, ImageWarp::ImgVertex v2, 
							   int paramID0, int paramID1, int paramID2);
				void GetFittedTriangle(double x0p, double y0p, double x1p, double y1p, double x2p, double y2p);
			
				int m_iParamID0;
				int m_iParamID1;
				int m_iParamID2;
			
				double x0f, y0f, x1f, y1f, x2f, y2f;
			
			protected:
				void FillMatrixF(double x01, double y01);			
				SymVectorReal * C;
				SymVectorReal * w;		
				SymMatrix2DReal * FInv;
				double x01, y01;
		};
		std::vector<FittedTriangle *> m_FittedTriangles;
	
	// Step 2.2
		SymMatrix2DReal * m_pHpCholeskyDecomp;
		SymMatrix2DReal * m_pD;
		SymVectorReal * m_pf0;
		SymVectorReal * m_pf0AndDq;
		
	// Output (also used in step 1 & 2)
		SymVectorReal * m_pConstrainedParams;
		SymVectorReal * m_pFreeParams;		
	
	
};


class ImageWarpLinearMesh : public ImageWarp
{
	public:	
		enum RetrieveImageType
		{
			IMAGE_TYPE_REGULAR = 0,
			IMAGE_TYPE_TEXTURE_SPACE = 1
		};
	
		enum FaceType
		{
			FACE_QUAD,
			FACE_TRIANGLE1,
			FACE_TRIANGLE2,
			FACE_TRIANGLE3
		};
	
		enum InitialSeeding
		{
			IMAGE_SEED_REGULAR_LATTICE,
			IMAGE_SEED_FEATURE_PNTS,
			IMAGE_SEED_MANUAL
		};
		
	/**
	 * Triangulate on a regular lattice
	 */
		ImageWarpLinearMesh(QImage & img, int maxWidth, int maxHeight, int numRegularCtrlPnts[2], 
							bool inverse = false, bool fixBoundary=true, double paramStep = 1, 
							double minParamStep = 0, ImagePenalization foldover = ImagePenalization(), 
							ImageHierarchyDescriptor * hierarchyDescr = NULL, 
							FaceType faceType=FACE_TRIANGLE3, double zoomWindow = 1);
	/**
	 * Triangulate putting control points at feature points
	 */
		ImageWarpLinearMesh(QImage & img, int maxWidth, int maxHeight, int numPoints, 
							bool inverse = false, bool fixBoundary=true, double paramStep = 1, 
							double minParamStep = 0, ImagePenalization foldover = ImagePenalization(), 
							ImageHierarchyDescriptor * hierarchyDescr = NULL, double zoomWindow = 1, 
							bool addMorePnts = false);

	/**
	 * Manual mesh (e.g. for testing purposes)
	 */
		ImageWarpLinearMesh(QImage & img, int maxWidth, int maxHeight, std::vector<ImgVertex> &vertices, 
							std::vector<ImgFace> &faces, bool inverse = false, bool fixBoundary=true, 
							double paramStep = 1, double minParamStep = 0, 
							ImagePenalization foldover = ImagePenalization(), 
							ImageHierarchyDescriptor * hierarchyDescr = NULL, double zoomWindow = 1);
	
		virtual DeformationModel * CloneSelfUsingBestParams(SymVectorReal & bestParams);
	
		void InitializeMesh(QImage & img, std::vector<ImgVertex> &vertices, std::vector<ImgFace> & faces, bool inverse);
		void InitializeBuffers(QImage & img);
#ifdef USE_QT_FRAMEBUFFERS
		QGLPixelBuffer * InitializeBuffer(GLuint * textureIDs, bool initGrayscale, QImage & img);
#else
		BufferID InitializeBuffer(bool initGrayscale, QImage & img);
#endif
	
		virtual void ChildUpdateAtHierarchySwitch(int currLevel, int newLevel);
		virtual void CreateAsRigidAsPossibleHierarchy(int * numNodes, ImageHierarchyDescriptor * hierarchyDescr, double initStep, double minParamStep);
		virtual void CreateLinearWeightsHierarchy(int * numNodes, ImageHierarchyDescriptor * hierarchyDescr, double initStep, double minParamStep);
		virtual void CreateHierarchy(int * numNodes, ImageHierarchyDescriptor * hierarchyDescr, double initStep, double minParamStep);
		std::vector<int> GetChildrenIndices(double minX, double x1, double x2, double maxX, int numX);
	
		void GetCenterOfMass(double * cx, double * cy);	
		void GetOutputDimensions(int * width, int * height);
		
		virtual ~ImageWarpLinearMesh();
		virtual void GenerateHierarchicalParameterVectorChildImplement(SymVectorReal & params, SymVectorReal & trg);
		virtual SymMatrix & Deform(SymVectorReal & params, double * outPenalty = NULL);
		virtual QImage GetDeformedImage(int imageType = 0);
		virtual int GetNumOutputImageTypes();
		virtual void Draw(bool highResVersion);
	
		virtual void SetPostFilter(SymFilter * filter);
		virtual QImage DrawGeodesicSorting();
		
		int m_NumCtrlPnts[2];
		bool m_bShowWireframe;
		bool m_bPreserveScale;
		
		double CalculateDeformationPenalty();
		bool CheckTriangle(int v1, int v2, int v3);
		
		static void TestImageWarpMeshing();
		static void TestImageWarpHierarchies();
		static void TestImageWarpHierarchiesUsingSwitch();	
		static void TestImageWarpFoldover();
		static void TestAsRigidAsPossibleWarp();
		static void TestAsRigidAsPossibleWarpHierarchySwitchConsistency();	
		static void TestFeatureBasedMeshing();
	
//	protected:
	
		double m_fZoomWindow;
		int m_iMaxWidth;
		int m_iMaxHeight;
		QImage m_HighResImg;
		void DrawControlPoints();
		bool m_bRetrievingImage;
		void GenHierarchicalParamsAsRigidAsPossible(SymVectorReal & params, SymVectorReal & trg);
	
		std::vector<int> GetHierarchicalSortingByGeodesic();
	
		ImageHierarchyDescriptor * m_HierarchyDescriptor;
		FaceType m_FaceType;
		bool m_bFixBoundary;
		bool m_bInverseMapping;
		InitialSeeding m_SeedingType;
		double m_fInitialParamStep;
		double m_fMinParamStep;
		std::vector<ImgVertex> m_Vertices;
		std::vector<ImgFace> m_Faces;
#ifdef USE_QT_FRAMEBUFFERS
		GLuint m_TextureIds[2];
		GLuint m_HighResTextureIds[1];
		QGLPixelBuffer * m_pFramebuffer;
		QGLPixelBuffer * m_pHighResBuffer;
#else
		BufferID m_iFramebuffer;
		BufferID m_iHighResBuffer;
		QImage DrawAndReadBuffer(bool highResBuffer, bool grayScale);
		void LoadBuffer(bool highResBUffer, bool grayScale);
		QImage ReadBuffer(bool highResBUffer, bool grayScale);	
#endif
	
		ImagePenalization m_Penalization;
				
		std::vector<AsRigidAsPossibleLevelDescriptor *> m_AsRigidHierarchyLevels;
		QImage m_ControlPntsImg;
};

#endif

