#include "DeformImageWarpLinearMesh.h"
#include "DrawingRoutines.h"
#include <iostream>
#include <time.h>
#include "TimeProfiler.h"

/********************* Code for AsRigidAsPossibleLevelDescriptor ********************/
AsRigidAsPossibleLevelDescriptor::FittedTriangle::FittedTriangle(ImageWarp::ImgVertex v0, ImageWarp::ImgVertex v1, ImageWarp::ImgVertex v2, 
																 int paramID0, int paramID1, int paramID2)
{
	double x0 = v0.x;	double y0 = v0.y;	double x1 = v1.x;	double y1 = v1.y;	double x2 = v2.x;	double y2 = v2.y;
	m_iParamID0 = paramID0;					m_iParamID1 = paramID1;					m_iParamID2 = paramID2;
	
	double v1x = x1-x0;
	double v1y = y1-y0;
	double n1 = sqrt(v1x*v1x + v1y*v1y);
	v1x/=n1;	v1y/=n1;
	
	double v2x = x2-x0;
	double v2y = y2-y0;
	double n2 = sqrt(v2x*v2x + v2y*v2y);
	v2x/=n2;	v2y/=n2;
	
	double x01 = (v1x * v2x + v1y * v2y) * n2 / n1;
	double y01 = (v1y * v2x - v1x * v2y) * n2 / n1;
	FInv = new SymMatrix2DReal(4, 4);
	w = new SymVectorReal(4);
	C = new SymVectorReal(4);
	
	FillMatrixF(x01, y01);
	
	GetFittedTriangle(x0, y0, x1, y1, x2, y2);
	double relative0 = 0.000001;
	assert(qAbs(x0-x0f)<relative0 && qAbs(y0-y0f)<relative0 && qAbs(x1-x1f)<relative0 
		   && qAbs(y1-y1f)<relative0 && qAbs(x2-x2f)<relative0 && qAbs(y2-y2f)<relative0);
}

void AsRigidAsPossibleLevelDescriptor::FittedTriangle::FillMatrixF(double x01, double y01)
{
	this->x01 = x01;
	this->y01 = y01;

	//dE/dx0f
	(*FInv)(0, 0) = 4 - 4*x01 + 2*x01*x01 + 2*y01*y01;		//x0f
	(*FInv)(0, 1) = 0;										//y0f
	(*FInv)(0, 2) = 2*x01 - 2*y01*y01 - 2*x01*x01;			//x1f
	(*FInv)(0, 3) = 2*y01;									//y1f
	
	//dE/dy0f
	(*FInv)(1, 0) = 0;										//x0f
	(*FInv)(1, 1) = 4 - 4*x01+2*y01*y01 + 2*x01*x01;		//y0f
	(*FInv)(1, 2) = -2*y01;									//x1f
	(*FInv)(1, 3) = -2*y01*y01 + 2*x01 - 2*x01*x01;			//y1f

	//dE/dx1f
	(*FInv)(2, 0) = 2*x01 - 2*x01*x01 - 2*y01*y01;			//x0f
	(*FInv)(2, 1) = -2*y01;									//y0f
	(*FInv)(2, 2) = 2 + 2*x01*x01 + 2*y01*y01;				//x1f
	(*FInv)(2, 3) = 0;										//y1f

	//dE/dy1f
	(*FInv)(3, 0) = 2*y01;									//x0f
	(*FInv)(3, 1) = 2*x01 - 2*y01*y01 - 2*x01*x01;			//y0f
	(*FInv)(3, 2) = 0;										//x1f
	(*FInv)(3, 3) = 2 + 2*y01*y01 + 2*x01*x01;				//y1f
	
	(*FInv).Inverse();
}

void AsRigidAsPossibleLevelDescriptor::FittedTriangle::GetFittedTriangle(double x0p, double y0p, double x1p, double y1p, double x2p, double y2p)
{
	//Fill C
	(*C)(0) = -1*(-2*x0p - 2*x2p + 2*x2p*x01 - 2*y01*y2p);		//dE/dx0f
	(*C)(1) = -1*(-2*y0p + 2*y01*x2p - 2*y2p + 2*y2p*x01);		//dE/dy0f
	(*C)(2) = -1*(-2*x1p - 2*x2p*x01 + 2*y01*y2p);				//dE/dx1f
	(*C)(3) = -1*(-2*y1p - 2*y01*x2p - 2*y2p*x01);				//dE/dy1f	
	
	// w = FInv * -C;
	FInv->Multiply(*C, *w);
	
	x0f = (*w)(0);
	y0f = (*w)(1);
	x1f = (*w)(2);
	y1f = (*w)(3);	
	x2f = x0f + x01*(x1f-x0f) + y01*(y1f-y0f);
	y2f = y0f + x01*(y1f-y0f) + y01*(-x1f+x0f);	
	
	double dxf = x0f-x1f;		double dyf = y0f-y1f;
	double dxp = x0p-x1p;		double dyp = y0p-y1p;	
	double scaleFactor = sqrt(dxf*dxf + dyf*dyf) / sqrt(dxp*dxp + dyp*dyp);
	
	double mx = (x0f + x1f + x2f)/3.;
	double my = (y0f + y1f + y2f)/3.;
	
	dxf = x0f - mx;						dyf = y0f - my;
	x0f = mx + dxf * scaleFactor;		y0f = my + dyf * scaleFactor;	

	dxf = x1f - mx;						dyf = y1f - my;
	x1f = mx + dxf * scaleFactor;		y1f = my + dyf * scaleFactor;	

	dxf = x2f - mx;						dyf = y2f - my;
	x2f = mx + dxf * scaleFactor;		y2f = my + dyf * scaleFactor;	
}

AsRigidAsPossibleLevelDescriptor::AsRigidAsPossibleLevelDescriptor(int constrainedVars, 
										int freeVars, std::vector<DeformationHierarchyNode*> & breadthFirstSorted)
{
//	std::cout<<"Creating new level descriptor"<<std::endl;
//	std::cout<<"freeVars="<<freeVars<<", constraintedVars="<<constrainedVars<<std::endl;
	assert(freeVars%2==0 && constrainedVars%2==0);
	assert(freeVars>0 && constrainedVars>0);
	m_iFreeVars = freeVars;
	m_iConstrainedVars = constrainedVars;
	m_iFreeVertices = freeVars/2;
	m_iConstrainedVertices = constrainedVars/2;
	
	m_aParamMapConstrainted = new int[m_iConstrainedVars];
	m_aParamMapFree = new int[m_iFreeVars];	
	m_aParamMapToG = new int[m_iFreeVars+m_iConstrainedVars];
	
	m_pGInverseB = new SymMatrix2DReal(m_iFreeVars, m_iConstrainedVars);
	m_pConstrainedParams = new SymVectorReal(m_iConstrainedVars);
	m_pFreeParams = new SymVectorReal(m_iFreeVars);
	
	m_pHpCholeskyDecomp = new SymMatrix2DReal(m_iFreeVars, m_iFreeVars, 0);
	m_pD = new SymMatrix2DReal(m_iFreeVars, m_iConstrainedVars, 0);
	m_pf0 = new SymVectorReal(m_iFreeVars);
	m_pf0AndDq = new SymVectorReal(m_iFreeVars);
	

	assert((int)breadthFirstSorted.size() == (m_iConstrainedVertices + m_iFreeVertices));	
	for (int i=0; i<m_iFreeVertices; i++)
	{
		m_aParamMapFree[i*2+0] = breadthFirstSorted[m_iConstrainedVertices + i]->m_ParamIDs[0];
		m_aParamMapFree[i*2+1] = breadthFirstSorted[m_iConstrainedVertices + i]->m_ParamIDs[1];	
		m_aParamMapToG[breadthFirstSorted[m_iConstrainedVertices + i]->m_ParamIDs[0]] = i*2+0;
		m_aParamMapToG[breadthFirstSorted[m_iConstrainedVertices + i]->m_ParamIDs[1]] = i*2+1;		
	}
	
	for (int i=0; i<m_iConstrainedVertices; i++)
	{
		m_aParamMapConstrainted[i*2+0] = breadthFirstSorted[i]->m_ParamIDs[0];
		m_aParamMapConstrainted[i*2+1] = breadthFirstSorted[i]->m_ParamIDs[1];		
		m_aParamMapToG[breadthFirstSorted[i]->m_ParamIDs[0]] = m_iFreeVars + i*2+0;
		m_aParamMapToG[breadthFirstSorted[i]->m_ParamIDs[1]] = m_iFreeVars + i*2+1;		
	}
//	std::cout<<"Creating new level descriptor - Done!"<<std::endl;	
}

void AsRigidAsPossibleLevelDescriptor::InitializeMatrices(std::vector<ImageWarp::ImgVertex> & vertices, std::vector<ImageWarp::ImgFace> & faces)
{
					///////////// STEP 1 - scale free ///////////////
	assert((int)vertices.size()==(m_iConstrainedVertices+m_iFreeVertices));
	SymMatrix2DReal G(m_iFreeVars+m_iConstrainedVars, m_iFreeVars+m_iConstrainedVars);
	
	for (int f=0; f<(int)faces.size(); f++)
	{
		ImageWarp::ImgFace & face = faces[f];
		assert(face.v4==-1);	//must be triangles
		AddCoeffsForTriangleStep1(G, vertices[face.v1], vertices[face.v2], vertices[face.v3]);
		AddCoeffsForTriangleStep1(G, vertices[face.v1], vertices[face.v3], vertices[face.v2]);
		AddCoeffsForTriangleStep1(G, vertices[face.v2], vertices[face.v3], vertices[face.v1]);		
	}
	
	SymMatrix2DReal G00(m_iFreeVars, m_iFreeVars);
	SymMatrix2DReal G01(m_iFreeVars, m_iConstrainedVars);
	SymMatrix2DReal G10(m_iConstrainedVars, m_iFreeVars);	
	//Fill G00, G01, G10
	for (int i=0; i<m_iFreeVars; i++)
	{
		for (int j=0; j<m_iFreeVars; j++)
			G00(i,j) = G(i,j);
		for (int j=0; j<m_iConstrainedVars; j++)
		{
			G01(i,j) = G(i, j+m_iFreeVars);
			G10(j,i) = G(j+m_iFreeVars, i);
		}
	}
	
	//Fill Gp and B
	SymMatrix2DReal Gp(m_iFreeVars, m_iFreeVars, 0);			//Gp = G00 + G00'
	SymMatrix2DReal B(m_iFreeVars, m_iConstrainedVars, 0);		//B  = G01 + G10'
	G00.Transpose(Gp);
	Gp.Add(G00);
	G10.Transpose(B);
	B.Add(G01);
	
	//Find Gp^-1
	Gp.Inverse();

	//Set m_pGInverseB = -Gp^-1*B
	Gp.DivideBy(-1);
	Gp.Multiply(B, *m_pGInverseB);
	
					///////////// STEP 2.1 fitting individual triangles ///////////////	
	//Fill F^-1 for each triangle fit
	for (int i=0; i<(int)faces.size(); i++)
	{
		//find param id's for triangles		
		int id0 = m_aParamMapToG[2*vertices[faces[i].v1].m_iParamID];
		int id1 = m_aParamMapToG[2*vertices[faces[i].v2].m_iParamID];
		int id2 = m_aParamMapToG[2*vertices[faces[i].v3].m_iParamID];
		FittedTriangle * newTriangle = new FittedTriangle(vertices[faces[i].v1], vertices[faces[i].v2], vertices[faces[i].v3], id0, id1, id2);
		m_FittedTriangles.push_back(newTriangle);
	}
	
					///////////// STEP 2.2 obtaining scale-adjusted result ///////////////	
	//Find H
	SymMatrix2DReal H(m_iFreeVars+m_iConstrainedVars, m_iFreeVars+m_iConstrainedVars);	
	for (int i=0; i<(int)m_FittedTriangles.size(); i++)	
		AddCoeffsForTriangleStep2(H, *(m_FittedTriangles[i]));
	
	//Fill H00, H00', H01, H10'
	SymMatrix2DReal H00(m_iFreeVars, m_iFreeVars);
	SymMatrix2DReal H01(m_iFreeVars, m_iConstrainedVars);
	SymMatrix2DReal H10(m_iConstrainedVars, m_iFreeVars);	
	for (int i=0; i<m_iFreeVars; i++)
	{
		for (int j=0; j<m_iFreeVars; j++)
			H00(i,j) = H(i,j);
		for (int j=0; j<m_iConstrainedVars; j++)
		{
			H01(i,j) = H(i, j+m_iFreeVars);
			H10(j,i) = H(j+m_iFreeVars, i);
		}
	}		
	
	//find Hp, D
	H00.Transpose(*m_pHpCholeskyDecomp);
	m_pHpCholeskyDecomp->Add(H00);
	H10.Transpose(*m_pD);
	m_pD->Add(H01);
	
	//Decompose Hp
	m_pHpCholeskyDecomp->DecomposeCholesky();
}

void AsRigidAsPossibleLevelDescriptor::FillOutputParams(SymVectorReal & outParamVector)
{
  	for (int i=0; i<m_pFreeParams->m; i++)
		outParamVector(m_aParamMapFree[i]) = (*m_pFreeParams)(i);

  	for (int i=0; i<m_pConstrainedParams->m; i++)
		outParamVector(m_aParamMapConstrainted[i]) = (*m_pConstrainedParams)(i);
}

void AsRigidAsPossibleLevelDescriptor::FillFreeParams(bool preserveScale)
{
	// Step 1: u = m_pGInverseB * q		(scale-free construction)
	m_pGInverseB->Multiply(*m_pConstrainedParams, *m_pFreeParams);
	
	if (preserveScale)
	{
		//Step 2.1: Per triangle: w=F^-1 * (-C)		(fit triangles)
		m_pf0AndDq->SetAll(0);
		m_pf0->SetAll(0);			
		for (int i=0; i<(int)m_FittedTriangles.size(); i++)
		{
			//find local params for triangle vertices
			int paramID0 = m_FittedTriangles[i]->m_iParamID0; 
			int paramID1 = m_FittedTriangles[i]->m_iParamID1;
			int paramID2 = m_FittedTriangles[i]->m_iParamID2;

			//find triangle vertex locations calculated at step 1
			double x0 = (paramID0 >= m_iFreeVars) ? (*m_pConstrainedParams)(paramID0 - m_iFreeVars + 0) : (*m_pFreeParams)(paramID0 + 0);
			double y0 = (paramID0 >= m_iFreeVars) ? (*m_pConstrainedParams)(paramID0 - m_iFreeVars + 1) : (*m_pFreeParams)(paramID0 + 1);
			double x1 = (paramID1 >= m_iFreeVars) ? (*m_pConstrainedParams)(paramID1 - m_iFreeVars + 0) : (*m_pFreeParams)(paramID1 + 0);
			double y1 = (paramID1 >= m_iFreeVars) ? (*m_pConstrainedParams)(paramID1 - m_iFreeVars + 1) : (*m_pFreeParams)(paramID1 + 1);
			double x2 = (paramID2 >= m_iFreeVars) ? (*m_pConstrainedParams)(paramID2 - m_iFreeVars + 0) : (*m_pFreeParams)(paramID2 + 0);
			double y2 = (paramID2 >= m_iFreeVars) ? (*m_pConstrainedParams)(paramID2 - m_iFreeVars + 1) : (*m_pFreeParams)(paramID2 + 1);		
			
			//fit triangles
			m_FittedTriangles[i]->GetFittedTriangle(x0, y0, x1, y1, x2, y2);
			
			AddCoeffsForTriangleStep2(*m_pf0, *(m_FittedTriangles[i]));	
		}

		//Step 2.2: H' * u = -f0-D*q		(scale adjusted result). Use LU decomp to solve for u
		m_pD->Multiply(*m_pConstrainedParams, *m_pf0AndDq);
		m_pf0AndDq->Add(*m_pf0);
		m_pf0AndDq->Multiply(-1);
		m_pHpCholeskyDecomp->SolveUsingCholesky(*m_pf0AndDq, *m_pFreeParams);
	}	
}

void AsRigidAsPossibleLevelDescriptor::FillConstrainedParams(SymVectorReal & inParamVector)
{
  	for (int i=0; i<m_pConstrainedParams->m; i++)
		(*m_pConstrainedParams)(i) = inParamVector(m_aParamMapConstrainted[i]);
}

void AsRigidAsPossibleLevelDescriptor::AddValueG(SymMatrix2DReal & G,  ImageWarp::ImgVertex & v1, ImageWarp::ImgVertex & v2, int v1_xory, int v2_xory, double coefficient)
{
	int paramID1 = 2*v1.m_iParamID + v1_xory;
	int paramID2 = 2*v2.m_iParamID + v2_xory;	
	assert(paramID1 < m_iFreeVars+m_iConstrainedVars && paramID2 < m_iFreeVars + m_iConstrainedVars);
	int idInG1 = m_aParamMapToG[paramID1];
	int idInG2 = m_aParamMapToG[paramID2];	
	assert(idInG1 < G.m && idInG2<G.m && G.m==G.n);
	if (idInG1==idInG2)
	{
		G(idInG1, idInG2) += coefficient;
	}
	else
	{
		G(idInG1, idInG2) += coefficient;
		G(idInG2, idInG1) += coefficient;		
	}
}

/**
 * This function adds energy penalties due to vertex v2 in triangle v0, v1, v2. 
 * After all triangles are added energy should be E = v' * G * v, where v=new vertex locations.
 * See equation (5) in section 4.1, of As-Rigid-As-Possible Shape Manipulation by Igarashi et al.
 */
void AsRigidAsPossibleLevelDescriptor::AddCoeffsForTriangleStep1(SymMatrix2DReal & G, ImageWarp::ImgVertex & v0, 
																 ImageWarp::ImgVertex & v1, ImageWarp::ImgVertex & v2)
{
	//Find: x01, y01
	double v1x = v1.m_fAsRigidRestX - v0.m_fAsRigidRestX;
	double v1y = v1.m_fAsRigidRestY - v0.m_fAsRigidRestY;
	double n1 = sqrt(v1x*v1x + v1y*v1y);
	v1x/=n1;	v1y/=n1;

	double v2x = v2.m_fAsRigidRestX - v0.m_fAsRigidRestX;
	double v2y = v2.m_fAsRigidRestY - v0.m_fAsRigidRestY;
	double n2 = sqrt(v2x*v2x + v2y*v2y);
	v2x/=n2;	v2y/=n2;
	
	double x01 = (v1x * v2x + v1y * v2y) * n2 / n1;
	double y01 = (v1y * v2x - v1x * v2y) * n2 / n1;
		
	assert(qAbs(-v2.m_fAsRigidRestX+v0.m_fAsRigidRestX+x01*(v1.m_fAsRigidRestX - v0.m_fAsRigidRestX) + y01*(v1.m_fAsRigidRestY - v0.m_fAsRigidRestY))<0.000001);
	assert(qAbs(-v2.m_fAsRigidRestY+v0.m_fAsRigidRestY+x01*(v1.m_fAsRigidRestY - v0.m_fAsRigidRestY) + y01*(-v1.m_fAsRigidRestX + v0.m_fAsRigidRestX))<0.000001);
	
	int VX = 0;
	int VY = 1;
	//set x-coord
		//quadratic terms
	AddValueG(G, v0, v0, VX, VX,	(1-x01)*(1-x01));
	AddValueG(G, v1, v1, VX, VX,	x01*x01);
	AddValueG(G, v1, v1, VY, VY,	y01*y01);
	AddValueG(G, v0, v0, VY, VY,	y01*y01);
	AddValueG(G, v2, v2, VX, VX,	1);	
		//2 * terms 
	AddValueG(G, v0, v1, VX, VX,	(1-x01)*x01);		//starts with x0
	AddValueG(G, v0, v1, VX, VY,	(1-x01)*y01);
	AddValueG(G, v0, v0, VX, VY,	(1-x01)*-y01);
	AddValueG(G, v0, v2, VX, VX,	(1-x01)*-1);	
	AddValueG(G, v1, v1, VX, VY,	x01*y01);			//starts with x1
	AddValueG(G, v1, v0, VX, VY,	x01*-y01);
	AddValueG(G, v1, v2, VX, VX,	x01*-1);
	AddValueG(G, v1, v0, VY, VY,	y01*-y01);			//starts with y1
	AddValueG(G, v1, v2, VY, VX,	y01*-1);
	AddValueG(G, v0, v2, VY, VX,	-y01*-1);			//starts with y0
	
	//set y-coord
		//quadratic terms
	AddValueG(G, v0, v0, VY, VY,	(1-x01)*(1-x01));
	AddValueG(G, v1, v1, VY, VY,	x01*x01);
	AddValueG(G, v0, v0, VX, VX,	y01*y01);
	AddValueG(G, v1, v1, VX, VX,	y01*y01);
	AddValueG(G, v2, v2, VY, VY,	1);	
		//2 * terms 
	AddValueG(G, v0, v1, VY, VY,	(1-x01)*x01);		//starts with y0
	AddValueG(G, v0, v0, VY, VX,	(1-x01)*y01);
	AddValueG(G, v0, v1, VY, VX,	(1-x01)*-y01);
	AddValueG(G, v0, v2, VY, VY,	(1-x01)*-1);
	AddValueG(G, v1, v0, VY, VX,	x01*y01);			//starts with y1
	AddValueG(G, v1, v1, VY, VX,	x01*-y01);
	AddValueG(G, v1, v2, VY, VY,	x01*-1);
	AddValueG(G, v0, v1, VX, VX,	y01*-y01);			//starts with x0
	AddValueG(G, v0, v2, VX, VY,	y01*-1);
	AddValueG(G, v1, v2, VX, VY,	-y01*-1);			//starts with x1	
}

void AsRigidAsPossibleLevelDescriptor::AddCoeffsForTriangleStep2(SymMatrix2DReal & H, FittedTriangle & triangle)
{
	int x0 = triangle.m_iParamID0+0;		int y0 = triangle.m_iParamID0+1;
	int x1 = triangle.m_iParamID1+0;		int y1 = triangle.m_iParamID1+1;
	int x2 = triangle.m_iParamID2+0;		int y2 = triangle.m_iParamID2+1;
	
	H(x0, x0)+=2;	H(y0, y0)+=2;		H(x1, x1)+=2;	H(y1, y1)+=2;		H(x2, x2)+=2;	H(y2, y2)+=2;	//squared
	H(x1, x0)-=1;	H(x0, x1)-=1;		H(y1, y0)-=1;	H(y0, y1)-=1;	//all multiple of 2 symmetric terms
	H(x2, x1)-=1;	H(x1, x2)-=1;		H(y2, y1)-=1;	H(y1, y2)-=1;			
	H(x2, x0)-=1;	H(x0, x2)-=1;		H(y2, y0)-=1;	H(y0, y2)-=1;
}

void AsRigidAsPossibleLevelDescriptor::AddCoeffsForTriangleStep2(SymVectorReal & f0, FittedTriangle & triangle)
{
	int x0 = triangle.m_iParamID0+0;		int y0 = triangle.m_iParamID0+1;
	int x1 = triangle.m_iParamID1+0;		int y1 = triangle.m_iParamID1+1;
	int x2 = triangle.m_iParamID2+0;		int y2 = triangle.m_iParamID2+1;

	if (x0<f0.m)
		f0(x0) += -4*triangle.x0f + 2*triangle.x1f  + 2*triangle.x2f;

	if (x1<f0.m)
		f0(x1) += -4*triangle.x1f + 2*triangle.x0f + 2*triangle.x2f;

	if (x2<f0.m)
		f0(x2) += -4*triangle.x2f + 2*triangle.x0f + 2*triangle.x1f;
	
	if (y0<f0.m)
		f0(y0) += -4*triangle.y0f + 2*triangle.y1f + 2*triangle.y2f;

	if (y1<f0.m)
		f0(y1) += -4*triangle.y1f + 2*triangle.y0f + 2*triangle.y2f;

	if (y2<f0.m)
		f0(y2) += -4*triangle.y2f + 2*triangle.y0f + 2*triangle.y1f;
	
}


void AsRigidAsPossibleLevelDescriptor::DrawControlPoints(QImage & trg, bool inverse, std::vector<ImageWarp::ImgVertex> & vertices)
{
	for (int i=0; i<m_iFreeVertices; i++)
	{
		double x = (*m_pFreeParams)(2*i+0);
		double y = (*m_pFreeParams)(2*i+1);		
		if (inverse)
		{
			x = vertices[m_aParamMapFree[2*i]/2].tu;
			y = vertices[m_aParamMapFree[2*i]/2].tv;
		}		
		DrawingRoutines::DrawCircle(trg, x / trg.width(), y / trg.height(), 0.005, 0, qRgb(0, 0, 220), qRgb(0,0,0));
	}	
	
	for (int i=0; i<m_iConstrainedVertices; i++)
	{
		double x = (*m_pConstrainedParams)(2*i+0);
		double y = (*m_pConstrainedParams)(2*i+1);		
		if (inverse)
		{
			x = vertices[m_aParamMapConstrainted[2*i]/2].tu;
			y = vertices[m_aParamMapConstrainted[2*i]/2].tv;
		}
		DrawingRoutines::DrawCircle(trg, x / trg.width(), y / trg.height(), 0.01, 0, qRgb(0, 0, 255), qRgb(0,0,0));
	}	
}

/************************* Code For ImageWarpLinearMesh ***************************/
void ImageWarpLinearMesh::CreateAsRigidAsPossibleHierarchy(int * numNodes, ImageHierarchyDescriptor * hierarchyDescr, 
														   double initStep, double minParamStep)
{
	assert(m_Vertices.size()>0);
	m_iHierarchyLevel=-1;
	//no need for extra control points, the only thing that matters is which vertices are control points.
	*numNodes = (int)m_Vertices.size();
	
	int minPntsInHierarchy = hierarchyDescr->m_iMinPnts;
	int maxPntsInHierarchy = qMin(hierarchyDescr->m_iMaxPnts,  (int)m_Vertices.size()-1);
	
	// Select control points. Sort vertices according to 'hierarchy'. 
//	std::cout<<"\tGeodesic sorting (TODO: speed up using Dijkstra algorithm): "<<std::flush;
//	std::vector<int> verticesByImportance = GetHierarchicalSortingByGeodesic();
	std::vector<int> verticesByImportance;
	for (int i=0; i<(int)m_Vertices.size(); i++)	//just for debugging let vertices have same sorting
		verticesByImportance.push_back(i);
	assert(verticesByImportance.size()==m_Vertices.size());
	
	m_iMaxHierarchyLevels = 0;
	int pntsAtCurrentLevel = minPntsInHierarchy;
	int numSteps = 0;
	assert((int)m_Vertices.size()>0);
	m_pAllLastParams = new SymVectorReal(2 * (int)m_Vertices.size());
	
//	for (int i=0; i<(int)m_Vertices.size(); i++)
//		m_pHierarchyBreadthFirstList.push_back(NULL);
	// Fill m_pHierarchyBreadthFirstList=[sorted by 'importance'] control points go first
	for (int i=0; i<(int)m_Vertices.size(); i++)
	{		
		m_Vertices[i].m_fAsRigidRestX = m_Vertices[i].x;
		m_Vertices[i].m_fAsRigidRestY = m_Vertices[i].y;		
		(*m_pAllLastParams)(2*i+0) = m_Vertices[i].x;
		(*m_pAllLastParams)(2*i+1) = m_Vertices[i].y;		
		
		DeformationHierarchyNode * node = new DeformationHierarchyNode();
		node->m_ParamIDs.push_back(2*verticesByImportance[i]+0);
		node->m_ParamIDs.push_back(2*verticesByImportance[i]+1);		
		//m_pHierarchyBreadthFirstList[(int)m_Vertices.size() - i - 1] = node;
		m_pHierarchyBreadthFirstList.push_back(node);
	}
	//fill steps, hierarchical single layers
	std::vector<double> minSteps;
	std::vector<int> suggestedSteps;
	std::vector<double> distinctSteps;
	while (pntsAtCurrentLevel <= maxPntsInHierarchy)	//find number of hierarchy levels
	{
		m_iMaxHierarchyLevels++;
		pntsAtCurrentLevel*=2;
	}		
	pntsAtCurrentLevel = minPntsInHierarchy;
	int depth=-1;
	while (pntsAtCurrentLevel <= maxPntsInHierarchy)	//fill params
	{
		depth++;
		m_iLayerToFirstElement.push_back(0);
		m_iLayerToMaxElements.push_back(pntsAtCurrentLevel);
//		std::cout<<"pnts["<<m_iMaxHierarchyLevels<<"] = "<<pntsAtCurrentLevel<<std::endl;
		
		std::vector<int> nodes;		nodes.push_back(2*numSteps);		nodes.push_back(2*numSteps+1);
		m_ParamPrintNodes.push_back(HierarchyPrintNode("", nodes, m_iMaxHierarchyLevels - depth - 1));
		
		//fill per level param maps for each parameter
		for (int i=0; i<(int)m_pHierarchyBreadthFirstList.size(); i++)
		{
			DeformationHierarchyNode * node = m_pHierarchyBreadthFirstList[i];
			node->m_LevelParamStepIDs.push_back(std::vector<int>());
			for (int i=0; i<pntsAtCurrentLevel; i++)
			{
				node->m_LevelParamStepIDs[node->m_LevelParamStepIDs.size()-1].push_back(2*(numSteps+i)+0);
				node->m_LevelParamStepIDs[node->m_LevelParamStepIDs.size()-1].push_back(2*(numSteps+i)+1);				
			}
		}

		//fill steps
		for (int i=0; i<2*pntsAtCurrentLevel; i++)
		{
			suggestedSteps.push_back(distinctSteps.size());
			minSteps.push_back(minParamStep);
		}
		
		numSteps += pntsAtCurrentLevel;
		pntsAtCurrentLevel *= 2;	//suppose each next layer has twice more points
		distinctSteps.push_back(initStep);		
		initStep *= sqrt(2);		//TODO: relate to bounding box
//		std::cout<<"initStep="<<initStep<<std::endl;
	}
	assert((int)minSteps.size()==numSteps*2 && (int)suggestedSteps.size()==numSteps*2);
	assert(numSteps>0);
	m_pAllParamStep = new SymVectorReal(2*numSteps);
	m_pAllMinParamStep = new SymVectorReal(2*numSteps);	
	
	//copy step values into vectors of all parameters
	for (int i=0; i<2*numSteps; i++)
	{
		(*m_pAllMinParamStep)(i) = minSteps[i];
		(*m_pAllParamStep)(i) = distinctSteps[distinctSteps.size()-1-suggestedSteps[i]];		
	}
//	std::cout<<"minParamSteps=";	m_pAllMinParamStep->PrintVector(0.0001);	std::cout<<std::endl;
//	std::cout<<"suggestedSteps=";	m_pAllParamStep->PrintVector(0.0001);	std::cout<<std::endl;	

	std::cout<<"\tInitializing As-Rigid-As-Possible deformation matrices: "<<m_iMaxHierarchyLevels<<" : "<<std::flush;	
	//create as-rigid-as-possible hierarchy levels
	for (int i=0; i<m_iMaxHierarchyLevels; i++)
	{
		int constrained = m_iLayerToMaxElements[i]*2;
		int free = 2*m_Vertices.size() - constrained;
		AsRigidAsPossibleLevelDescriptor * asRigidLevel = new AsRigidAsPossibleLevelDescriptor( constrained, free, m_pHierarchyBreadthFirstList);
		asRigidLevel->InitializeMatrices(m_Vertices, m_Faces);
		m_AsRigidHierarchyLevels.push_back(asRigidLevel);
		std::cout<<"."<<std::flush;
	}
	std::cout<<std::endl;
}

/**
 * Lists m_Vertices indices according to 'importance', e.g. picks center one first, and then
 * iteratively picks the one that is furthest from the other vertices (in geodesic sense)
 * NOTE 1: complexity is O(N^3), where N = # vertices
 * NOTE 2: this algorithm assumes triangle inequality
 * NOTE 3: Highly inefficient. Should use Dijkstar Algorithm O(N^2) (repeat for each vertex)
 */
std::vector<int> ImageWarpLinearMesh::GetHierarchicalSortingByGeodesic()
{
	TimeProfiler profiler;
	profiler.startedProcess("SortingByGeodesic");		
	std::vector<int> sortedVertices;
	std::vector<bool> visited;
	std::vector<bool> visitedForBreadthFirstTraversal;	
	
	// find center vertex
	double ctrX = ((double)m_Source.width())/2.;
	double ctrY = ((double)m_Source.height())/2.;
	int bestVertex = -1;
	double bestDist = -1;
	for (int i=0; i<(int)m_Vertices.size(); i++)
	{
		visited.push_back(false);
		visitedForBreadthFirstTraversal.push_back(false);
		double dx = m_Vertices[i].x - ctrX;
		double dy = m_Vertices[i].y - ctrY;
		if (dx*dx + dy*dy < bestDist || bestDist<0)
		{
			bestVertex = i;
			bestDist = dx*dx + dy*dy;
		}
	}
	
	visited[bestVertex] = true;
	sortedVertices.push_back(bestVertex);
	
	int revealed = -1;
	// iteratively find vertices furtherst (geoedsically) from all visited vertices
	while (sortedVertices.size() < m_Vertices.size())	//add vertices one-by-one
	{
		if ((revealed+1) == (int)((sortedVertices.size()*20) / m_Vertices.size()))
		{
			std::cout<<"."<<std::flush;
			revealed++;
		}
		bestDist = -1;
		bestVertex = -1;
		for (int i=0; i<(int)m_Vertices.size(); i++)	//loop through non-visited vertices
		{
			if (!visited[i] && sortedVertices.size()==m_Vertices.size()-1)	//terminal step
			{
				bestVertex = i;
				break;
			}
			else if (!visited[i])	//Calculate minimal of distances to all other vertices
			{
				double currDist = -1;
				double bestCurrDist = -1;
				for (int v=0; v<(int)m_Vertices.size(); v++)
					visitedForBreadthFirstTraversal[v] = false;
				visitedForBreadthFirstTraversal[i] = true;
				//using breadth-first search find geodesic distance to all visited nodes
				std::vector<int> bfsList;
				std::vector<double> bfsDistances;
				bfsList.push_back(i);
				bfsDistances.push_back(0);
				int bfsID = 0;
				while (bfsID<(int)bfsList.size())	//for neighbors of the current vertex (i.e. all other vertices)
				{
					int currID = bfsList[bfsID];	//add all children
					for (int chld=0; chld<(int)m_Vertices[currID].m_AdjacentVertices.size(); chld++)	//loop through children
					{
						int childID = m_Vertices[currID].m_AdjacentVertices[chld];
						if (!visitedForBreadthFirstTraversal[childID])	// if child already added - do not add again
						{
							double dx = m_Vertices[childID].x - m_Vertices[i].x;
							double dy = m_Vertices[childID].y - m_Vertices[i].y;								
							currDist = bfsDistances[bfsID] + sqrt(dx*dx + dy*dy);
							
							if (visited[childID])	//if child is marked red (i.e. visited) - find best distance
							{
								if (currDist < bestCurrDist || bestCurrDist<0)
								{
									bestCurrDist = currDist;
									//bestVertex = i;
								}
							}
							else
							{
								bfsList.push_back(childID);
								bfsDistances.push_back(currDist);
							}
							visitedForBreadthFirstTraversal[childID] = true;
						}
					}
					bfsID++;
				}
				
				assert(bestCurrDist>=0);
				
				if (bestDist<0 || bestCurrDist > bestDist)
				{
					bestDist = bestCurrDist; 
					bestVertex = i;
				}
			}
		}

		visited[bestVertex] = true;
		sortedVertices.push_back(bestVertex);
	}
	profiler.finishedProcess("SortingByGeodesic", true);		
	return sortedVertices;
}

void ImageWarpLinearMesh::GenHierarchicalParamsAsRigidAsPossible(SymVectorReal & params, SymVectorReal & trg)
{
//	std::cout<<"InParams=";		params.PrintVector(0.0001);		std::cout<<std::endl;	
	assert(m_iHierarchyLevel>=0);
	AsRigidAsPossibleLevelDescriptor * asRigidLevel = m_AsRigidHierarchyLevels[m_iHierarchyLevel];
	asRigidLevel->FillConstrainedParams(params);
	asRigidLevel->FillFreeParams(m_bPreserveScale);
	if (m_bShowWireframe && m_bRetrievingImage)
	{
		if (m_ControlPntsImg.width()!=m_Deformed.width() || m_ControlPntsImg.height()!=m_Deformed.height())
			m_ControlPntsImg = QImage(m_Deformed.width(), m_Deformed.height(), QImage::Format_RGB32);
		DrawingRoutines::ClearImage(m_ControlPntsImg, qRgb(0, 0, 0));
		asRigidLevel->DrawControlPoints(m_ControlPntsImg, m_bInverseMapping, m_Vertices);
	}	
	asRigidLevel->FillOutputParams(trg);
//	std::cout<<"OutParams=";	trg.PrintVector(0.0001);		std::cout<<std::endl;	
}

QImage ImageWarpLinearMesh::DrawGeodesicSorting()
{
	SymVectorReal &params = GetParameterVector();
	DeformAtCurrentHierarchyLevel(params);
	QImage img = GetDeformedImage();
	double minRad = 1. / qMax(img.width(), img.height());
	double maxRad = 20. / qMax(img.width(), img.height());
	double red = 1.2;
	int maxSize=4;
	int numDrawCircles = qMin(20, (int)m_pHierarchyBreadthFirstList.size());
	for (int i=0; i<numDrawCircles; i++)
	{
		double x = (*m_pFinalParams)(m_pHierarchyBreadthFirstList[i]->m_ParamIDs[0]);
		double y = (*m_pFinalParams)(m_pHierarchyBreadthFirstList[i]->m_ParamIDs[1]);		
		if (m_bInverseMapping)
		{
			x = m_Vertices[m_pHierarchyBreadthFirstList[i]->m_ParamIDs[0]/2].tu;
			y = m_Vertices[m_pHierarchyBreadthFirstList[i]->m_ParamIDs[0]/2].tv;			
		}
		DrawingRoutines::DrawCircle(img, x / img.width(), y / img.height(), maxRad, 0, qRgb(255 - (i%maxSize)*196/maxSize, 0, (i%maxSize)*196/maxSize), qRgb(0,0,0));
		
//		maxRad = qMax(maxRad - red, minRad);		
		maxRad = qMax(maxRad/red, minRad);		
	}
	return img;
}



