#include "Numerical.h"
#include "Deformations.h"
#include "R2Shapes/R2Shapes.h"
#include "R3Shapes/R3Shapes.h"
#include "R3Graphics/R3Graphics.h"
#include <QtOpenGL/QGLContext>
#include <QtOpenGL/QGLFramebufferObject>
#include <QtOpenGL/QGLPixelBuffer>
#include <QtOpenGL/QGLFormat>
#include <Qt/qimage.h>

#ifndef __DEFORM_TEXTURE_SYNTH_H
#define __DEFORM_TEXTURE_SYNTH_H

class DeformTextureSynth : public DeformationModel
{
	public:
		struct DynProgCell
		{
			double m_fCost;
			int m_iPrevCell;
		};
	
		struct SymSeam
		{
			enum SeamType
			{
				HORIZONTAL_SEAM,
				VERTICAL_SEAM
			};
			
			int m_iWidth;
			int m_iHeight;
			
			int m_iPatchID;
			int m_iIntersection1;
			int m_iIntersection2;			
			
			std::vector<int> m_SeamBoundary;
//			std::vector<double> m_SeamCost;				//TODO: not using me
		};
	
		enum TextureSynthAlgorithm
		{
			IMAGE_QUILTING
		};
	
		DeformTextureSynth(QImage & src, int resWidth, int resHeight, int patchWidth, int patchHeight, 
						   int overlapWidth, int overlapHeight, int optimNumPatches, int optimNumBestPatches,
						   bool minCutCarving);
		virtual ~DeformTextureSynth(){}	
	
		virtual bool DistortsInput();
		virtual bool HasOutput(OutputType type);
		virtual void GetOutputDimensions(int * inWidth, int * inHeight);
		virtual SymMatrix & Deform(SymVectorReal & params, double * outPenalty = NULL);
		virtual QImage GetDeformedImage(int imageType = 0);
		virtual SymMatrix & TakeRandomStep(bool & enforceStep);
		virtual void RevertLastRandomStep();
		virtual void UpdateOnlyOnePatch(int patchID);
		int m_iOldPatchX;
		int m_iOldPatchY;

		virtual void FillVerticalScanlines(int patchID);
		virtual void FillHorizontalScanlines(int patchID);
		virtual void FillBottomRightCorner(int patchID);

		static void GetMinCutSeam(std::vector<DynProgCell> & overlap, SymSeam & trgSeam);	
		static double GetColorDifference(QImage & img1, int i1x, int i1y, QImage & img2, int i2x, int i2y);
		static bool ValidPixel(SymSeam * topVerticalSeam, SymSeam * leftHorizontalSeam, 
							   int leftBndX, int topBndY, int imgWidth, int imgHeight, int x, int y,
							   int overlapWidth, int overlapHeight);
		static void FillMinOverlapCandidates(QImage & srcImg, int patchWidth, int patchHeight,
											 int overlapWidth, int overlapHeight, int leftX, int leftY,
											 int topX, int topY, int rightX, int rightY,
											 int bottomX, int bottomY, std::vector<int> & candidates, 
											 int numPatches, int numBestPatches);
		static double CompareRegions(QImage & srcImg, int regionWidth, int regionHeight, 
									 int x1, int y1, int x2, int y2);
//		int TopMultiSeam(int x, int y, SymSeam * leftSeam, SymSeam * rightSeam, SymSeam * topSeam);
//		int LeftMultiSeam(int x, int y, SymSeam * topSeam, SymSeam * bottomSeam, SymSeam * rightSeam);
		int GetRegionID(int x, int y, SymSeam * leftSeam, SymSeam * rightSeam, SymSeam * topSeam, SymSeam * bottomSeam);
		
		int m_iPatchNumX;
		int m_iPatchNumY;
		std::vector<int> m_iPatchPermutations;
	
	protected:
		void ComputeAllSeams();
		void RecomputeSeamsForPatch(int patchID);
		void ComputeSeamBetweenPatches(int patchID1, int patchID2);

		std::vector<SymSeam> m_HorizontalSeams;
		std::vector<SymSeam> m_VerticalSeams;
	
		
		std::vector<int> m_PickedCandidatesPatches;
		QImage m_TempTesterCheckIfAllPatches;
		int m_iCurrentPatch;
	
		TextureSynthAlgorithm m_SynthAlgorithm;
		QImage m_Source;
		QImage m_Deformed;
		SymMatrix2DReal * m_pSourceMatrix;
		int m_iPatchWidth;
		int m_iPatchHeight;
		int m_iOverlapWidth;
		int m_iOverlapHeight;
		int m_iOptimNumPatches;
		int m_iOptimNumBestPatches;
		std::vector<DynProgCell> m_OverlapFieldVertical;
		std::vector<DynProgCell> m_OverlapFieldHorizontal;
		bool m_bMinCutCarving;
};

#endif

