#include "DeformationLensDistortion.h"
#include "DeformImageWarpLinearMesh.h"
#include "DrawingRoutines.h"
#include "SymSpace.h"
#include "SymEnergy.h"
#include <iostream>
#include <time.h>
#include "CVLib.h"

DeformationLensDistortion::DeformationLensDistortion(QImage & img, int maxWidth, int maxHeight, CameraModel model, int flags)
: ImageWarp(img)
{
	int numParams = 0;
	std::cout<<"flags="<<flags<<std::endl;
	
	for (int i=0; i<TSAI_MAX; i++)
	{
		if ((flags&1)!=0)
		{
			m_FlagToParam.push_back(numParams);
			numParams++;
		}
		else
		{
			m_FlagToParam.push_back(-1);
		}
		flags /= 2;		
	}
	

	int numPnts[2];
	numPnts[0] = img.width()/10;
	numPnts[1] = img.height()/10;	
	ImagePenalization noPenalty;
	noPenalty.m_iFields = 0;
	m_pMesh = new ImageWarpLinearMesh(img, maxWidth, maxHeight, numPnts, false, false, 0, 0, 
									  noPenalty, NULL, ImageWarpLinearMesh::FACE_TRIANGLE3);
	m_pOriginalParams = new SymVectorReal(m_pMesh->GetParameterVector().m);

	m_fScaleX = img.width();
	m_fScaleY = img.height();
	
	for (int i=0; i<m_pMesh->GetParameterVector().m; i+=2)
	{
		(*m_pOriginalParams)(i+0) = m_pMesh->GetParameterVector()(i+0)/m_fScaleX;
		(*m_pOriginalParams)(i+1) = m_pMesh->GetParameterVector()(i+1)/m_fScaleY;		
	}
	m_pAllMinParamStep = new SymVectorReal(numParams, 0.01);
	m_pAllParamStep = new SymVectorReal(numParams, 1);
	m_pAllLastParams = new SymVectorReal(numParams);		
	double pi = 3.14159265;
	if (m_FlagToParam[0]>=0)	//kx
	{
		(*m_pAllLastParams)(m_FlagToParam[0]) = 0;
		(*m_pAllParamStep)(m_FlagToParam[0]) = 0.1;
	}
	if (m_FlagToParam[1]>=0)	//ky
	{
		(*m_pAllLastParams)(m_FlagToParam[1]) = 0;
		(*m_pAllParamStep)(m_FlagToParam[1]) = 0.1;
	}
	if (m_FlagToParam[2]>=0)	//cx
	{
		(*m_pAllLastParams)(m_FlagToParam[2]) = 0.5;
		(*m_pAllParamStep)(m_FlagToParam[2]) = 0.05;		
	}
	if (m_FlagToParam[3]>=0)	//cy
	{
		(*m_pAllLastParams)(m_FlagToParam[3]) = 0.5;
		(*m_pAllParamStep)(m_FlagToParam[3]) = 0.05;				
	}
	if (m_FlagToParam[4]>=0)	//Rx
	{
		(*m_pAllLastParams)(m_FlagToParam[4]) = 0;
		(*m_pAllParamStep)(m_FlagToParam[4]) = pi/30;				
	}
	if (m_FlagToParam[5]>=0)	//Ry
	{
		(*m_pAllLastParams)(m_FlagToParam[5]) = 0;
		(*m_pAllParamStep)(m_FlagToParam[5]) = pi/30;				
	}
	if (m_FlagToParam[6]>=0)	//Rz
	{
		(*m_pAllLastParams)(m_FlagToParam[6]) = 0;
		(*m_pAllParamStep)(m_FlagToParam[6]) = pi/30;				
	}
	if (m_FlagToParam[7]>=0)	//Tx
	{
		(*m_pAllLastParams)(m_FlagToParam[7]) = 0;
		(*m_pAllParamStep)(m_FlagToParam[7]) = 0.1;
	}
	if (m_FlagToParam[8]>=0)	//Ty
	{
		(*m_pAllLastParams)(m_FlagToParam[8]) = 0;
		(*m_pAllParamStep)(m_FlagToParam[8]) = 0.1;
	}
	if (m_FlagToParam[9]>=0)	//Tz
	{
		(*m_pAllLastParams)(m_FlagToParam[9]) = -1;
		(*m_pAllParamStep)(m_FlagToParam[9]) = 0.1;
	}
	if (m_FlagToParam[10]>=0)	//f
	{
		(*m_pAllLastParams)(m_FlagToParam[10]) = 1;
		(*m_pAllParamStep)(m_FlagToParam[10]) = 0.1;
	}
	
	m_pInitParams = m_pAllLastParams;
	m_pParameterSteps = m_pAllParamStep;
	m_pMinimalParamSteps = m_pAllMinParamStep;	
	m_Model = model;
	m_iHierarchyLevel = -1;
}

SymMatrix & DeformationLensDistortion::Deform(SymVectorReal & params, double * outPenalty)
{
	// for each vertex convert to new params
	double f = 1;
	double kx = 0;
	double ky = 0;	
	double cx = 0.5;
	double cy = 0.5;

	double Rx = 0;
	double Ry = 0;
	double Rz = 0;
	double Tx = 0;
	double Ty = 0;
	double Tz = -1;
	
	if (m_FlagToParam[0]>=0)	//kx
		kx = (*m_pAllLastParams)(m_FlagToParam[0]);
	if (m_FlagToParam[1]>=0)	//ky
		ky = (*m_pAllLastParams)(m_FlagToParam[1]);
	else if (m_FlagToParam[0]>=0)
		ky = kx;
	if (m_FlagToParam[2]>=0)	//cx
		cx = (*m_pAllLastParams)(m_FlagToParam[2]);
	if (m_FlagToParam[3]>=0)	//cy
		cy = (*m_pAllLastParams)(m_FlagToParam[3]);
	if (m_FlagToParam[4]>=0)	//Rx
		Rx = (*m_pAllLastParams)(m_FlagToParam[4]);
	if (m_FlagToParam[5]>=0)	//Ry
		Ry = (*m_pAllLastParams)(m_FlagToParam[5]);
	if (m_FlagToParam[6]>=0)	//Rz
		Rz = (*m_pAllLastParams)(m_FlagToParam[6]);
	if (m_FlagToParam[7]>=0)	//Tx
		Tx = (*m_pAllLastParams)(m_FlagToParam[7]);
	if (m_FlagToParam[8]>=0)	//Ty
		Ty = (*m_pAllLastParams)(m_FlagToParam[8]);
	if (m_FlagToParam[9]>=0)	//Tz
		Tz = (*m_pAllLastParams)(m_FlagToParam[9]);
	if (m_FlagToParam[10]>=0)	//f
		f = (*m_pAllLastParams)(m_FlagToParam[10]);
	
//	std::cout<<"k = ["<<kx<<", "<<ky<<"]"<<std::endl;
	
	double minX, maxX, minY, maxY, temp;
	double minX2, maxX2, minY2, maxY2;	
	GetUnprojectedValue( 0,    0.5, 1, &minX, &temp, kx, ky, cx, cy, f, 1, 1, Rx, Ry, Rz, Tx, Ty, Tz);
	GetUnprojectedValue( 1,    0.5, 1, &maxX, &temp, kx, ky, cx, cy, f, 1, 1, Rx, Ry, Rz, Tx, Ty, Tz);
	GetUnprojectedValue( 0.5,  0,   1, &temp, &minY, kx, ky, cx, cy, f, 1, 1, Rx, Ry, Rz, Tx, Ty, Tz);
	GetUnprojectedValue( 0.5,  1,   1, &temp, &maxY, kx, ky, cx, cy, f, 1, 1, Rx, Ry, Rz, Tx, Ty, Tz);

	GetUnprojectedValue( 0, 0, 1, &minX2, &minY2, kx, ky, cx, cy, f, 1, 1, Rx, Ry, Rz, Tx, Ty, Tz);
	GetUnprojectedValue( 1, 1, 1, &maxX2, &maxY2, kx, ky, cx, cy, f, 1, 1, Rx, Ry, Rz, Tx, Ty, Tz);
	
	minX = qMax(minX, minX2);		minY = qMax(minY, minY2);
	maxX = qMin(maxX, maxX2);		maxY = qMin(maxY, maxY2);		
	
	double kScaleX = maxX - minX;
	double kScaleY = maxY - minY;	
//	kScaleX = 1;
//	kScaleY = 1;
//	std::cout<<"kSCale=["<<kScaleX<<", "<<kScaleY<<"], k="<<k<<" x=["<<minX<<", "<<maxX<<"] y=["<<minY<<", "<<maxY<<"]"<<std::endl;
	assert(kScaleX<1000000000 && kScaleX>-100000000);

	for (int i=0; i<(int)m_pOriginalParams->m; i+=2)
	{
		double x = (*m_pOriginalParams)(i+0);
		double y = (*m_pOriginalParams)(i+1);		
		double z = 1;
		double newX, newY;		
		GetUnprojectedValue(x, y, z, &newX,  &newY, kx, ky, cx, cy, f, kScaleX, kScaleY, Rx, Ry, Rz, Tx, Ty, Tz);
		m_pMesh->GetParameterVector()(i+0) = newX*m_fScaleX;
		m_pMesh->GetParameterVector()(i+1) = newY*m_fScaleY;
	}
	
	SymMatrix & retMatrix = m_pMesh->DeformAtCurrentHierarchyLevel(m_pMesh->GetParameterVector(), outPenalty);
	//TODO: check kx, ky are valid for all pixels
	return retMatrix;
}

QImage DeformationLensDistortion::GetDeformedImage(int imageType)
{
	return m_pMesh->GetDeformedImage(imageType);
}

void DeformationLensDistortion::GetOutputDimensions(int * w, int * h)
{
	m_pMesh->GetOutputDimensions(w, h);
}

void DeformationLensDistortion::SetPostFilter(SymFilter * filter)
{
//	std::cout<<"Setting filter"<<std::endl;
	m_pMesh->SetPostFilter(filter);
}

void DeformationLensDistortion::GetUnprojectedValue(double x, double y, double z, 
													double * newX, double * newY, 
													double kx, double ky, double cx, double cy, 
													double f, double locSx, double locSy, 
													double Rx, double Ry, double Rz, 
													double Tx, double Ty, double Tz)
{
//	std::cout<<"\nxy=["<<x<<", "<<y<<"] k=["<<kx<<", "<<ky<<"] r2="<<kx*(x-cx)*(x-cx) + ky*(y-cy)*(y-cy)<<" c=["<<cx<<", "<<cy<<"]"<<std::endl;
	x -= cx; 
	y -= cy;
	double r2 = kx*x*x + ky*y*y;

	assert(1+r2>0);
	x = x / (1 + r2);
	y = y / (1 + r2);	
	x = z * x / f;
	y = z * y / f;
	// x = R * (x-T)
	*newX = x / locSx + cx;
	*newY = y / locSy + cy;	
}

//void DeformationLensDistortion::GetUnprojectedValue(double x, double y, double z, 
//													double * newX, double * newY, 
//													double kx, double ky, double f)
//{
//	double r2 = kx*x*x + ky*y*y;
//	assert(1+r2>0);	
//	*newX = x / ((1 + r2)*f/z);
//	*newY = y / ((1 + r2)*f/z);
////	std::cout<<x<<"/"<<(1 + k * r2)<<"*"<<f<<"/"<<z<<" k="<<k<<" r="<<r2 <<std::endl;
////	*newX = (*newX+cx);
////	*newY = (*newY+cy);
//}


void DeformationLensDistortion::TestLensDistortion()
{
	QImage img("../../../data/imageSymmetrization/RadialLens/21.png");
	img.save("InputImage.png");
	DeformationLensDistortion lensUnwarp(img, 250, 250, TSAI_SIMPLE, TSAI_kx);
	//DeformationLensDistortion lensUnwarp(img, 250, 250, TSAI_SIMPLE, TSAI_Rx);
	lensUnwarp.SetPostFilter(new SymFilterGradient(0.01, 3));
	lensUnwarp.DeformAtCurrentHierarchyLevel(lensUnwarp.GetParameterVector());
	lensUnwarp.GetDeformedImage(-1).save("InputImage_mesh.png");

	int w, h;
	lensUnwarp.GetOutputDimensions(&w, &h);
	SymSpaceTransToroidal2D symSpace(w, h, w*0.5, h*0.5);
	SymEnergySymmetrization symEnergy(&symSpace, "SquaredDistancenFromMean");

	double minK = 1.999;
	int order = 0;
	
	//1-param	
	double stepK = minK/8;	
	std::cout<<"plot([";
	for (double k=-minK; k<=minK; k+=minK/8)
	{
		order++;	
		std::cout<<order<<" ";	//k
	}
	std::cout<<"], [";
	order=0;
	for (double k=-minK; k<=minK; k+=minK/8)
	{
		order++;
		lensUnwarp.GetParameterVector()(0) = k;
		lensUnwarp.GetParameterVector()(1) = k;		
		QString suffix = "_k="+QString::number(order)+".png";
		SymMatrix2DReal & f1D = ((SymMatrix2DReal&) lensUnwarp.DeformAtCurrentHierarchyLevel(lensUnwarp.GetParameterVector()));
		f1D.ToImage().save("f_1D"+suffix);
		lensUnwarp.GetDeformedImage(-1).save("f_"+suffix);
		SymMatrix2DReal & S = (SymMatrix2DReal &)symSpace.CalculateSymmetryAndProcess(f1D);
		S.ToImage().save("S"+suffix);
		//std::cout<<"E["<<order<<"] = "<<symEnergy.GetEnergy(f1D)<<"\tk="<<k<<std::endl;
		std::cout<<symEnergy.GetEnergy(f1D)<<" "<<std::flush;
	}
	std::cout<<"]);"<<std::endl;

	
//	// 2-param
//	double stepK = minK/4;		
//	double minEnergy = SYM_INFINITY;
//	int bestOrder = -1;
//	double bestKx, bestKy;
//	std::cout<<"surf([";
//	for (double k=-minK; k<=minK; k+=stepK)
//		std::cout<<k<<" ";
//	std::cout<<"], [";
//	for (double k=-minK; k<=minK; k+=stepK)
//		std::cout<<k<<" ";	
//	std::cout<<"], ["<<std::flush;	
//	for (double kx=-minK; kx<=minK; kx+=stepK)
//	{
//		for (double ky=-minK; ky<=minK; ky+=stepK)
//		{
//			order++;
//			lensUnwarp.GetParameterVector()(0) = kx;
//			lensUnwarp.GetParameterVector()(1) = ky;		
//			QString suffix = "_k="+QString::number(order)+".png";
//			SymMatrix2DReal & f1D = ((SymMatrix2DReal&) lensUnwarp.DeformAtCurrentHierarchyLevel(lensUnwarp.GetParameterVector()));
//			f1D.ToImage().save("f_1D"+suffix);
//			lensUnwarp.GetDeformedImage(-1).save("f_"+suffix);
//			SymMatrix2DReal & S = (SymMatrix2DReal &)symSpace.CalculateSymmetryAndProcess(f1D);
//			S.ToImage().save("S"+suffix);
//			//std::cout<<"E["<<order<<"] = "<<symEnergy.GetEnergy(f1D)<<"\tk="<<k<<std::endl;
//			double energy = symEnergy.GetEnergy(f1D);
//			if (minEnergy>energy)
//			{
//				minEnergy = energy;
//				bestOrder = order;
//				bestKx = kx;
//				bestKy = ky;
//				f1D.ToImage().save("BEST_f_1D.png");
//				S.ToImage().save("BEST_S.png");
//				lensUnwarp.GetDeformedImage(-1).save("BEST_f.png");
//			}
//			std::cout<<energy<<" "<<std::flush;
//		}
//		if (kx+stepK<minK)
//			std::cout<<"; ";
//	}
//	std::cout<<"]);"<<std::endl;
//	std::cout<<"bestOrder="<<bestOrder<<" bestK=["<<bestKx<<", "<<bestKy<<"]"<<std::endl;
}

