#include "Deformations.h"

/**************** ExampleWithAnswer ********************/
ExampleWithAnswer::ExampleWithAnswer(QString name)
{
	m_pValidSolutionSpace = NULL;
	m_pPerfectSolution = NULL;
	m_sName = name;
}

double ExampleWithAnswer::GetParameterSpaceError(SymVectorReal * optimizerOutput)
{
	if (m_pPerfectSolution==NULL || optimizerOutput->m!=m_pPerfectSolution->m)
		return -1;
//	assert(m_pPerfectSolution!=NULL && m_pPerfectSolution->m == optimizerOutput->m);
	double totalErr = 0;
	std::vector<double> errors;
	for (int i=0; i<m_pPerfectSolution->m; i++)
	{
		double err = ((*m_pPerfectSolution)(i) - (*optimizerOutput)(i))*((*m_pPerfectSolution)(i) - (*optimizerOutput)(i));
		errors.push_back(err);
		totalErr += err;
	}
	//return sqrt(Err)/optimizerOutput->m;
	sort(errors.begin(), errors.end());
	//return errors[errors.size()/2];
	//std::cout<<"\n\t\tTotalError="<<totalErr<<" without outliers=";
	totalErr = 0;
	int errorsToInspect = (int)errors.size()*8 / 10;
	for (int i=0; i<errorsToInspect; i++)
		totalErr += errors[i];
	//std::cout<<totalErr<<" half="<<errors[errors.size()/2]<<" 3/4="<<errors[errors.size()*3/4]<<std::endl;
	return totalErr * 100 / errorsToInspect;
}

double ExampleWithAnswer::GetImageSpaceError(QImage & optimizerSolution)
{
	optimizerSolution.save("ComparingMe.png");
	m_ImagePerfectDeformed.save("ComparingTo.png");	
	assert(m_ImagePerfectDeformed.width()==optimizerSolution.width() 
		   && m_ImagePerfectDeformed.height()==optimizerSolution.height());
	double Err = 0;
//	int totalEnergyOptimal = 0;
//	int totalEnergyMy = 0;
	QImage diffImg(m_ImagePerfectDeformed);
//	for (int i=0; i<m_ImagePerfectDeformed.width(); i++)
//	{
//		for (int j=0; j<m_ImagePerfectDeformed.height(); j++)
//		{
//			double valBest = ((double)qGray(m_ImagePerfectDeformed.pixel(i,j)))/255.;
//			double valMy = ((double)qGray(optimizerSolution.pixel(i,j)))/255.;
//			totalEnergyOptimal+=valBest;
//			totalEnergyMy+=valMy;
//		}
//	}
	for (int i=0; i<m_ImagePerfectDeformed.width(); i++)
	{
		for (int j=0; j<m_ImagePerfectDeformed.height(); j++)
		{
			double valBest = ((double)qGray(m_ImagePerfectDeformed.pixel(i,j)))/255.;
			double valMy = ((double)qGray(optimizerSolution.pixel(i,j)))/255.;			
			//double diff = valBest - valMy*totalEnergyOptimal/totalEnergyMy;
			double diff = valBest-valMy;
			int c = (int)(diff*diff*255);
			assert(c>=0 && c<=255);
			
			diffImg.setPixel(i,j,qRgb(c,c,c));
			Err+= diff*diff;
		}
	}
	diffImg.save("imgSpaceDiff.png");
	
	return sqrt(Err) / (m_ImagePerfectDeformed.width() * m_ImagePerfectDeformed.height());
}

double ExampleWithAnswer::GetSymSpaceError(SymMatrix & optimizerSolution)
{
	optimizerSolution.Subtract(*m_pPerfectSymmetrySpace);
	return optimizerSolution.SumElements(2) / optimizerSolution.GetNumElements();
}

/**************** DeformationHierarchyNode ********************/
DeformationHierarchyNode::DeformationHierarchyNode()
{
	m_bAdded = false;
	m_fTempWeightValue = 0;
	m_iDepth = 0;
}

/**************** HierarchyLevelDescriptor ********************/
HierarchyLevelDescriptor::HierarchyLevelDescriptor(int depth, int firstElement, int maxElements, 
												   std::vector<DeformationHierarchyNode*> & breadthFirstSorted)
{
	std::vector<int> tempMaps;
	std::vector<int> tempStepMaps;
	m_bSeparateStepMapping = breadthFirstSorted[0]->m_LevelParamStepIDs.size()!=0;
	for (int i=0; i<maxElements; i++)
	{
		DeformationHierarchyNode * currNode = breadthFirstSorted[firstElement + i];
//		assert(m_bSeparateStepMapping || ((int)currNode->m_LevelParamStepIDs.size()>depth
//										  && currNode->m_LevelParamStepIDs[depth].size()==currNode->m_ParamIDs.size()));
		for (int paramID = 0; paramID < (int)currNode->m_ParamIDs.size(); paramID++)
			tempMaps.push_back(currNode->m_ParamIDs[paramID]);
		
		for (int paramID = 0; m_bSeparateStepMapping && paramID < (int)currNode->m_ParamIDs.size(); paramID++)
			tempStepMaps.push_back(currNode->m_LevelParamStepIDs[depth][paramID]);
	}
	m_iDepth = depth;
	m_aiMapToConcatParams = new int[tempMaps.size()];
	m_pConcatSteps = new SymVectorReal(tempMaps.size());
	m_pConcatParams  = new SymVectorReal(tempMaps.size());;
	m_pConcatMinSteps = new SymVectorReal(tempMaps.size());	
	if (m_bSeparateStepMapping)
		m_aiMapToConcatParamsSteps = new int[tempMaps.size()];
	else
		m_aiMapToConcatParamsSteps = NULL;
	
	for (int i=0; i<(int)tempMaps.size(); i++)
	{
		m_aiMapToConcatParams[i] = tempMaps[i];	
		if (m_bSeparateStepMapping)
			m_aiMapToConcatParamsSteps[i] = tempStepMaps[i];
	}
}

void HierarchyLevelDescriptor::CopyFromConcatParams(SymVectorReal * trgParams)
{
	for (int i=0; i<m_pConcatParams->m; i++)
		(*trgParams)(m_aiMapToConcatParams[i]) = (*m_pConcatParams)(i);	
}

void HierarchyLevelDescriptor::CopyFromConcatSteps(SymVectorReal * trgParams)
{
	int * paramMap = m_bSeparateStepMapping ? m_aiMapToConcatParamsSteps : m_aiMapToConcatParams;
	for (int i=0; i<m_pConcatSteps->m; i++)
		(*trgParams)(paramMap[i]) = (*m_pConcatSteps)(i);	
}	

void HierarchyLevelDescriptor::CopyFromConcatMinSteps(SymVectorReal * trgParams)
{
	int * paramMap = m_bSeparateStepMapping ? m_aiMapToConcatParamsSteps : m_aiMapToConcatParams;	
	for (int i=0; i<m_pConcatMinSteps->m; i++)
		(*trgParams)(paramMap[i]) = (*m_pConcatMinSteps)(i);	
}

void HierarchyLevelDescriptor::CopyToConcatParams(SymVectorReal * srcParams)
{
	for (int i=0; i<m_pConcatParams->m; i++)
		(*m_pConcatParams)(i) = (*srcParams)(m_aiMapToConcatParams[i]);	
}

void HierarchyLevelDescriptor::CopyToConcatSteps(SymVectorReal * srcParams)
{
	int * paramMap = m_bSeparateStepMapping ? m_aiMapToConcatParamsSteps : m_aiMapToConcatParams;	
	for (int i=0; i<m_pConcatSteps->m; i++)
		(*m_pConcatSteps)(i) = (*srcParams)(paramMap[i]);
}

void HierarchyLevelDescriptor::CopyToConcatMinSteps(SymVectorReal * srcParams)
{
	int * paramMap = m_bSeparateStepMapping ? m_aiMapToConcatParamsSteps : m_aiMapToConcatParams;	
	for (int i=0; i<m_pConcatMinSteps->m; i++)
		(*m_pConcatMinSteps)(i) = (*srcParams)(paramMap[i]);		
}

/**************** DeformationModel::HierarchyPrintNode ********************/
DeformationModel::HierarchyPrintNode::HierarchyPrintNode(QString name, std::vector<int> & paramIDs, int depth)
{
	m_sName = name;
	m_iParamIDs = paramIDs;
	m_iDepth = depth;
}


/**************** DeformationModel ********************/
DeformationModel * DeformationModel::CloneSelfUsingBestParams(SymVectorReal &)
{
	assert(false);
	return NULL;
}

void DeformationModel::PrintParams(bool printMin)
{
	std::cout<<"Params("<<m_pAllParamStep->m<<") :";
	int initDepth = -1;
	for (int i=(int)m_ParamPrintNodes.size()-1; i>=0; i--)
	{
		if (initDepth != m_ParamPrintNodes[i].m_iDepth)
		{
			QString mod = " ";
			if (m_iHierarchyLevel==m_iMaxHierarchyLevels-m_ParamPrintNodes[i].m_iDepth-1 && m_iHierarchyLevel!=-1)
				mod="*";
			std::cout<<"\n\tDepth["<<m_iMaxHierarchyLevels-m_ParamPrintNodes[i].m_iDepth-1<<"]"<<mod.toStdString().c_str()<<" : "<<std::flush;
			initDepth = m_ParamPrintNodes[i].m_iDepth;
		}
		std::cout<<m_ParamPrintNodes[i].m_sName.toStdString().c_str()<<" Ids=[ "<<std::flush;
		for (int id=0; id<(int)m_ParamPrintNodes[i].m_iParamIDs.size(); id++)
			std::cout<<m_ParamPrintNodes[i].m_iParamIDs[id]<<" ";
		std::cout<<"] \t Values=[";
		for (int id=0; id<(int)m_ParamPrintNodes[i].m_iParamIDs.size(); id++)
			std::cout<<(*m_pAllParamStep)(m_ParamPrintNodes[i].m_iParamIDs[id])<<" ";
		std::cout<<"]";
	}
	std::cout<<std::endl;
	
	if (printMin)
	{
		std::cout<<"MinParams("<<m_pAllParamStep->m<<") :";
		initDepth = -1;
		for (int i=(int) m_ParamPrintNodes.size()-1; i>=0; i--)
		{
			if (initDepth != m_ParamPrintNodes[i].m_iDepth)
			{
				QString mod = " ";
				if (m_iHierarchyLevel==m_iMaxHierarchyLevels-m_ParamPrintNodes[i].m_iDepth-1 && m_iHierarchyLevel!=-1)
					mod="*";				
				std::cout<<"\n\tDepth["<<m_iMaxHierarchyLevels-m_ParamPrintNodes[i].m_iDepth-1<<"]"<<mod.toStdString().c_str()<<" : "<<std::flush;
				initDepth = m_ParamPrintNodes[i].m_iDepth;
			}
			
			std::cout<<m_ParamPrintNodes[i].m_sName.toStdString().c_str()<<"=[ "<<std::flush;
			for (int id=0; id<(int)m_ParamPrintNodes[i].m_iParamIDs.size(); id++)
			{
				double paramValue = (*m_pAllMinParamStep)(m_ParamPrintNodes[i].m_iParamIDs[id]);
				std::cout<<paramValue<<" ";
			}
			std::cout<<"]\t"<<std::flush;
		}
		std::cout<<std::endl;
		
	}
}
		 

DeformationModel::DeformationModel()
{
	m_bInvertColors = false;
	m_bPolarCoords = false;
	m_iLastChangedParam = -1;
	m_pPostFilter = NULL;
	m_pInitParams = NULL;
	m_pParameterSteps = NULL;
	m_pMinimalParamSteps = NULL;
	
	m_pAllLastParams = NULL;	//in child dimensions
	m_pAllParamStep = NULL;
	m_pAllMinParamStep = NULL;
	
	m_pFinalParams = NULL;
	
	m_pDeformedMatrix = NULL;

	m_iHierarchyLevel = -1;
	m_iMaxHierarchyLevels = 0;
}

DeformationModel::~DeformationModel()
{
}

void DeformationModel::InvertColors()
{
	m_bInvertColors = true;
}

SymVectorReal & DeformationModel::GetParameterVector()
{
	assert(m_pInitParams!=NULL);

	return *m_pInitParams;
}

SymVectorReal & DeformationModel::GetSuggestedParamStep()
{
	assert(m_pParameterSteps!=NULL);
	return (*m_pParameterSteps);
}

SymVectorReal & DeformationModel::GetMinimalParamStep()
{
	assert(m_pMinimalParamSteps!=NULL);
	return (*m_pMinimalParamSteps);
}

void DeformationModel::TestRandomizeParams(double minRand, double maxRand)
{
	double displacement = 0;
	double sumDisplacementsOdd = 0;
	double sumDisplacementsEven = 0;
	srand(time(NULL));
	
	if (m_pAllLastParams->GetEntryType()==SYM_ENTRY_REAL)
	{
		SymVectorReal & vec = (SymVectorReal &)*m_pAllLastParams;
		for (int i=0; i<vec.m; i++)
		{
			double rnd = (double)(rand() % 10000) / 10000.;
			displacement = rnd * minRand + (1-rnd) * maxRand;
			if (i%2==0)
				sumDisplacementsEven += displacement;
			else
				sumDisplacementsOdd += displacement;
			vec(i) += displacement;
		}
	}
	//DeformAtCurrentHierarchyLevel(*m_pAllLastParams);
	Deform(*m_pAllLastParams);
}

bool DeformationModel::HasOutput(OutputType)
{
	return false;
}

void DeformationModel::GetCenterOfMass(double * , double * )
{
	assert(false);
}

void DeformationModel::GetOutputDimensions(int * , int * )
{
	assert(false);
}

SymMatrix & DeformationModel::DeformAtCurrentHierarchyLevel(SymVectorReal & params, double * outPenalty)
{	
	if (m_iHierarchyLevel>=0)
	{
		assert(m_PrecomputedSingleLayers[m_iHierarchyLevel]!=NULL 
			   && (int)m_PrecomputedSingleLayers.size()>m_iHierarchyLevel);
		if (m_PrecomputedSingleLayers[m_iHierarchyLevel]->m_pConcatParams != &params)
			*(m_PrecomputedSingleLayers[m_iHierarchyLevel]->m_pConcatParams) = params;
		m_PrecomputedSingleLayers[m_iHierarchyLevel]->CopyFromConcatParams(m_pAllLastParams);
	}
	else
	{
		*m_pAllLastParams = params;
	}
	
	if (m_bPolarCoords)	
		ConvertPolar2Cart(*m_pAllLastParams, *m_pAllLastParams);
		
	SymMatrix * ret = NULL;
	if (m_pHierarchyBreadthFirstList.size()>0)
	{
		GenerateHierarchicalParameterVector(*m_pAllLastParams);
		ret = &(Deform(*m_pFinalParams, outPenalty));
	}
	else
	{
		ret = &(Deform(*m_pAllLastParams, outPenalty));
	}
	if (m_pPostFilter!=NULL)
	{
		m_pPostFilter->Process(ret);
	}
	
	if (m_bPolarCoords)	
		ConvertCart2Polar(*m_pAllLastParams, *m_pAllLastParams);
		
	return *ret;
}




SymMatrix & DeformationModel::TakeStepInDimension(int dimension, double dx, double & penalty)
{
	m_iLastChangedParam = dimension;
	m_fLastParamChangedValue = GetParameterVector()(dimension);
	
	GetParameterVector()(dimension) += dx;
	return DeformAtCurrentHierarchyLevel(GetParameterVector(), &penalty);
}

void DeformationModel::RevertLastStepInDimension()
{
	GetParameterVector()(m_iLastChangedParam) = m_fLastParamChangedValue;	
}


QImage DeformationModel::GetDeformedImage(int )
{
	assert(false);	//	should never be reached
	return QImage();
}

int DeformationModel::GetNumOutputImageTypes()
{
	return 0;
}

int DeformationModel::GetNumHierarchyLevels()
{
	return m_iMaxHierarchyLevels;
}

void DeformationModel::SetPostFilter(SymFilter * filter)
{
	m_pPostFilter = filter;
}

bool DeformationModel::SwitchToHierarchy(int hierarchy)
{
	assert(hierarchy>=-1 && hierarchy < m_iMaxHierarchyLevels);
	assert(m_pAllLastParams!=NULL && m_pAllParamStep!=NULL && m_pAllMinParamStep!=NULL);

	if (hierarchy==m_iHierarchyLevel)
		return true;

	if (m_PrecomputedSingleLayers.size()==0)
	{
		for (int i=0; i<m_iMaxHierarchyLevels; i++)
			m_PrecomputedSingleLayers.push_back(NULL);
	}
	
	if (hierarchy>=0 && m_PrecomputedSingleLayers[hierarchy]==NULL)
	{
		m_PrecomputedSingleLayers[hierarchy] = new HierarchyLevelDescriptor(hierarchy, 
			     m_iLayerToFirstElement[hierarchy], m_iLayerToMaxElements[hierarchy], m_pHierarchyBreadthFirstList);
	}
	
	ChildUpdateAtHierarchySwitch(m_iHierarchyLevel, hierarchy);		
	
	if (hierarchy==-1)
	{
		m_PrecomputedSingleLayers[m_iHierarchyLevel]->CopyFromConcatParams(m_pAllLastParams);
		m_PrecomputedSingleLayers[m_iHierarchyLevel]->CopyFromConcatSteps(m_pAllParamStep);
	}
	else if (hierarchy>=0 && m_iHierarchyLevel!=-1)
	{
		m_PrecomputedSingleLayers[m_iHierarchyLevel]->CopyFromConcatParams(m_pAllLastParams);
		m_PrecomputedSingleLayers[m_iHierarchyLevel]->CopyFromConcatSteps(m_pAllParamStep);		
	}
		
	m_iHierarchyLevel = hierarchy;
	if (m_iHierarchyLevel==-1)
	{		
		m_pInitParams = m_pAllLastParams;
		m_pParameterSteps = m_pAllParamStep;
		m_pMinimalParamSteps = m_pAllMinParamStep;
	}
	else 
	{
		m_PrecomputedSingleLayers[m_iHierarchyLevel]->CopyToConcatParams(m_pAllLastParams);
		m_PrecomputedSingleLayers[m_iHierarchyLevel]->CopyToConcatSteps(m_pAllParamStep);
		m_PrecomputedSingleLayers[m_iHierarchyLevel]->CopyToConcatMinSteps(m_pAllMinParamStep);		
		
		m_pInitParams = m_PrecomputedSingleLayers[m_iHierarchyLevel]->m_pConcatParams;
		m_pParameterSteps = m_PrecomputedSingleLayers[m_iHierarchyLevel]->m_pConcatSteps;
		m_pMinimalParamSteps = m_PrecomputedSingleLayers[m_iHierarchyLevel]->m_pConcatMinSteps;
	}
		
	return true;
}

/**
 * Implement in child. Must use m_pFinalParams to fill m_pAllLastParams
 * See note (in code) for further details.
 */
void DeformationModel::ChildUpdateAtHierarchySwitch(int, int)
{
	//NOTE: deformation model only copies m_pAllLastParams values, which basically
	//		correspond to control vertices (i.e. all dependent vertices are 
	//		re-assigned in m_pFinalParams). However, when jumping from one
	//		hierarchy level to another, some control points might have to be
	//		re-assigned (if they depended on control points at lower-level hierarchy).
}

void DeformationModel::ReduceSuggestedStepByFactor(double factor)
{
	if (m_iHierarchyLevel==-1)
	{
		for (int i=0; i<m_pAllParamStep->m; i++)
			(*m_pAllParamStep)(i)*=factor;
	}
	else
	{
		for (int i=0; i<m_pParameterSteps->m; i++)
		{
			(*m_pParameterSteps)(i)*=factor;
			int tempHierarchy = m_iHierarchyLevel;
			SwitchToHierarchy(-1);
			SwitchToHierarchy(tempHierarchy);
		}
	}
}

int DeformationModel::GetCurrentHierarchy()
{
	return m_iHierarchyLevel;
}

bool DeformationModel::GoUpHieararchy()		//higher-resolution (i.e. more local)
{
	if (m_iHierarchyLevel < m_iMaxHierarchyLevels-1)
		return SwitchToHierarchy(m_iHierarchyLevel+1);
	else
		return false;
}

bool DeformationModel::GoDownHierarchy()	//lower-resolution (i.e. more global)
{
	if (m_iHierarchyLevel > 0)
		return SwitchToHierarchy(m_iHierarchyLevel-1);
	else
		return false;
}


void DeformationModel::InitializeParameterHierearchy(std::vector<DeformationHierarchyNode *> &hierarchyTreeRoot, 
													 SymVectorReal & params, HierarchyTreeTraversal traversalType, 
													 DeduceWeightsMethod weightingMethod, DeduceParentsLocations pLocMethod)
{
	m_TraversalType = traversalType;
	
	if ((&m_pHierarchyTreeRoots)!=&(hierarchyTreeRoot))
		m_pHierarchyTreeRoots = hierarchyTreeRoot;
		
	std::cout<<"Breadth Fill Value from: InitializeParameterHierearchy"<<std::endl;
	if (m_TraversalType==HIERARCHY_TRAVERSE_BREADTH)
		BreadthFirstFillValues(params, weightingMethod, pLocMethod);
//	else if (m_TraversalType==HIERARCHY_TRAVERSE_DEPTH)
//	{
//		for (int i=0; i<(int)m_pHierarchyTreeRoots.size(); i++)
//			RecursivelyFillValue(m_pHierarchyTreeRoots[i], params);
//	}
	else if (m_TraversalType!=HIERARCHY_TRAVERSE_NONE)
	{
		std::cout<<"DeformationModel::InitializeParameterHierearchy unknown hierarchy traversal type!"<<std::endl;
		exit(0);
	}
}

//double DeformationModel::RecursivelyFillValue(DeformationHierarchyNode * root, SymVectorReal & params)
//{
//	//1. set parent's value to average of children
//	double & fillme = params(root->m_iParamID);
//	for (int i=0; i<(int)root->m_apChildren.size(); i++)	
//		fillme += RecursivelyFillValue(root->m_apChildren[i], params);
//
//	//2. set parent's value
//	if (root->m_apChildren.size()>0)
//		fillme = fillme/root->m_apChildren.size();	
//		
//	//3. set children offsets
//	for (int i=0; i<(int)root->m_apChildren.size(); i++)	
//	{
//		if (root->m_apChildren[i]->m_RelationshipToParent==DeformationHierarchyNode::HIERARCH_REL_ADD)
//			params(root->m_apChildren[i]->m_iParamID) -= fillme;
//		else
//		{
//			std::cout<<"DeformationModel::RecursivelyFillValue error: unknown child-parent relationship"<<std::endl;
//			exit(0);
//		}
//	}
//	return fillme;
//}

void DeformationModel::BreadthFirstFillValues(SymVectorReal & params, DeduceWeightsMethod weightingMethod, DeduceParentsLocations pLocMethod )
{
	//std::cout<<"Breadth-first, paramInit=";		params.PrintVector(0.0001);	std::cout<<std::endl;
	//1. figure out breadth-first order (store it for all future calculations)
	m_iMaxHierarchyLevels = -1;
	for (int i=0; i<(int)m_pHierarchyTreeRoots.size(); i++)
		m_pHierarchyBreadthFirstList.push_back(m_pHierarchyTreeRoots[i]);
	int visit = 0;
//	QMap<DeformationHierarchyNode *, int> visitedValue;
//	std::cout<<"Filling values for "<<m_pHierarchyBreadthFirstList.size()<<std::endl;
//	std::cout<<"ParentCompare=["<<std::flush;
	while (visit < (int)m_pHierarchyBreadthFirstList.size())
	{
		DeformationHierarchyNode * currNode = m_pHierarchyBreadthFirstList[visit++];
//		if (visitedValue.contains(currNode))
//			std::cout<<"\nRe-Visiting\n "<<currNode<<std::endl;
//		else
//			visitedValue[currNode] = 1;
		int currDepth = currNode->m_iDepth;
		assert(currDepth-m_iMaxHierarchyLevels<2 && currDepth>=m_iMaxHierarchyLevels);
		if (currDepth > m_iMaxHierarchyLevels)
		{
			m_iLayerToFirstElement.push_back(visit-1);
			if (m_iMaxHierarchyLevels>=0)
			{
				int lastEl = m_iLayerToFirstElement.size()-1;
//				std::cout<<"elements["<<m_iLayerToMaxElements.size()<<"] = "<<m_iLayerToFirstElement[lastEl] - m_iLayerToFirstElement[lastEl-1]<<std::endl;				
				m_iLayerToMaxElements.push_back(m_iLayerToFirstElement[lastEl] - m_iLayerToFirstElement[lastEl-1]);
			}
			m_iMaxHierarchyLevels = currDepth;
		}
//		std::cout<<"Method="<<weightingMethod<<" KeepAs="<<WEIGHTS_KEEP_AS_IT_IS<<std::endl;
		assert((weightingMethod!=WEIGHTS_KEEP_AS_IT_IS && currNode->m_afChildrenWeights.size()==0) 
			   || currNode->m_apChildren.size()==currNode->m_afChildrenWeights.size());
		for (int i=0; i<(int)currNode->m_apChildren.size(); i++)
		{
			if (weightingMethod!=WEIGHTS_KEEP_AS_IT_IS)
				currNode->m_afChildrenWeights.push_back(1);
			if (currNode->m_apChildren[i]->m_bAdded==false)
			{
				currNode->m_apChildren[i]->m_bAdded = true;
				currNode->m_apChildren[i]->m_iDepth = currDepth+1;	//initialize node's depth
				m_pHierarchyBreadthFirstList.push_back(currNode->m_apChildren[i]);
			}
		}
//		std::cout<<currNode->m_apChildren.size()<<" == "<<currNode->m_afChildrenWeights.size()<<", "<<std::flush;
		assert(currNode->m_apChildren.size()==currNode->m_afChildrenWeights.size());
	}
	m_iMaxHierarchyLevels++;
	int lastEl = m_iLayerToFirstElement.size()-1;
//	std::cout<<"elements["<<m_iLayerToMaxElements.size()<<"] = "<<(m_pHierarchyBreadthFirstList.size() - m_iLayerToFirstElement[lastEl])<<std::endl;	
	m_iLayerToMaxElements.push_back(m_pHierarchyBreadthFirstList.size() - m_iLayerToFirstElement[lastEl]);	//num els at last hierarchy

	//2a. fill all parent's values to average of its children (weighted according to provided weights)
	double maxParentToChildDist = 0;
	int numDims = -1;
	for (int i=(int)m_pHierarchyBreadthFirstList.size()-1; i>=0; i--)
	{
		DeformationHierarchyNode * currNode = m_pHierarchyBreadthFirstList[i];
		if (currNode->m_apChildren.size() > 0)	//for each non-terminal node
		{
			std::vector<int> & currParamIDs = currNode->m_ParamIDs;

			double sumWeights = 0;	//for parent children's weights do not have to sum to 1, so find norm
			if (pLocMethod==PARENTSLOC_AVERAGE_CHILDREN)	//parent's location = mean(children)
			{
				for (int paramID=0; i<paramID<(int)currParamIDs.size(); paramID++)
					params(currParamIDs[paramID]) = 0;
				for (int c=0; c<(int)currNode->m_apChildren.size(); c++)
				{
					sumWeights += currNode->m_afChildrenWeights[c];
					std::vector<int> & childrenIDs = currNode->m_apChildren[c]->m_ParamIDs;
					for (int paramID=0; i<paramID<(int)currParamIDs.size(); paramID++)	// sum up all children
						params(currParamIDs[paramID]) += currNode->m_afChildrenWeights[c] * params(childrenIDs[paramID]);
				}
				for (int paramID=0; i<paramID<(int)currParamIDs.size(); paramID++)	// rescale parent's location 
					params(currParamIDs[paramID]) /= sumWeights;
			}
			for (int c=0; c<(int)currNode->m_apChildren.size(); c++)
			{
				double currParentToChildrenDist = 0;
				std::vector<int> & childIDs = currNode->m_apChildren[c]->m_ParamIDs;
				for (int paramID=0; paramID < numDims; paramID++)
				{
					double childVal =  params(childIDs[paramID]);
					double parentVal = params(currParamIDs[paramID]);
					currParentToChildrenDist+=(childVal-parentVal)*(childVal-parentVal);
				}
				maxParentToChildDist = qMax(sqrt(currParentToChildrenDist), maxParentToChildDist);				
			}
		}
	}
	
	//std::cout<<"Breadth-first, paramLocations=";		params.PrintVector();	std::cout<<std::endl;
	
	//2b. Fill parents list for each child (used for some weight-calculators).
	for (int i=0; i<(int)m_pHierarchyBreadthFirstList.size(); i++)	//find sums for children weights
	{
		DeformationHierarchyNode * currNode = m_pHierarchyBreadthFirstList[i];
		for (int c=0; c<(int)currNode->m_apChildren.size(); c++)
		{
			currNode->m_apChildren[c]->m_apParents.push_back(currNode);
			currNode->m_apChildren[c]->m_iIDinParents.push_back(c);
		}
	}
	
	//2c. calculate children weights (or sum them up for normalization)	
	maxParentToChildDist *= 2;
	if (maxParentToChildDist==0)
		maxParentToChildDist=1;	
	
	for (int i=0; i<(int)m_pHierarchyBreadthFirstList.size(); i++)
		m_pHierarchyBreadthFirstList[i]->m_fTempWeightValue = 0;
	
	for (int i=(int)m_pHierarchyBreadthFirstList.size()-1; i>=0; i--)	//find sums for children weights
	{
		DeformationHierarchyNode * currNode = m_pHierarchyBreadthFirstList[i];
		std::vector<int> & currParamIDs = currNode->m_ParamIDs;
		for (int c=0; c<(int)currNode->m_apChildren.size(); c++)
		{
			DeformationHierarchyNode * child = currNode->m_apChildren[c];
			std::vector<int>& childIDs = child->m_ParamIDs;
			if (weightingMethod==WEIGHTS_CONST)
			{
				currNode->m_afChildrenWeights[c] = 1;
			}
			else if (weightingMethod==WEIGHTS_DISTANCE_LINEAR)	//don't use
			{
				double currDist = 0;
				for (int paramID=0; paramID < numDims; paramID++)
				{
					double childVal =  params(childIDs[paramID]);
					double parentVal = params(currParamIDs[paramID]);
					currDist+=(childVal-parentVal)*(childVal-parentVal);
				}				
				currNode->m_afChildrenWeights[c] = maxParentToChildDist - sqrt(currDist);
			}
			else if (weightingMethod==WEIGHTS_BILINEAR)	//use only for 2D images. Not very efficient, but it is done once
			{
					//assert that each child has 4 parents for interpolation & they are in a proper grid!
//				if (child->m_apParents.size()!=4)
//					std::cout<<"parents="<<child->m_apParents.size()<<" of child ["<<(*(child->m_pParamIDs))(0)<<", "<<(*(child->m_pParamIDs))(1)<<"]"<<std::endl;
				assert(child->m_apParents.size()==4);
				double x[4];
				double y[4];
				for (int t=0; t<4; t++)
				{
					x[t] = params(child->m_apParents[t]->m_ParamIDs[0]);
					y[t] = params(child->m_apParents[t]->m_ParamIDs[1]);
				}

				DeformationHierarchyNode * p[4];
				
				p[0] = NULL;				p[1] = NULL;				p[2] = NULL;				p[3] = NULL;
				double x1 = x[0];
				double x2 = x[0];
				double y1 = y[0];
				double y2 = y[0];
				for (int t=1; t<4; t++)
				{
					x1 = qMin(x1, x[t]);		x2 = qMax(x2, x[t]);
					y1 = qMin(y1, y[t]);		y2 = qMax(y2, y[t]);					
				}
				
				double properX[4];
				double properY[4];
				properX[0] = x1;		properX[1] = x2;		properX[2] = x1;		properX[3] = x2;
				properY[0] = y1;		properY[1] = y1;		properY[2] = y2;		properY[3] = y2;
				
				double w[4];
				double childX = params(child->m_ParamIDs[0]);
				double childY = params(child->m_ParamIDs[1]);
//				bool displayData = (*(child->m_pParamIDs))(0)==8;
				double div = (x2-x1)*(y2-y1);
				assert(childX>=x1 && childX<=x2 && childY>=y1 && childY<=y2);
				
				//just sort parents by positions
//				std::cout<<"==="<<std::endl;
				for (int sp=0; sp<4; sp++)
				{
					for (int prop=0; prop<4; prop++)
					{
						if (x[sp]==properX[prop] && y[sp]==properY[prop])
						{
							p[prop] = child->m_apParents[sp];
//							if (displayData)
//								std::cout<<"p["<<sp<<"] = ["<<params((*(p[sp]->m_pParamIDs))(0))<<", "<<params((*(p[sp]->m_pParamIDs))(1))<<"]"<<std::endl;
						}
					}					
				}
				
				assert(p[0]!=NULL && p[1]!=NULL && p[2]!=NULL && p[3]!=NULL);
//				std::cout<<"======="<<std::endl;
//				std::cout<<"\tchildIDs=["<<(*(child->m_pParamIDs))(0)<<", "<<(*(child->m_pParamIDs))(1)<<"]"<<std::endl;
//				std::cout<<"\tchild=["<<childX<<", "<<childY<<"] box=["<<x1<<", "<<x2<<"] to ["<<y1<<", "<<y2<<"]"<<std::endl;
//				for (int i=0; i<4; i++)
//					std::cout<<"\tp["<<i<<"]=id["<<(*(p[i]->m_pParamIDs))(0)<<", "<<(*(p[i]->m_pParamIDs))(1)<<"] l["<<params((*(p[i]->m_pParamIDs))(0))<<", "<<params((*(p[i]->m_pParamIDs))(0))<<"]"<<std::endl;
				w[3] = (childX-x1)*(childY-y1)/div;
				w[2] = (x2-childX)*(childY-y1)/div;
				w[1] = (childX-x1)*(y2-childY)/div;
				w[0] = (x2-childX)*(y2-childY)/div;				
				
				for (int wid = 0; wid<4; wid++)
				{
					if (currNode==p[wid])
					{
//						if (displayData)
//						{
//							std::cout<<"\tweight["<<wid<<"] of parent["<<params((*(p[wid]->m_pParamIDs))(0))<<", "<<params((*(p[wid]->m_pParamIDs))(1))<<"] = "<<w[wid]<<std::endl;
//							std::cout<<"\ttarget=["<<childX<<", "<<childY<<"] "<<std::endl;
//						}
						currNode->m_afChildrenWeights[c] = w[wid];
					}
				}
				
			}
			else if (weightingMethod!=WEIGHTS_KEEP_AS_IT_IS)
			{
				std::cout<<"DeformationModel::BreadthFirstFillValues - ERROR: can't resolve children "; 
				std::cout<<"weight assignment at this level. Assign weights at subclass. MethodID="<<weightingMethod<<std::endl;
				exit(0);
			}
			currNode->m_apChildren[c]->m_fTempWeightValue += currNode->m_afChildrenWeights[c];
		}
	}

	//2c. normalize children's weights 
	for (int i=0; i<(int)m_pHierarchyBreadthFirstList.size(); i++)	//rescale weights
	{
		DeformationHierarchyNode * currNode = m_pHierarchyBreadthFirstList[i];
		for (int c=0; c<(int)currNode->m_apChildren.size(); c++)
			currNode->m_afChildrenWeights[c] /= currNode->m_apChildren[c]->m_fTempWeightValue;				
	}		
		 
		
	//3. For each child set offset
	for (int i=m_pHierarchyBreadthFirstList.size()-1; i>=0; i--)
	{
		DeformationHierarchyNode * parent = m_pHierarchyBreadthFirstList[i];
		std::vector<int> & parentIds = parent->m_ParamIDs;
		for (int c=0; c<(int)parent->m_apChildren.size(); c++)
		{
			double childWeight = parent->m_afChildrenWeights[c];
			//int childID = parent->m_apChildren[c]->m_iParamID;
			std::vector<int> & childIds = parent->m_apChildren[c]->m_ParamIDs;
			for (int id=0; id<(int)parentIds.size(); id++)
			{
				if (parent->m_apChildren[c]->m_RelationshipToParent==DeformationHierarchyNode::HIERARCH_REL_ADD)
					params(childIds[id]) -= params(parentIds[id]);
				else if (parent->m_apChildren[c]->m_RelationshipToParent==DeformationHierarchyNode::HIERARCH_REL_OFFSET_FROM_AVERAGE)
					params(childIds[id]) -= params(parentIds[id]) * childWeight;
			}
		}
	}
	//std::cout<<"Breadth-first, paramFinal=";		params.PrintVector(0.0001);	std::cout<<std::endl;	
}

void DeformationModel::GenerateHierarchicalParameterVectorChildImplement(SymVectorReal & , SymVectorReal &)
{
	assert(false);	//must be implemented in the inherited class
}

void DeformationModel::GenerateHierarchicalParameterVector(SymVectorReal & params)
{
	assert(m_pFinalParams!=NULL && m_pFinalParams->m==params.m);	
	
	*m_pFinalParams = params;
	
	if (m_TraversalType==HIERARCHY_TRAVERSE_BREADTH)
	{
		BreadthFirstHierarchicalParameterVector();
	}
	else if (m_TraversalType==HIERARCHY_TRAVERSE_NONE)
	{
		GenerateHierarchicalParameterVectorChildImplement(params, *m_pFinalParams);
	}
//	else if (m_TraversalType==HIERARCHY_TRAVERSE_DEPTH)
//	{
//		for (int i=0; i<(int)m_pHierarchyTreeRoots.size(); i++)
//			DepthFirstHierarchicalParameterVector(m_pHierarchyTreeRoots[i]);
//	}
}

//void DeformationModel::DepthFirstHierarchicalParameterVector(DeformationHierarchyNode * root)
//{	
//	std::vector<int> visitStack;
//	std::vector<DeformationHierarchyNode *> visitStackPtrs;
//	visitStack.push_back(0);	
//	visitStackPtrs.push_back(root);
//	while (visitStack.size()>0)
//	{
//		int & visitID = visitStack[visitStack.size()-1];
//		DeformationHierarchyNode * currParent = visitStackPtrs[visitStackPtrs.size()-1];
//
//		if (visitID >= (int)currParent->m_apChildren.size())
//		{
//			visitStack.pop_back();
//			visitStackPtrs.pop_back();
//		}
//		else
//		{		
//			DeformationHierarchyNode * currChild = currParent->m_apChildren[visitID];
//			int parentID = currParent->m_iParamID;
//			int childID = currChild->m_iParamID;	
//			
//			if (currChild->m_RelationshipToParent==DeformationHierarchyNode::HIERARCH_REL_ADD)
//				(*m_pFinalParams)(childID) += (*m_pFinalParams)(parentID);
//			else if (currChild->m_RelationshipToParent==DeformationHierarchyNode::HIERARCH_REL_OFFSET_FROM_AVERAGE)
//				(*m_pFinalParams)(childID) += (*m_pFinalParams)(parentID) * currParent->m_afChildrenWeights[visitID];
//			
//			visitID++;
//			
//			visitStackPtrs.push_back(currChild);
//			visitStack.push_back(0);
//		}
//	}
//}

void DeformationModel::BreadthFirstHierarchicalParameterVector()
{
	for (int i=0; i<(int)m_pHierarchyBreadthFirstList.size(); i++)
	{
		DeformationHierarchyNode * parent = m_pHierarchyBreadthFirstList[i];
		std::vector<int> & parentIDs = parent->m_ParamIDs;
		for (int c=0; c<(int)parent->m_apChildren.size(); c++)
		{
			double childWeight = parent->m_afChildrenWeights[c];
			std::vector<int> & childIDs = parent->m_apChildren[c]->m_ParamIDs;
			
			if (parent->m_apChildren[c]->m_RelationshipToParent==DeformationHierarchyNode::HIERARCH_REL_ADD)
			{
				for (int currID=0; currID < (int)parentIDs.size(); currID++)
					(*m_pFinalParams)(childIDs[currID]) += (*m_pFinalParams)(parentIDs[currID]);
			}
			else if (parent->m_apChildren[c]->m_RelationshipToParent==DeformationHierarchyNode::HIERARCH_REL_OFFSET_FROM_AVERAGE)
			{
				for (int currID=0; currID < (int)parentIDs.size(); currID++)
					(*m_pFinalParams)(childIDs[currID]) += (*m_pFinalParams)(parentIDs[currID]) * childWeight;
			}
		}
	}	
}

SymMatrix & DeformationModel::TakeRandomStep(bool & enforceStep)
{
	enforceStep = false;
	int randomParam = rand() % GetParameterVector().m;
	m_iLastChangedParam = randomParam;	
	assert(GetMinimalParamStep()(randomParam)>0);	
	if (rand()%2==0)
	{
		m_fLastParamChangedValue = GetMinimalParamStep()(randomParam);
		GetParameterVector()(randomParam) += GetMinimalParamStep()(randomParam);
	}
	else
	{
		m_fLastParamChangedValue = -GetMinimalParamStep()(randomParam);		
		GetParameterVector()(randomParam) -= GetMinimalParamStep()(randomParam);
	}
	
	return DeformAtCurrentHierarchyLevel(GetParameterVector());
}

void DeformationModel::RevertLastRandomStep()
{
	if (m_iLastChangedParam!=-1)
		GetParameterVector()(m_iLastChangedParam) -= m_fLastParamChangedValue;
	m_iLastChangedParam = -1;
}

bool DeformationModel::TerminalState()
{
	return false;
}

void DeformationModel::SetToPolarCoords(double centerX, double centerY)
{	
//	std::cout<<"SETTING TO POLAR:";
//	m_pInitParams->PrintVector();
//	std::cout<<"\n=============\n"<<std::endl;
	m_bPolarCoords = true;
	m_fPolarCoordsCX = centerX;
	m_fPolarCoordsCY = centerY;
	ConvertCart2Polar(*m_pAllLastParams, *m_pAllLastParams);
	ConvertCart2Polar(*m_pInitParams, *m_pInitParams);
	
	for (int i=1; i<m_pAllMinParamStep->m; i+=2)
	{
		(*m_pAllMinParamStep)(i) *= 2 * 3.14159265 / 360;
		(*m_pAllParamStep)(i) *=  2 * 3.14159265 / 360;
	}
	
	for (int i=1; i<m_pMinimalParamSteps->m; i+=2)
	{
		(*m_pMinimalParamSteps)(i) *=  2 * 3.14159265 / 360;
		(*m_pParameterSteps)(i) *=  2 * 3.14159265 / 360;
	}
	
//	std::cout<<"DONE SETTING TO POLAR:";
//	m_pInitParams->PrintVector();
//	std::cout<<"\n=============\n"<<std::endl;
	
}

void DeformationModel::ConvertCart2Polar(SymVectorReal & vecFrom, SymVectorReal & vecTo)
{
	assert(vecFrom.m==vecTo.m);
	for (int i=0; i<vecFrom.m/2; i++)
	{
		double x = vecFrom(2*i+0)-m_fPolarCoordsCX;
		double y = vecFrom(2*i+1)-m_fPolarCoordsCY;		
		double r = sqrt(x*x+y*y);
		double theta = atan2(y, x);
		vecTo(2*i+0) = r;
		vecTo(2*i+1) = theta;		
//		vecTo(i) = r;
//		vecTo(vecFrom.m/2+i) = theta;
	}
}

bool DeformationModel::DistortsInput()
{
	return false;
}

void DeformationModel::ConvertPolar2Cart(SymVectorReal & vecFrom, SymVectorReal & vecTo)
{
	assert(vecFrom.m==vecTo.m);	
	for (int i=0; i<vecFrom.m/2; i++)
	{
		double r = vecFrom(2*i+0);
		double theta = vecFrom(2*i+1);		
//		double r = vecFrom(i);
//		double theta = vecFrom(vecFrom.m/2+i);
		vecTo(2*i+0) = m_fPolarCoordsCX + r * cos(theta);
		vecTo(2*i+1) = m_fPolarCoordsCY + r * sin(theta);
	}	
	
}



