#include "Numerical.h"
#include "qt/qstringlist.h"
#include "qt/qprocess.h"

#ifndef __SYM_FILTER_SEQUENCE_H
#define __SYM_FILTER_SEQUENCE_H

class SymFilter
{
	public:
		SymFilter(QString specialDirections="none");
//		SymFilter(bool high);	
		virtual ~SymFilter(){}
		virtual void Process(SymMatrix * inOutMatrix);
		virtual void Process(SymMatrix * inMatrix, SymMatrix ** outMatrix);
		QString m_sSpecialDirections;
};

class SymFilterScript : public SymFilter
{
	public:
		SymFilterScript(QString excutable, QStringList params);
		virtual void Process(SymMatrix * inOutMatrix);
		virtual void Process(SymMatrix * inMatrix, SymMatrix ** outMatrix);
	
		QString m_sFilterType;
		QString m_sParameters;
};

class SymFilterGaussian : public SymFilter
{
	public:
		SymFilterGaussian(double stdDev);
		virtual ~SymFilterGaussian(){}
		virtual void Process(SymMatrix * inOutMatrix);
		static void Process2DReal(SymMatrix2DReal * inMatrix, double std_dev);
	
	protected:
		double m_fStdDev;
	
};

class SymFilterPntAbsPow : public SymFilter
{
	public:
		SymFilterPntAbsPow(double p, double maxVal=-1, double cutRad=-1);
		virtual ~SymFilterPntAbsPow(){}
		virtual void Process(SymMatrix * inOutMatrix);
		static void Process2DReal(SymMatrix2DReal * inMatrix, double p, double maxVal, double maxRad);
	
	protected:
		double m_fPow;
		double m_fMaxVal;
		double m_fCutRadius;
};

class SymFilterGradient : public SymFilter
{
	public:
		SymFilterGradient(double blurStdDev, int window);
		virtual ~SymFilterGradient(){}
		virtual void Process(SymMatrix * inOutMatrix);
		static void Process2DReal(SymMatrix2DReal * inMatrix, double blurStdDev, int window);
	
	protected:
		double m_fBlurStdDev;
		int m_iWindow;
};

class SymFilterCanny : public SymFilter
{
	public:
		SymFilterCanny(double afterBlurStdDev, double threshold1, double threshold2, int window);
		virtual ~SymFilterCanny(){}
		virtual void Process(SymMatrix * inOutMatrix);
		static void Process2DReal(SymMatrix2DReal * inMatrix, double afterBlurStdDev, double threshold1, double threshold2, int window);
		
	protected:
		double m_fThrehsold1;
		double m_fThrehsold2;
		double m_fAfterBlur;
		int m_iWindow;
};


/**
 * General 
 */
class SymFilterSequence : public SymFilter
{
	public:
		SymFilterSequence();
		virtual ~SymFilterSequence();
		virtual int GetNumFilters();
		virtual SymFilter * GetFilter(int fltID);	
		virtual void AddFilter(SymFilter * filter, int id=-1);
		virtual void Process(SymMatrix * inOutMatrix);
	
	protected:
		std::vector<SymFilter *> m_FilterSequence;
};

#endif



