#include "SymParams.h"
#include <Qt/qimage.h>
#include "DeformImageWarpLinearMesh.h"
#include "DeformTextureSynth.h"
#include "DeformationLensDistortion.h"
#include "SymSpaceReflective.h"
#include "SymSpaceAdditional.h"
#include <Qt/qprocess.h>
#include <Qt/qfile.h>
#include <Qt/qstring.h>
#include <Qt/qtextStream.h>

SymParamsText::SymParamsText(QString inFilename, char ** argv, int argc)
{
	QFile file(inFilename);
	std::cout<<"Opening: "<<inFilename.toStdString().c_str()<<std::endl;
	if (!file.open(QIODevice::ReadOnly | QIODevice::Text))
		assert(argc>0);	//all params must be here
	
	QTextStream textStream(&file);
	
	ParamModuleID currParam = MAX_PARAMS_INVALID;
		//read file, fill values
	QStringList commands = textStream.readAll().replace("\n", " ").replace("\t", " ").split(" ", QString::SkipEmptyParts);	
	for (int i=0; i<argc; i++)	//read arguments, add or refill arguments	
		commands.push_back(argv[i]);
	for (int i=0; i<commands.count(); i++)
	{
		//std::cout<<"command["<<i<<"] = "<<commands[i].toStdString().c_str()<<std::endl;
		if (commands[i]=="Optimizer:" || commands[i]=="Optimizer" || commands[i]=="-Optimizer")
			currParam = OPTIM_PARAMS;
		else if (commands[i]=="SymmetrySpace:" || commands[i]=="SymmetrySpace" || commands[i]=="-SymmetrySpace")
			currParam = SYM_SPACE_PARAMS;
		else if (commands[i]=="EnergyFn:" || commands[i]=="EnergyFn" || commands[i]=="-EnergyFn")
			currParam = ENERGY_PARAMS;
		else if (commands[i]=="DeformationModel:" || commands[i]=="DeformationModel" || commands[i]=="-DeformationModel")
			currParam = DEFORM_PARAMS;
		else if (commands[i]=="FilterTarget:" || commands[i]=="FilterTarget" || commands[i]=="-FilterTarget")
			currParam = FILTER_TRG_PARAMS;
		else if (commands[i]=="FilterDeform:" || commands[i]=="FilterDeform" || commands[i]=="-FilterDeform")
			currParam = FILTER_DEFORM_PARAMS;
		else if (currParam!=MAX_PARAMS_INVALID)
		{
			QString field = commands[i++];
			m_ParamModule[currParam][field].clear();
			assert(commands.count()>i);			
			if (commands[i].startsWith("[") && !commands[i].endsWith("]"))				//loop over multiple param values
			{
				assert(commands.count()>i);
				if (commands[i]=="[")
					i++;
				while(commands[i]!="]")
				{
					bool terminate = commands[i].endsWith("]");
					m_ParamModule[currParam][field].push_back(commands[i].replace("[", "").replace("]", ""));
					if (terminate)	break;
					i++;
				}
			}
			else		//set single param value
			{
				m_ParamModule[currParam][field].push_back(commands[i].replace("[", "").replace("]", ""));
			}
		}
	}
	std::cout<<"done"<<std::endl;
}

QStringList SymParamsText::GetStrValues(ParamModuleID module, QString valName, bool & valid)
{
	valid = m_ParamModule[module].contains(valName);
	if (valid)
		return m_ParamModule[module][valName];
	else
		return QStringList();
}

QString SymParamsText::GetStrValue(ParamModuleID module, QString valName, bool & valid)
{
	valid = m_ParamModule[module].contains(valName);
	if (valid)
		return m_ParamModule[module][valName][0];
	else 
		return QString();
}

double SymParamsText::GetDoubleValue(ParamModuleID module, QString valName, bool & valid)
{
	valid = m_ParamModule[module].contains(valName);
	if (valid)
		return m_ParamModule[module][valName][0].toDouble(&valid);
	else
		return -1.;
}

int SymParamsText::GetIntValue(ParamModuleID module, QString valName, bool & valid)
{
	valid = m_ParamModule[module].contains(valName);
	if (valid)
		return m_ParamModule[module][valName][0].toInt(&valid);	
	else
		return -1;
}

std::vector<double> SymParamsText::GetDoubleValues(ParamModuleID module, QString valName, bool & valid)
{
	valid = m_ParamModule[module].contains(valName);
	std::vector<double> newVals;
	if (valid)
	{
		QStringList list = m_ParamModule[module][valName];

		bool currValid;
		for (int i=0; i<list.count(); i++)
		{
			newVals.push_back(list[i].toDouble(&currValid));
			valid = currValid && valid; 
		}
	}
	return newVals;
}

std::vector<int> SymParamsText::GetIntValues(ParamModuleID module, QString valName, bool & valid)
{
	valid = m_ParamModule[module].contains(valName);
	std::vector<int> newVals;
	
	if (valid)
	{
		QStringList list = m_ParamModule[module][valName];

		bool currValid;
		for (int i=0; i<list.count(); i++)
		{
			newVals.push_back(list[i].toInt(&currValid));
			valid = currValid && valid; 
		}
	}
	return newVals;	
}

QImage SymPipelineParams::LoadImage(QString srcFile, QString defaultSrcDir, ExampleWithAnswer ** syntheticExample)
{
	*syntheticExample = NULL;
	
	//parse necessary args	
	if (srcFile.startsWith("-"))
		srcFile = QString(defaultSrcDir) + srcFile.right(srcFile.length()-1);
	else if (srcFile.startsWith("~"))
	{
		std::cout<<std::endl;
		srcFile = srcFile.right(srcFile.length()-1);
		QString srcFile_copy = srcFile;
		srcFile_copy.remove(srcFile_copy.length()-4, 4);
		QStringList synthImgList = srcFile_copy.split('_');
//		std::cout<<"list="<<synthImgList.count()<<std::endl;
//		for (int i=0; i<synthImgList.count(); i++)
//			std::cout<<"list["<<i<<"]="<<synthImgList[i].toStdString().c_str()<<std::endl;
		
		int rndSeed = 0;
		assert(synthImgList.count()>=1);
		if (synthImgList[synthImgList.count()-1] > 0)	//seed is given
			rndSeed = synthImgList[synthImgList.count()-1].toInt();
		else		//seed is given
			rndSeed = time(NULL);
		
		if (synthImgList[0] == "regGlyph")
		{
			*syntheticExample = new ImageWarpTestExample(srcFile);
			ImageWarpTestExample & newExample = (ImageWarpTestExample&)(**syntheticExample);
			
			// lattice
			int LATTICE_OFFSET = 1;
			assert(synthImgList.count()>LATTICE_OFFSET+14);
			newExample.m_BackgroundColor = synthImgList[LATTICE_OFFSET+1].toDouble();
			newExample.m_NumX = synthImgList[LATTICE_OFFSET+3].toInt();
			newExample.m_NumY = synthImgList[LATTICE_OFFSET+4].toInt();
			newExample.m_Blur = synthImgList[LATTICE_OFFSET+6].toDouble();
			newExample.m_V1 = R2Vector(synthImgList[LATTICE_OFFSET+8].toDouble(), synthImgList[LATTICE_OFFSET+9].toDouble());
			newExample.m_V2 = R2Vector(synthImgList[LATTICE_OFFSET+11].toDouble(), synthImgList[LATTICE_OFFSET+12].toDouble());	
			newExample.m_OddShift = synthImgList[LATTICE_OFFSET+14].toDouble();
			
			// glyphs
			int GLYPHS_OFFSET = LATTICE_OFFSET + 15;
			assert(synthImgList.count()>GLYPHS_OFFSET+8);			
			newExample.m_GlyphName = synthImgList[GLYPHS_OFFSET+1];
			newExample.m_GlyphX = synthImgList[GLYPHS_OFFSET+3].toDouble();
			newExample.m_GlyphY = synthImgList[GLYPHS_OFFSET+4].toDouble();
			newExample.m_GlyphMeanColor = synthImgList[GLYPHS_OFFSET+6].toDouble();
			newExample.m_bPutCtrlAtGlyphCorners = (synthImgList[GLYPHS_OFFSET+8] == "true");
			
			// noise
			int NOISE_OFFSET = GLYPHS_OFFSET + 9;
			assert(synthImgList.count()>NOISE_OFFSET+17);						
			newExample.m_FractionMissing = synthImgList[NOISE_OFFSET+1].toDouble();
			newExample.m_FractionExtra = synthImgList[NOISE_OFFSET+3].toDouble();
			newExample.m_ExtraBlobs = synthImgList[NOISE_OFFSET+5].toInt();
			newExample.m_ExtraBlobsFractionArea = synthImgList[NOISE_OFFSET+7].toDouble();
			newExample.m_GlyphLocationVariationHighFreq = synthImgList[NOISE_OFFSET+9].toDouble();
			newExample.m_GlyphLocationVariationLowFreq = synthImgList[NOISE_OFFSET+11].toDouble();
			newExample.m_GlyphDimensionsVariation = synthImgList[NOISE_OFFSET+13].toDouble();
			newExample.m_GlyphColorVariations = synthImgList[NOISE_OFFSET+15].toDouble();
			newExample.m_GlyphQuadVariations = synthImgList[NOISE_OFFSET+17].toDouble();
			
			newExample.GenerateExample(rndSeed);
		}
		else if (synthImgList[0] == "checkerboard")
		{			
			*syntheticExample = new ImageWarpTestExample(srcFile);
			ImageWarpExampleCheckerboard & newExample = (ImageWarpExampleCheckerboard&)(**syntheticExample);
			newExample.m_NumX = synthImgList[2].toInt();
			newExample.m_NumY = synthImgList[3].toInt();
			newExample.m_WidthX = synthImgList[5].toInt();
			newExample.m_WidthY = synthImgList[6].toInt();
			newExample.m_fNoise = synthImgList[8].toDouble();	
			newExample.GenerateExample(rndSeed);
		}
		else
		{
			assert(false);
		}
		std::cout<<"SynthInput="<<(**syntheticExample).m_ImageInputExample.width()<<" x "<<(**syntheticExample).m_ImageInputExample.height()<<std::endl;
		(**syntheticExample).m_ImageInputExample.save("SyntheticInput.png");
		(**syntheticExample).m_ImagePerfectOutput.save("SyntheticPerfect.png");		
		(**syntheticExample).m_ImagePerfectDeformed.save("SyntheticPerfectDeform.png");
	}
	
	QImage img;
	if (*syntheticExample==NULL)
		img = QImage(srcFile);
	else
		img = (*syntheticExample)->m_ImageInputExample;
	
	if (img.width()==0 || img.height()==0)
	{
		std::cout<<"ERROR: Image does not exist or invalid: "<<srcFile.toStdString().c_str()<<std::endl;
		exit(0);
	}	
	return img;
}

void SymPipelineParams::InitializeToDefault()
{
	m_pSymSpaceTargetFilter = NULL;
	m_pDeformPostFilter = NULL;
	
	m_pSg = NULL;
	m_bVerbous = true;
	m_pDeformationModel = NULL;
	m_pSymmetrySpace = NULL;
	m_pEnergyFn = NULL;
	m_Algorithm = SymOptimizer::F_SIMPLEX_NELDER_MEAD;
	m_iMaxNumIterations = 2000;
	m_iUpdate = 200;
	m_fRescaleAtUpdate = 1;
	numCtrl[0] = 10;
	numCtrl[1] = 10;
	invertMapping = false;
	m_fStep = 1;
	m_fToLineMin = 0.001;
	m_fStepVector = NULL;
	m_fMinChange = 0.001;
//	m_FilterTrgParams.m_sName = "none";
}

void SymPipelineParams::InitializeSymmetrySpace(int width, int height, int zDim, 
												SymParamsText & params, int minWidth, int minHeight,
												DeformationModel * deformationModel)
{
	bool valid;
	QString symSpaceType = params.GetStrValue(SymParamsText::SYM_SPACE_PARAMS, "Type", valid);	
	assert(valid);	// symmetry space has to be defined!
	double p = params.GetDoubleValue(SymParamsText::SYM_SPACE_PARAMS, "Pow", valid);
	if (!valid)		p = 1;
	QString norm = params.GetStrValue(SymParamsText::SYM_SPACE_PARAMS, "Normalization", valid);
	SymSpace::NormalizationType normType = SymSpace::NORM_SYMMETRY_SPACE;
	if (valid)
	{
		if (norm=="SymSpace")
			normType = SymSpace::NORM_SYMMETRY_SPACE;
		else if (norm=="ImgSpace")
			normType = SymSpace::NORM_IMAGE_SPACE;
		else if (norm=="GaussMean0Std1")
			normType = SymSpace::NORM_GAUSS_MEAN0_STD1;
		else if (norm=="SymSpaceMax1")
			normType = SymSpace::NORM_SYMMETRY_SPACE_MAX_TO_1;
		else if (norm=="none")
			normType = SymSpace::NORM_NONE;
		
	}	
	double constScale = 1;	
	//data for some (translational) symmetries
	std::vector<double> cropDims = params.GetDoubleValues(SymParamsText::SYM_SPACE_PARAMS, "CropSymDims", valid);
	int symDimsX = -1;
	int symDimsY = -1;
	if (valid)
	{
		if (symSpaceType=="Trans2D" || symSpaceType=="PointwisePadded")
		{
			minWidth*=2;
			minHeight*=2;
		}
		
		assert(cropDims[0]==-1 || cropDims[0]>0);
		if (cropDims[0]>1)
			symDimsX = cropDims[0];
		else
			symDimsX = cropDims[0] * minWidth;
		assert(cropDims[1]==-1 || cropDims[1]>0);		
		if (cropDims[1]>1)
			symDimsY = cropDims[1];		
		else
			symDimsY = cropDims[1] * minHeight;
	}
	
	bool cropToDisk = params.GetStrValue(SymParamsText::SYM_SPACE_PARAMS, "CropToDisk", valid)=="true" && valid;

	QStringList centerOfMassFlags = params.GetStrValues(SymParamsText::SYM_SPACE_PARAMS, "CenterOfMass", valid);
	int COMflags = 0;
	for (int i=0; i<centerOfMassFlags.count(); i++)		
	{
		if (centerOfMassFlags[i]=="none")
			COMflags = 0;
		else if (centerOfMassFlags[i]=="areaMass")
		{
			COMflags/=4;	COMflags*=4;	//erase first 2 bits
		}
		else if (centerOfMassFlags[i]=="intensityMass")
		{
			COMflags/=4;	COMflags*=4;
			COMflags |= SymSpaceSinglePointReflection2D::COM_INTENSITY_MASS;
		}
		else if (centerOfMassFlags[i]=="symmetryMass")
		{
			COMflags/=4;	COMflags*=4;
			COMflags |= SymSpaceSinglePointReflection2D::COM_SYMMETRY_MASS;
		}	
		else if (centerOfMassFlags[i]=="update")
		{
			COMflags |= SymSpaceSinglePointReflection2D::COM_UPDATE;
		}			
	}
	
	int iterateS = params.GetIntValue(SymParamsText::SYM_SPACE_PARAMS, "IterateS", valid);
	if (!valid)
		iterateS = 1;
	
	std::vector<int> color1DSyms = params.GetIntValues(SymParamsText::SYM_SPACE_PARAMS, "VisColors", valid);
	std::vector<double> cropVlaues = params.GetDoubleValues(SymParamsText::SYM_SPACE_PARAMS, "VisCrop", valid);
	double attenuateGamma = params.GetDoubleValue(SymParamsText::SYM_SPACE_PARAMS, "VisGamma", valid);
	if (!valid)
		attenuateGamma = -1;
	
	//data for some (reflectional) symmetries
	int numAngles = params.GetIntValue(SymParamsText::SYM_SPACE_PARAMS, "NumAngles", valid);
	if (!valid)
		numAngles = 144;
	int numRadii = params.GetIntValue(SymParamsText::SYM_SPACE_PARAMS, "NumRadii", valid);
	if (!valid)
		numRadii = -1;	
	if (symSpaceType=="TransToroidal2D")
	{
//		std::cout<<"\tSymDims="<<symDimsX<<" x "<<symDimsY<<""<<std::endl;		
		m_pSymmetrySpace = new SymSpaceTransToroidal2D(width, height, symDimsX, symDimsY, normType, p, 
													   constScale, iterateS);
	}	
	else if (symSpaceType=="Trans2D")
	{
		m_pSymmetrySpace = new SymSpaceTrans2D(width, height, symDimsX, symDimsY, normType, p, 
											   constScale);
	}
	else if (symSpaceType=="PRST")
	{
//		new SymSpacePRST2D(width, height, numAngles, numRadii, 
//						   SymSpacePRST2D::PRST_CALC_SAMPLE_ORIENTATIONS, 
//						   p, normType, constScale, iterateS, cropToDisk);		
//		SymSpacePRST2D(int img_m, int img_n, int numAngles, int numR=-1, 
//					   PRSTCalculationMethod method=PRST_CALC_SAMPLE_ORIENTATIONS, 
//					   double p = 1, NormalizationType norm = SymSpace::NORM_IMAGE_SPACE, 
//					   double constNormScalar=1, bool cropToDisk=true);
		
		m_pSymmetrySpace = new SymSpacePRST2D(width, height, numAngles, numRadii, 
											  SymSpacePRST2D::PRST_CALC_SAMPLE_ORIENTATIONS, 
											  p, normType, constScale, cropToDisk);
	}	
	else if (symSpaceType=="SinglePointReflective")
	{
		assert(deformationModel!=NULL);
		double cx = width/2., cy=height/2.;
		deformationModel->GetCenterOfMass(&cx, &cy);
		std::cout<<"cropToDisk="<<cropToDisk<<std::endl;
		std::cout<<"Flags="<<COMflags<<std::endl;
		m_pSymmetrySpace = new SymSpaceSinglePointReflection2D(width, height, numAngles, numRadii, 
															   deformationModel, p, normType, 
															   constScale, cropToDisk, COMflags);
	}
	else if (symSpaceType=="SinglePointRotational")
	{
		std::vector<double> rotationAngleLim = params.GetDoubleValues(SymParamsText::SYM_SPACE_PARAMS, "IgnoreRot", valid);
		if (!valid)
		{
			rotationAngleLim.push_back(0);			rotationAngleLim.push_back(0);			
		}
		assert(deformationModel!=NULL);		
		m_pSymmetrySpace = new SymSpaceRotationalSinglePnt2D(width, height, numAngles, numRadii, 
															 deformationModel, normType, cropToDisk, 
															 COMflags, rotationAngleLim[0], rotationAngleLim[1]);	
	}
	else if (symSpaceType=="PointwiseToroidal")
	{
		m_pSymmetrySpace = new SymSpacePointwiseToroidal(width, height, symDimsX, symDimsY, normType);	
	}
	else if (symSpaceType=="PointwisePadded")
	{
		m_pSymmetrySpace = new SymSpacePointwisePadded(width, height, symDimsX, symDimsY, normType);	
	}
	else if (symSpaceType=="Identity")
	{
		QString identityType = params.GetStrValue(SymParamsText::SYM_SPACE_PARAMS, "IdentityType", valid);
		if (!valid)
			identityType = "none";
		m_pSymmetrySpace = new SymSpaceIdentity(width, height, normType);
	}
	else
	{
		std::cout<<"SymPipelineParams::InitializeSymmetrySpace: Unknown symSpace: "<<symSpaceType.toStdString().c_str();
		std::cout<<" with dims="<<width<<" x "<<height<<" x "<<zDim<<std::endl;
		assert(false);
	}
	m_pSymmetrySpace->SetVisualizationParams(color1DSyms, cropVlaues, attenuateGamma);
}

void SymPipelineParams::InitializeImageDeformation(QImage & f, SymParamsText & params)
{
	std::cout<<"SymPipelineParams::InitializeImageDeformation - 1"<<std::endl;
	bool valid;
	QString deformType = params.GetStrValue(SymParamsText::DEFORM_PARAMS, "Type", valid);	
	assert(valid);	// symmetry space has to be defined!
	double pStep = params.GetDoubleValue(SymParamsText::DEFORM_PARAMS, "ParameterStep", valid);
	if (!valid)		pStep = 1;
	std::cout<<"SymPipelineParams::InitializeImageDeformation - 2"<<std::endl;
	if (deformType=="TextureSynth")
	{
		std::vector<int> resolution = params.GetIntValues(SymParamsText::DEFORM_PARAMS, "TargetResolution", valid);
		assert(valid && resolution.size()==2);
		std::vector<int> patchDims = params.GetIntValues(SymParamsText::DEFORM_PARAMS, "PatchDims", valid);
		assert(valid && patchDims.size()==2);
		std::vector<int> overlapDims = params.GetIntValues(SymParamsText::DEFORM_PARAMS, "OverlapDims", valid);
		assert(valid && overlapDims.size()==2);
		int optimNumPatches = params.GetIntValue(SymParamsText::DEFORM_PARAMS, "OptimNumPatches", valid);
		assert(valid);
		int optimNumBestPatches = params.GetIntValue(SymParamsText::DEFORM_PARAMS, "OptimNumBestPatches", valid);
		assert(valid);
		bool minCutCarving = params.GetStrValue(SymParamsText::DEFORM_PARAMS, "MinCutCarving", valid)!="false" || !valid;
		m_pDeformationModel	= new DeformTextureSynth(f, resolution[0], resolution[1], patchDims[0], patchDims[1], 
													 overlapDims[0], overlapDims[1], optimNumPatches, 
													 optimNumBestPatches, minCutCarving);
		
		std::cout<<"Texture Synthesis Target created"<<std::endl;
	}
	else if (deformType=="ImageWarpLinearMesh")
	{
		std::cout<<"SymPipelineParams::ImageWarpLinearMesh - 1"<<std::endl;
		std::vector<double> ctrlPnts = params.GetDoubleValues(SymParamsText::DEFORM_PARAMS, "NumCtrl", valid);
		if (valid)
		{
			if (ctrlPnts.size()==2)
			{
				numCtrl[0] = ctrlPnts[0];			
				numCtrl[1] = ctrlPnts[1];
			}
			else if (ctrlPnts.size()==1)
			{
				numCtrl[0] = ctrlPnts[0];			
				numCtrl[1] = 0;
			}
			else
			{
				assert(false);
			}
		}
		
		std::cout<<"SymPipelineParams::ImageWarpLinearMesh - 2"<<std::endl;
		double defaultStep  = params.GetDoubleValue(SymParamsText::DEFORM_PARAMS, "ParameterStep", valid);
		if (valid)
			m_fStep = defaultStep;

		QString invert = params.GetStrValue(SymParamsText::DEFORM_PARAMS, "Invert", valid);
		if (valid)
			invertMapping = (invert=="true");	
		
		QStringList checkFoldover = params.GetStrValues(SymParamsText::DEFORM_PARAMS, "Penalize", valid);
		ImageWarp::ImagePenalization penalization;
			std::cout<<"SymPipelineParams::ImageWarpLinearMesh - 3"<<std::endl;
		if (valid)
		{
			for (int i=0; i<(int)checkFoldover.count(); i++)
			{
				if (checkFoldover[i]=="none")
				{
					assert(checkFoldover.count()==1);
					penalization.m_iFields = 0;
				} 
				else if (checkFoldover[i]=="binaryFoldover")
				{
					penalization.m_iFields |= ImageWarp::ImagePenalization::FOLDOVER_BINARY;
				}
				else if (checkFoldover[i]=="smoothFoldover")
				{
					penalization.m_iFields |= ImageWarp::ImagePenalization::FOLDOVER_SMOOTH;
				}
				else if (checkFoldover[i]=="stayOnCanvas")
				{
					penalization.m_iFields |= ImageWarp::ImagePenalization::STAY_ON_CANVAS;
				}
			}
		}
		std::cout<<"SymPipelineParams::ImageWarpLinearMesh - 4"<<std::endl;
		ImageWarpLinearMesh::FaceType faceType = ImageWarpLinearMesh::FACE_QUAD;
		QString meshing = params.GetStrValue(SymParamsText::DEFORM_PARAMS, "Meshing", valid);
		if (valid)
			meshing=="triangles1";
		
		if (meshing=="triangles1")
			faceType = ImageWarpLinearMesh::FACE_TRIANGLE1;
		else if (meshing=="triangles2")
			faceType = ImageWarpLinearMesh::FACE_TRIANGLE2;
		else if (meshing=="triangles3")
			faceType = ImageWarpLinearMesh::FACE_TRIANGLE3;
		else if (meshing=="quads")
			faceType = ImageWarpLinearMesh::FACE_QUAD;
		else
			std::cout<<"Meshing type: "<<meshing.toStdString().c_str()<<" not implemented!"<<std::endl;
		
			std::cout<<"SymPipelineParams::ImageWarpLinearMesh - 5"<<std::endl;
		ImageWarp::ImageHierarchyDescriptor * hierarchyDescr = NULL;
				
		QString hierarchyType = params.GetStrValue(SymParamsText::DEFORM_PARAMS, "HierarchyType", valid);
		if (valid && hierarchyType!="none")
		{
			hierarchyDescr = new ImageWarp::ImageHierarchyDescriptor();
			if (hierarchyType=="ConstChildren" || hierarchyType=="OverlapChildren")
			{
				hierarchyDescr->m_HierarchyType = hierarchyType=="ConstChildren" ? ImageWarp::CONST_CHILDREN : ImageWarp::OVERLAP_CHILDREN;
				std::vector<int> assocXY = params.GetIntValues(SymParamsText::DEFORM_PARAMS, "HierarchyFactor", valid);
				if (valid)
				{
					hierarchyDescr->m_iAssocX = assocXY[0];
					hierarchyDescr->m_iAssocY = assocXY[1];				
				}
				
				std::vector<int> overlapXY = params.GetIntValues(SymParamsText::DEFORM_PARAMS, "HierarchyOverlap", valid);
				if (!valid)
				{
					overlapXY.clear();	
					overlapXY.push_back(hierarchyDescr->m_iAssocX/2);	
					overlapXY.push_back(hierarchyDescr->m_iAssocY/2);
				}
				hierarchyDescr->m_iOverlapX = overlapXY[0];
				hierarchyDescr->m_iOverlapY = overlapXY[1];			
				
			}
			else if (hierarchyType=="AsRigidAsPossible")
			{
				hierarchyDescr->m_HierarchyType = ImageWarp::AS_RIGID_AS_POSSIBLE;
			}
			
			int minPnts = params.GetIntValue(SymParamsText::DEFORM_PARAMS, "HierarchyMin", valid);
			if (valid)
				hierarchyDescr->m_iMinPnts = minPnts;

			int maxPnts = params.GetIntValue(SymParamsText::DEFORM_PARAMS, "HierarchyMax", valid);
			if (valid)
				hierarchyDescr->m_iMaxPnts = qMin(maxPnts, numCtrl[1]==0 ? numCtrl[0] : (numCtrl[0] * numCtrl[1]-1));
			else
				hierarchyDescr->m_iMaxPnts = numCtrl[1]==0 ? (numCtrl[0] / 2) : (numCtrl[0] * numCtrl[1] / 2);
			
			hierarchyDescr->m_iInitNumPntsX = numCtrl[0];
			hierarchyDescr->m_iInitNumPntsY = numCtrl[1];
			
		}
		double minValidStep = params.GetDoubleValue(SymParamsText::DEFORM_PARAMS, "MinValidStep", valid);
		if (!valid)
			minValidStep = 1;	//e.g. 1 pixel

		bool fixBnd = params.GetStrValue(SymParamsText::DEFORM_PARAMS, "FixBoundary", valid)=="true" && valid;
		
		std::vector<int> maxDims = params.GetIntValues(SymParamsText::DEFORM_PARAMS, "MaxDims", valid);
		if (!valid)
		{
			maxDims.push_back(f.width());
			maxDims.push_back(f.height());			
		}
		
		double windowZoom = params.GetDoubleValue(SymParamsText::DEFORM_PARAMS, "ZoomWindow", valid);
		if (!valid)
			windowZoom = 1;
		
		bool addMorePoints = params.GetStrValue(SymParamsText::DEFORM_PARAMS, "AddMorePoints", valid)=="true" && valid;
		
		std::cout<<"SymPipelineParams::ImageWarpLinearMesh - 6: "<<numCtrl[0]<<" x "<<numCtrl[1]<<std::endl;	
		if (numCtrl[1]==0)	// feature-based mesh
			m_pDeformationModel = new ImageWarpLinearMesh(f, maxDims[0], maxDims[1], numCtrl[0], 
														  invertMapping, fixBnd, m_fStep, minValidStep, 
														  penalization, hierarchyDescr, windowZoom, addMorePoints);
		else	// regular lattice
			m_pDeformationModel = new ImageWarpLinearMesh(f, maxDims[0], maxDims[1], numCtrl, 
														  invertMapping, fixBnd, m_fStep, minValidStep, 
														  penalization, hierarchyDescr, faceType, 
														  windowZoom);
		std::cout<<"SymPipelineParams::ImageWarpLinearMesh - 7"<<std::endl;
		QString invertColors = params.GetStrValue(SymParamsText::DEFORM_PARAMS, "InvertColors", valid);
		if (invertColors=="true" && valid)
			m_pDeformationModel->InvertColors();
		
		QString wireframe = params.GetStrValue(SymParamsText::DEFORM_PARAMS, "Wireframe", valid);
		if (valid && wireframe=="true")
			((ImageWarpLinearMesh*)m_pDeformationModel)->m_bShowWireframe = true;
		
		QString coordinateSystem  = params.GetStrValue(SymParamsText::DEFORM_PARAMS, "CoordinateSystem", valid);
		if (!valid)
			coordinateSystem = "Cartesian";
		
		double cx, cy;
		m_pDeformationModel->GetCenterOfMass(&cx, &cy);
		
		if (coordinateSystem=="Polar")
			m_pDeformationModel->SetToPolarCoords(cx, cy);
		
		std::cout<<"ImageWarpLinearMesh created and adjusted"<<std::endl;		
	}
	else if (deformType=="LensUnwarp")
	{
		std::vector<int> maxDims = params.GetIntValues(SymParamsText::DEFORM_PARAMS, "MaxDims", valid);
		if (!valid)
		{
			maxDims.push_back(f.width());
			maxDims.push_back(f.height());			
		}
		
		QString cameraModelStr = params.GetStrValue(SymParamsText::DEFORM_PARAMS, "CameraModel", valid);
		DeformationLensDistortion::CameraModel cameraModel = DeformationLensDistortion::TSAI_SIMPLE;
		int flags = DeformationLensDistortion::TSAI_kx;
		if (valid && cameraModelStr=="TSAI_SIMPLE")
		{
			cameraModel = DeformationLensDistortion::TSAI_SIMPLE;
			flags = DeformationLensDistortion::TSAI_kx;
			QStringList flagsStr = params.GetStrValues(SymParamsText::DEFORM_PARAMS, "Flags", valid);
			if (valid)
			{
				flags = 0;
				for (int i=0; i<flagsStr.count(); i++)
				{
					if (flagsStr[i]=="kx")
						flags |= DeformationLensDistortion::TSAI_kx;
					else if (flagsStr[i]=="ky")
						flags |= DeformationLensDistortion::TSAI_ky;
					else if (flagsStr[i]=="cx")
						flags |= DeformationLensDistortion::TSAI_cx;
					else if (flagsStr[i]=="cy")
						flags |= DeformationLensDistortion::TSAI_cy;
					else if (flagsStr[i]=="Rx")
						flags |= DeformationLensDistortion::TSAI_Rx;
					else if (flagsStr[i]=="Ry")
						flags |= DeformationLensDistortion::TSAI_Ry;
					else if (flagsStr[i]=="Rz")
						flags |= DeformationLensDistortion::TSAI_Rz;
					else if (flagsStr[i]=="Tx")
						flags |= DeformationLensDistortion::TSAI_Tx;
					else if (flagsStr[i]=="Ty")
						flags |= DeformationLensDistortion::TSAI_Ty;
					else if (flagsStr[i]=="Tz")
						flags |= DeformationLensDistortion::TSAI_Tz;
					else if (flagsStr[i]=="f")
						flags |= DeformationLensDistortion::TSAI_f;
				}
			}
		}
		
		m_pDeformationModel = new DeformationLensDistortion(f, maxDims[0], maxDims[1], cameraModel, flags);
	}
	else
	{
		assert(false);
	}
	
	std::cout<<"SymPipelineParams::InitializeImageDeformation - 3"<<std::endl;
	if (m_pDeformPostFilter!=NULL)
		m_pDeformationModel->SetPostFilter(m_pDeformPostFilter);
	
}

void SymPipelineParams::InitializeCurveDeformation(QString curveFile, SymParamsText & params)
{
	assert(false);	// vk: temp. pruned out on Oct 7, 2011
//	bool valid;
//	QString deformType = params.GetStrValue(SymParamsText::DEFORM_PARAMS, "Type", valid);	
//	assert(valid);	// symmetry space has to be defined!
//	double pStep = params.GetDoubleValue(SymParamsText::DEFORM_PARAMS, "ParameterStep", valid);
//	if (!valid)		pStep = 1;
//	
//	//std::cout<<"Type=="<<deformType.toStdString().c_str()<<std::endl;
//	if (deformType=="Curve2D")
//	{
//		double defaultStep  = params.GetDoubleValue(SymParamsText::DEFORM_PARAMS, "ParameterStep", valid);
//		if (valid)
//			m_fStep = defaultStep;
//		
//		double wndX = -1;
//		double wndY = -1;
//		std::vector<double> windowDims = params.GetDoubleValues(SymParamsText::DEFORM_PARAMS, "ScreenDims", valid);
//		if (valid)
//		{
//			wndX = windowDims[0];			
//			wndY = windowDims[1];
//		}
//		double stdDev = params.GetDoubleValue(SymParamsText::DEFORM_PARAMS, "StdDev", valid);
//		if (!valid)
//			stdDev=-1;
//		
//		DeformCurve::PointDistributionParams pntsParams;
//
//		double poissonDiskRad = params.GetDoubleValue(SymParamsText::DEFORM_PARAMS, "PoissonCondition", valid);
//		if (valid)
//			pntsParams.m_fPoissonDisk = poissonDiskRad;
//		
//		bool pntsSpeedup = params.GetStrValue(SymParamsText::DEFORM_PARAMS, "FastPnts", valid)=="true";
//		if (valid)
//			pntsParams.m_bPointDistrSpeedup = pntsSpeedup;
//
//		bool sumGaussians = params.GetStrValue(SymParamsText::DEFORM_PARAMS, "SumGaussians", valid)=="true";
//		if (valid)
//			pntsParams.m_bSumGaussians = sumGaussians;
//		
//		int samples = params.GetIntValue(SymParamsText::DEFORM_PARAMS, "NumSamples1", valid);
//		if (valid)
//			pntsParams.m_iPointDistrDimX = samples;
//		samples = params.GetIntValue(SymParamsText::DEFORM_PARAMS, "NumSamples2", valid);
//		if (valid)
//			pntsParams.m_iPointDistrDimY = samples;		
//
//		int polyPower = params.GetIntValue(SymParamsText::DEFORM_PARAMS, "PolynomPower", valid);
//		if (valid)
//			pntsParams.m_PointDistrPolyPower = polyPower;		
//
//		QString integrationType = params.GetStrValue(SymParamsText::DEFORM_PARAMS, "IntegrationType", valid);
//		if (valid)
//		{
//			if (integrationType=="CARTESIAN_COORDS")
//				pntsParams.m_IntegrationType = DeformCurve::CARTESIAN_COORDS;		
//			else if (integrationType=="POLAR_COORDS")
//				pntsParams.m_IntegrationType = DeformCurve::POLAR_COORDS;		
//			else if (integrationType=="POLAR_COORDS_WITH_POLYNOMIAL")
//				pntsParams.m_IntegrationType = DeformCurve::POLAR_COORDS_WITH_POLYNOMIAL;		
//			else 
//				assert(false);
//		}
//		
//		int numPntsToDistr = params.GetIntValue(SymParamsText::DEFORM_PARAMS, "DistributePnts", valid);
//		if (!valid)
//			numPntsToDistr = -1;
//		int numFarthestCandM = params.GetIntValue(SymParamsText::DEFORM_PARAMS, "NumFarthestCandidatesM", valid);
//		if (!valid)
//			numFarthestCandM = 20;
//		int numBestEnergy = params.GetIntValue(SymParamsText::DEFORM_PARAMS, "NumBestEnergy", valid);
//		if (!valid)
//			numBestEnergy = 10;
//		
//		if (curveFile.startsWith("~"))
//		{
//		//	bool fastPoint = params.GetStrValue(SymParamsText::DEFORM_PARAMS, "FastPnts", valid)=="true" && valid;
//			QString curveType;
//			double *paramsDouble = DivideCurveParams(curveFile, &curveType);
//			m_pDeformationModel = new DeformCurve(curveType, paramsDouble, pStep, wndX, wndY, stdDev, 
//												  true, pntsParams);
//		}		
//		else if (curveFile=="noFile" || numPntsToDistr>0)
//			m_pDeformationModel = new DeformCurve(numBestEnergy, numFarthestCandM, numPntsToDistr, 
//												  wndX, wndY, stdDev, true, pntsParams);
//		else
//			m_pDeformationModel = new DeformCurve(curveFile, pStep, wndX, wndY, stdDev, true, pntsParams);
//		m_pDeformationModel->SetPostFilter(m_pDeformPostFilter);
//	}	
//	else
//		assert(false);
}

double * SymPipelineParams::DivideCurveParams(QString curveFile, QString * curveType)
{
	curveFile = curveFile.right(curveFile.length()-1);
	QString curveFileCopy = curveFile;
	curveFileCopy = curveFileCopy.left(curveFile.length()-4);
	QStringList synthCurveParams = curveFileCopy.split("_", QString::SkipEmptyParts);
	double * paramsFloat = new double[synthCurveParams.count()-1];
//	std::cout<<"Params for "<<curveFile.toStdString().c_str()<<std::endl;
//	std::cout<<"\tName="<<synthCurveParams[0].toStdString().c_str()<<std::endl;
	for (int i=0; i<synthCurveParams.count()-1; i++)
	{
//		std::cout<<"Params["<<i+1<<"] = "<<synthCurveParams[i+1].toDouble()<<std::endl;
		paramsFloat[i] = synthCurveParams[i+1].toDouble();
	}
	*curveType = synthCurveParams[0];
	return paramsFloat;
}

void SymPipelineParams::InitializeEnergy(SymMatrix * Sg, SymParamsText & params, ExampleWithAnswer * example)
{
//	assert(maxWidth>0 && maxHeight>0);	//just to use it to get rid of warnings
	bool valid;
	QString energyType = params.GetStrValue(SymParamsText::ENERGY_PARAMS, "Type", valid);	
	assert(valid);	// symmetry space has to be defined!
	
	double p = params.GetDoubleValue(SymParamsText::ENERGY_PARAMS, "Pow", valid);
	if (!valid)		p = 1;
	
	if (energyType=="EnergyDiff")
	{
		assert(Sg!=NULL && Sg->GetNumDims()==2 && Sg->GetEntryType()==SYM_ENTRY_REAL);	//must have valid target
		m_pEnergyFn = new SymEnergyDifference(m_pSymmetrySpace, p, (SymMatrix2DReal&)(*Sg));
	}
	else if (energyType=="EnergySymmetrization")
	{
		QString symmetrizationType = params.GetStrValue(SymParamsText::ENERGY_PARAMS, "SymmetrizationType", valid);
		assert(valid);
		bool assymmetrization = params.GetStrValue(SymParamsText::ENERGY_PARAMS, "Asymmetrization", valid)=="true" && valid;
		m_pEnergyFn = new SymEnergySymmetrization(m_pSymmetrySpace, symmetrizationType, assymmetrization, p);
	}
	
	assert(m_pEnergyFn!=NULL);
	
	if (example!=NULL && example->m_ImagePerfectOutput.width()>0 && example->m_ImagePerfectOutput.height()>0
		 && example->m_ImagePerfectDeformed.width()>0 && example->m_ImagePerfectDeformed.height()>0)
	{
		SymMatrix2DReal tmpPerfect(example->m_ImagePerfectOutput);
		example->m_fEnergyPerfect = m_pEnergyFn->GetEnergy(tmpPerfect);
		SymMatrix2DReal tmpPerfectDeform(example->m_ImagePerfectDeformed);
		example->m_fEnergyDeform = m_pEnergyFn->GetEnergy(tmpPerfectDeform);		
		std::cout<<"\nPerfectSymEnergy="<<example->m_fEnergyPerfect<<"\n";
		std::cout<<"PerfectDeformSymEnergy="<<example->m_fEnergyDeform<<"\n";
	}

}

void SymPipelineParams::InitializeOptimization(SymParamsText & params, ExampleWithAnswer * example)
{
	assert(m_pEnergyFn!=NULL);
	assert(m_pDeformationModel!=NULL);
	bool valid;
	QString algorithm = params.GetStrValue(SymParamsText::OPTIM_PARAMS, "Type", valid);	
	if (valid)
	{
		std::cout<<"Algorithm="<<algorithm.toStdString().c_str()<<std::endl;
		if (algorithm=="CONJ_FLETCHER_REEVES")
			m_Algorithm = SymOptimizer::FDF_CONJ_FLETCHER_REEVES;
		else if (algorithm=="CONJ_POLAK_RIBIERE")
			m_Algorithm = SymOptimizer::FDF_CONJ_POLAK_RIBIERE;
		else if (algorithm=="BFGS")
			m_Algorithm = SymOptimizer::FDF_BFGS;
		else if (algorithm=="GRAD_DESC")
			m_Algorithm = SymOptimizer::FDF_GRAD_DESC;
		else if (algorithm=="SIMPLEX_NELDER_MEAD")
			m_Algorithm = SymOptimizer::F_SIMPLEX_NELDER_MEAD;
		else if (algorithm=="GREEDY_WALK")
			m_Algorithm = SymOptimizer::GREEDY_WALK;
		else if (algorithm=="MY_GRADIENT_DESCENT")
			m_Algorithm = SymOptimizer::MY_GRADIENT_DESCENT;
		else
			assert(false);
	}
	
	int numIters = params.GetIntValue(SymParamsText::OPTIM_PARAMS, "Iterations", valid);
	if (valid)
		m_iMaxNumIterations = numIters;

	int updateEvery = params.GetIntValue(SymParamsText::OPTIM_PARAMS, "Update", valid);
	if (valid)
		m_iUpdate = updateEvery;

	double minChange = params.GetDoubleValue(SymParamsText::OPTIM_PARAMS, "MinChange", valid);
	if (valid)
		m_fMinChange = minChange;

	QString verb = params.GetStrValue(SymParamsText::OPTIM_PARAMS, "OutputIntermediate", valid);
	if (valid)
		m_bVerbous = (verb=="true");
	
	double rescaleAtUpd = params.GetDoubleValue(SymParamsText::OPTIM_PARAMS, "RescaleAtUpdate", valid);
	if (valid)
		m_fRescaleAtUpdate = rescaleAtUpd;
	
	QString refeedStr = params.GetStrValue(SymParamsText::OPTIM_PARAMS, "UseIntermediateImageInDeform", valid);
	bool refeedImgAtUpdate = valid && (refeedStr=="true");
	
	int numDims = params.GetIntValue(SymParamsText::OPTIM_PARAMS, "NumDims", valid);

	if (!valid)
		numDims = -1;
	
	if (m_pDeformationModel->GetNumHierarchyLevels()>0)	//hierarchical optimizer
	{
		if (m_bVerbous)
			std::cout<<"Initializing Hierarchical Optimizer..."<<std::endl;
		m_pOptimizer = new MultiLevelOptimizer(m_pEnergyFn, m_pDeformationModel, m_Algorithm, 
											   m_iMaxNumIterations, m_iMaxNumIterations, m_iUpdate, 
											   m_fRescaleAtUpdate, refeedImgAtUpdate, example, numDims);
		

	}
	else if (m_Algorithm==SymOptimizer::FDF_CONJ_FLETCHER_REEVES || m_Algorithm==SymOptimizer::FDF_CONJ_POLAK_RIBIERE
		|| m_Algorithm==SymOptimizer::FDF_GRAD_DESC || m_Algorithm==SymOptimizer::FDF_BFGS)
	{
		if (m_bVerbous)
			std::cout<<"Initializing GSL Smooth Optimizer..."<<std::endl;
		
		m_pOptimizer = new SymOptimizerGSLSmooth(m_pDeformationModel->GetParameterVector().m, m_pDeformationModel->GetMinimalParamStep()(0), 
												 0, m_fMinChange, m_pEnergyFn, m_pDeformationModel, m_Algorithm, 
												 m_iMaxNumIterations, m_iMaxNumIterations, m_iUpdate, m_fRescaleAtUpdate, example);
	}
	else if (m_Algorithm==SymOptimizer::F_SIMPLEX_NELDER_MEAD)
	{
		if (m_bVerbous)
			std::cout<<"Initializing GSL Discrete Optimizer..."<<std::endl;	
		m_pOptimizer = new SymOptimizerGSLDiscrete(m_pDeformationModel->GetParameterVector().m, m_pEnergyFn, 
												   m_pDeformationModel, m_iMaxNumIterations, m_iMaxNumIterations,
												   m_iUpdate, m_fRescaleAtUpdate, example);
	}
	else if (m_Algorithm==SymOptimizer::GREEDY_WALK)
	{
		if (m_bVerbous)
			std::cout<<"Initializing Greedy Walk Optimizer..."<<std::endl;	
		m_pOptimizer = new SymOptimizerGreedyWalk(m_pEnergyFn, m_pDeformationModel, 
												  m_iMaxNumIterations, m_iMaxNumIterations, m_iUpdate, 
												  m_fRescaleAtUpdate, example);
		bool noOptimization = params.GetStrValue(SymParamsText::OPTIM_PARAMS, "DoNotOptimize", valid)=="true" && valid;
		((SymOptimizerGreedyWalk*)m_pOptimizer)->m_bNoOpimization = noOptimization;
	}
	else if (m_Algorithm==SymOptimizer::MY_GRADIENT_DESCENT)
	{
		if (m_bVerbous)
			std::cout<<"Initializing My Gradient Descent Optimizer..."<<std::endl;	
		m_pOptimizer = new SymOptimizerNDGradientDescent(numDims, m_pEnergyFn, m_pDeformationModel, 
														 m_iMaxNumIterations, m_iMaxNumIterations, m_iUpdate, 
														 m_fRescaleAtUpdate, example);
	}
		
	assert(m_pOptimizer!=NULL);
}

SymFilter * SymPipelineParams::GetFilter(SymParamsText & params, SymParamsText::ParamModuleID paramID)
{
	bool valid;
	QString filterType = params.GetStrValue(paramID, "Type", valid);
	if (valid)
	{
		if (filterType=="power")
		{
			double p = params.GetDoubleValue(paramID, "Pow", valid);	
			assert(valid);	//deviation must be defined
			double maxVal = params.GetDoubleValue(paramID, "MaxVal", valid);
			if (!valid)
				maxVal=-1;
			double maxRad = params.GetDoubleValue(paramID, "MaxRad", valid);
			if (!valid)
				maxRad=-1;			
			return new SymFilterPntAbsPow(p, maxVal, maxRad);
		}
		else if (filterType=="gauss")
		{	
			double std_dev = params.GetDoubleValue(paramID, "StdDev", valid);	
			assert(valid);	//deviation must be defined
			return new SymFilterGaussian(std_dev);
		}
		else if (filterType=="sobel")
		{	
			double blur = params.GetDoubleValue(paramID, "Blur", valid);	
			if (!valid)
				blur = 0;
			int window = params.GetIntValue(paramID, "Window", valid);	
			if (!valid)
				window = 0;
			
			return new SymFilterGradient(blur, window);
		}
		else if (filterType=="canny")
		{
			double blur = params.GetDoubleValue(paramID, "Blur", valid);	
			if (!valid)
				blur = 0;
			int window = params.GetIntValue(paramID, "Window", valid);	
			if (!valid)
				window = 0;
			std::vector<double> thresholds = params.GetDoubleValues(paramID, "Threshold", valid);	
			assert(valid);
			
			return new SymFilterCanny(blur, thresholds[0], thresholds[1], window);
		}
		else if (filterType=="special")
		{
			SymFilter * flt = new SymFilter();
			flt->m_sSpecialDirections = params.GetStrValue(paramID, "Directions", valid);
			assert(valid);
			return flt;
		}
		else if (filterType=="script")
		{
			QString executable = params.GetStrValue(paramID, "Executable", valid);
			assert(valid);
			QStringList paramsFilter = params.GetStrValues(paramID, "Parameters", valid);
			assert(valid);
			return new SymFilterScript(executable, paramsFilter);
		}
		else if (filterType=="none")
		{
			return NULL;
		}
	}	

	return NULL;
}

void SymPipelineParams::InitializeFilterParams(SymParamsText & params)
{
	m_pSymSpaceTargetFilter = GetFilter(params, SymParamsText::FILTER_TRG_PARAMS);
	m_pDeformPostFilter = GetFilter(params, SymParamsText::FILTER_DEFORM_PARAMS);
}

SymPipelineParams::SymPipelineParams(QImage &f, QImage &g, SymParamsText & params, ExampleWithAnswer * example)			// given target symmetry
{
	InitializeToDefault();
	InitializeFilterParams(params);	
	DeformationModel * imageG = NULL;
	if (example==NULL)
	{
		InitializeImageDeformation(g, params);
		if (!m_pDeformationModel->DistortsInput())
			imageG = m_pDeformationModel;
		InitializeImageDeformation(f, params);
	}
	else
	{
		m_pDeformationModel = example->m_pValidSolutionSpace;
		if (m_pDeformPostFilter!=NULL)
			m_pDeformationModel->SetPostFilter(m_pDeformPostFilter);
	}
	assert(m_pDeformationModel!=NULL);
	int fWidth, fHeight;
	m_pDeformationModel->GetOutputDimensions(&fWidth, &fHeight);

	double minWidth = qMin(fWidth, g.width());
	double minHeight = qMin(fHeight, g.height());
		
	std::cout<<"Initializing Symmetry space (Sg)"<<std::endl;
	
	InitializeSymmetrySpace(g.width(), g.height(), 0, params, minWidth, minHeight, imageG);
	SymMatrix * mg = NULL;
	if (imageG!=NULL)
		mg = &(imageG->DeformAtCurrentHierarchyLevel(imageG->GetParameterVector()));
	else
		mg = new SymMatrix2DReal(g);

	SymMatrix * Sg = m_pSymmetrySpace->GetFilteredSymSpace(mg, m_pSymSpaceTargetFilter); 
	
	if (m_bVerbous)
	{
		((SymMatrix2DReal*)Sg)->ToImage().save("targetSymmetry.png");
		QImage imgOfG = ((SymMatrix2DReal*)mg)->ToImage();
		QImage SgVis = ((SymSpace2D*) m_pSymmetrySpace)->GetVisualizationOfSymSpace(((SymMatrix2DReal&)*Sg), &imgOfG);	
		SgVis.save("targetSymmetryVis.png");
		
	}
	m_pSg = Sg;	
	
	std::cout<<"Initializing Symmetry space (Sf) "<<std::endl;	
	InitializeSymmetrySpace(fWidth, fHeight, 0, params, minWidth, minHeight, m_pDeformationModel);
	
	std::cout<<"Initializing Energy Fn"<<std::endl;
	InitializeEnergy(Sg, params, example);
	InitializeOptimization(params, example);
	std::cout<<"Initialized everything"<<std::endl;
}

SymPipelineParams::SymPipelineParams(QImage &f, SymParamsText & params, ExampleWithAnswer * example)		// use filtered f as a target
{	
	std::cout<<"SymPipelineParams::SymPipelineParams - 1"<<std::endl;
	InitializeToDefault();	
	std::cout<<"SymPipelineParams::SymPipelineParams - 2"<<std::endl;	
	InitializeFilterParams(params);	
	std::cout<<"SymPipelineParams::SymPipelineParams - 3. Example: "<<example<<std::endl;
	
	if (example==NULL)
		InitializeImageDeformation(f, params);
	else
	{
		m_pDeformationModel = example->m_pValidSolutionSpace;	
		if (m_pDeformPostFilter!=NULL)
			m_pDeformationModel->SetPostFilter(m_pDeformPostFilter);
	}
	std::cout<<"SymPipelineParams::SymPipelineParams - 3.1";
	int minWidth, minHeight;
	int inWidth, inHeight;
	m_pDeformationModel->GetOutputDimensions(&inWidth, &inHeight);
	minWidth = inWidth;
	minHeight = inHeight;
	std::cout<<"SymPipelineParams::SymPipelineParams - 4"<<std::endl;
	
	SymMatrix2DReal * gf = NULL;
	SymMatrix2DReal * FLTSgf = NULL;
	if (m_pSymSpaceTargetFilter!=NULL)	//add filtered self as a target
	{
		if (m_pDeformationModel->DistortsInput())
		{
			gf = new SymMatrix2DReal(f);
			minWidth = qMin(gf->m, inWidth);
			minHeight = qMin(gf->n, inHeight);
			
			InitializeSymmetrySpace(gf->m, gf->n, 0, params, minWidth, minHeight, m_pDeformationModel);			
			FLTSgf = ((SymMatrix2DReal*)m_pSymmetrySpace->GetFilteredSymSpace(gf, m_pSymSpaceTargetFilter));
			//NOTE: this is not right. Input size might have to be truncated different from Sg
			// since it takes dimensions of g as input
			minWidth = qMin(inWidth, ((SymMatrix2DReal*)FLTSgf)->m);
			minHeight = qMin(inWidth, ((SymMatrix2DReal*)FLTSgf)->n);
//			std::cout<<"!!!!!!!! minWidth="<<minWidth<<" minHeight="<<minHeight<<std::endl;
			InitializeSymmetrySpace(inWidth, inHeight, 0, params, minWidth, minHeight, m_pDeformationModel);
//			m_pSg = new SymMatrix2DReal(*FLTSgf);			
			m_pSg = new SymMatrix2DReal(*((SymMatrix2DReal*)m_pSymmetrySpace->RescaleAndNormalize(FLTSgf)));
		}
		else
		{
			gf = (SymMatrix2DReal*)&(m_pDeformationModel->DeformAtCurrentHierarchyLevel(m_pDeformationModel->GetParameterVector()));
			InitializeSymmetrySpace(inWidth, inHeight, 0, params, f.width(), f.height(), m_pDeformationModel);
			FLTSgf = (SymMatrix2DReal*)m_pSymmetrySpace->GetFilteredSymSpace(gf, m_pSymSpaceTargetFilter);
			m_pSg = new SymMatrix2DReal(*FLTSgf);
		}
	  if (m_bVerbous)
		  ((SymMatrix2DReal*)m_pSg)->ToImage().save("targetSymmetry.png");							  
	}
	else
	{
		InitializeSymmetrySpace(inWidth, inHeight, 0, params, f.width(), f.height(), m_pDeformationModel);
	}
	
	std::cout<<"SymPipelineParams::SymPipelineParams - 5"<<std::endl;
	InitializeEnergy(m_pSg, params, example);
	std::cout<<"SymPipelineParams::SymPipelineParams - 6"<<std::endl;
	InitializeOptimization(params, example);
	std::cout<<"SymPipelineParams::SymPipelineParams - DONE"<<std::endl;
}

SymPipelineParams::SymPipelineParams(QString curveFile, SymParamsText & params)
{
	std::cout<<"Initializing Modules"<<std::endl;
	InitializeToDefault();	
	std::cout<<"Initializing Filters"<<std::endl;	
	InitializeFilterParams(params);	
	std::cout<<"Initializing Curve Deformation Model"<<std::endl;
	InitializeCurveDeformation(curveFile, params);
	std::cout<<"Initializing 2D Symmetry Space"<<std::endl;	
//	int dimM = ((DeformCurve*)m_pDeformationModel)->m;
//	int dimN = ((DeformCurve*)m_pDeformationModel)->n;
	assert(m_pDeformationModel!=NULL);
	int dimM, dimN;
	m_pDeformationModel->GetOutputDimensions(&dimM, &dimN);	
	std::cout<<"\tDims="<<dimM<<" x "<<dimN<<std::endl;
	InitializeSymmetrySpace(dimM, dimN, 0, params, dimM, dimN, m_pDeformationModel);
	std::cout<<"Initializing Energy Function"<<std::endl;	
	if (m_pSymSpaceTargetFilter!=NULL)
	{
		SymMatrix * mf = &(m_pDeformationModel->DeformAtCurrentHierarchyLevel(m_pDeformationModel->GetParameterVector()));
		SymMatrix * FLTSf = m_pSymmetrySpace->GetFilteredSymSpace(mf, m_pSymSpaceTargetFilter);
		m_pSg = new SymMatrix2DReal((SymMatrix2DReal&)*FLTSf);
		if (m_bVerbous)
			((SymMatrix2DReal*)m_pSg)->ToImage().save("targetSymmetry.png");
	}		
	InitializeEnergy(m_pSg, params);
	std::cout<<"Initializing Optimization"<<std::endl;	
	InitializeOptimization(params);	
	std::cout<<"Done Initializing Modules"<<std::endl;		
}



