#include "Deformations.h"
#include "DeformCurve.h"
#include "SymEnergy.h"
#include "Numerical.h"
#include "SymOptimizer.h"
#include "SymFilters.h"

#include <gsl/gsl_multimin.h>
#include <Qt/qstringlist.h>
#include <Qt/qmap.h>
#include <Qt/qstring.h>


#ifndef __SYM_PARAMS_H
#define __SYM_PARAMS_H

class ImageWarpTestExample;

class SymParamsText
{
	public:
		enum ParamModuleID
		{
			SYM_SPACE_PARAMS = 0,
			DEFORM_PARAMS = 1,
			ENERGY_PARAMS = 2,
			OPTIM_PARAMS = 3,
			FILTER_TRG_PARAMS = 4,
			FILTER_DEFORM_PARAMS = 5,
			MAX_PARAMS_INVALID = 6
		};
		
		SymParamsText(QString inFilename, char ** argv, int argc);
		
		QMap<QString, QStringList> m_ParamModule[MAX_PARAMS_INVALID];	

		QStringList GetStrValues(ParamModuleID module, QString valName, bool & valid);
		QString GetStrValue(ParamModuleID module, QString valName, bool & valid);	
		double GetDoubleValue(ParamModuleID module, QString valName, bool & valid);	
		int GetIntValue(ParamModuleID module, QString valName, bool & valid);	
		std::vector<double> GetDoubleValues(ParamModuleID module, QString valName, bool & valid);	
		std::vector<int> GetIntValues(ParamModuleID module, QString valName, bool & valid);	
};

class SymPipelineParams
{
	public:		
		//Need to be initialized for all!
		DeformationModel * m_pDeformationModel;
		SymSpace * m_pSymmetrySpace;
		SymEnergy * m_pEnergyFn;
		SymOptimizer * m_pOptimizer;
				
		//Can be initialized (otherwise set to default)
		int numCtrl[2];
		bool invertMapping;
		int m_iMaxNumIterations;
		bool m_bVerbous;	
		int m_iUpdate;	
		SymOptimizer::OptimizationAlgirithm m_Algorithm;
		double m_fStep;
		SymVectorReal * m_fStepVector;	
		double m_fToLineMin; 
		double m_fMinChange;
		double m_fRescaleAtUpdate;
		
		SymMatrix * m_pSg;
	
//		SymSpace::FilterParams m_FilterTrgParams;
		SymFilter * m_pSymSpaceTargetFilter;
		SymFilter * m_pDeformPostFilter;
		
		
		//Auto-filled (do not initialize!)
		gsl_vector * m_pTempVec;
		int m_iFunctionEvals;
		int m_iDerivativeEvals;	
			
		static QImage LoadImage(QString filename, QString defaultSrc, ExampleWithAnswer ** syntheticExample);
	
		void InitializeToDefault();
		void InitializeSymmetrySpace(int width, int height, int zDim, SymParamsText & params, 
									 int minWidth, int minHeight, DeformationModel * deformationModel);
		void InitializeImageDeformation(QImage & f, SymParamsText & params);
		void InitializeEnergy(SymMatrix * Sg, SymParamsText & params, ExampleWithAnswer * example = NULL);
		void InitializeOptimization(SymParamsText & params, ExampleWithAnswer * example = NULL);
		void InitializeFilterParams(SymParamsText & params);
		void InitializeCurveDeformation(QString curveFile, SymParamsText & params);
		SymFilter * GetFilter(SymParamsText & params, SymParamsText::ParamModuleID paramID);
		
		SymPipelineParams(QImage &f, QImage &g, SymParamsText & params, ExampleWithAnswer * example = NULL);			// given target symmetry
		SymPipelineParams(QImage &f, SymParamsText & params, ExampleWithAnswer * example = NULL);	// use filtered f as a target
		SymPipelineParams(QString curveFile, SymParamsText & params);
	
		static double * DivideCurveParams(QString file, QString * outFilename);
};


#endif

