#include "Deformations.h"
#include "SymEnergy.h"
#include "Numerical.h"
#include <gsl/gsl_multimin.h>
#include "SymParams.h"
#include "SymOptimizer.h"

#ifndef __SYM_PIPELINE_H
#define __SYM_PIPELINE_H


//this function should return the result f(x,params) for argument x and parameters params. 
//If the function cannot be computed, an error value of GSL_NAN should be returned.
double optimF(const gsl_vector * x, void * params);
//this function should store the n-dimensional gradient g_i = d f(x,params) / d x_i in the vector 
//g for argument x and parameters params, returning an appropriate error code if the function cannot be computed.
void optimDF(const gsl_vector * x, void * params, gsl_vector * g);
//This function should set the values of the f and g as above, for arguments x and parameters params. 
//This function provides an optimization of the separate functions for f(x) and g(x)—it is always faster 
//to compute the function and its derivative at the same time. 
void optimFDF(const gsl_vector * x, void * params, double * f, gsl_vector * g);


class SymPipelineParams;

/**
 * Class that handles optimization and output of results.
 */
class SymPipeline
{
	public:
		SymPipeline(SymPipelineParams * params);
		virtual ~SymPipeline();
		void FindOptimalSolution();

		void OutputResultsImg(QString outDir, QImage * f, SymMatrix2DReal * Sf = NULL, QImage * g = NULL, 
							  SymMatrix2DReal * Sg=NULL, QString fileExt = ".png", bool printHistoryGifs = true);
		static QString AddTrailingZeros(int val, int numChars);
			
		//static void TestSimple();
		SymPipelineParams * m_pOptFnParams;
};	

#endif

