#include "SymSpaceAdditional.h"
#include "Deformations.h"

/****
 * POINTWISE (TOROIDAL)
 */

SymSpacePointwiseToroidal::SymSpacePointwiseToroidal(int im_m, int im_n, int S_m, int S_n, 
													 NormalizationType norm)
: SymSpace2D(im_m, im_n, (S_m<0 ? im_m : S_m), (S_n<0 ? im_n : S_n), 1., norm, 1., 1.)
{
//	m_pTmpFFTConj = new SymMatrix2DComplex(im_m, im_n);
	m_pTmpFFT = new SymMatrix2DComplex(im_m, im_n);
	m_pTmpConvFFT = new SymMatrix2DComplex(im_m, im_n);
	m_pTmpFullAnswer = new SymMatrix2DReal(im_m, im_n);	
}

void SymSpacePointwiseToroidal::CalculateSymmetry(SymMatrix & f, double * L2f)
{
	if (L2f!=NULL)
		*L2f = f.SumElements(2);
	
#ifdef CALCULATE_TIMING
	TimeProfiler::m_pGlobalSuperProfiler->startedProcess("SymSpacePointwiseTorus:FFT2D");
#endif
	SymMatrix2DReal & Sreal = (SymMatrix2DReal &)*m_pSymmetryValues;
	m_FFTManager.FFT2D((SymMatrix2DReal &)f, *m_pTmpFFT);
#ifdef CALCULATE_TIMING
	TimeProfiler::m_pGlobalSuperProfiler->finishedProcess("SymSpacePointwiseTorus:FFT2D", false);
#endif	
//	m_pTmpFFT->Conj(*m_pTmpFFTConj);
	m_pTmpFFT->PointwiseMultiply(*m_pTmpFFT, *m_pTmpConvFFT);
	
#ifdef CALCULATE_TIMING
	TimeProfiler::m_pGlobalSuperProfiler->startedProcess("SymSpacePointwiseTorus:IFFT2D");
#endif	
	m_FFTManager.IFFT2D(*m_pTmpConvFFT, *m_pTmpFFT);
#ifdef CALCULATE_TIMING
	TimeProfiler::m_pGlobalSuperProfiler->finishedProcess("SymSpacePointwiseTorus:IFFT2D", false);
#endif	
	
	if (m_pTmpFFT->m==Sreal.m && m_pTmpFFT->n==Sreal.n)
	{
		Sreal = *m_pTmpFFT;
	}
	else
	{
		*m_pTmpFullAnswer = *m_pTmpFFT;
		CropCorners(*m_pTmpFullAnswer);
	}	
}

/****
 * POINTWISE (PADDED)
 */

SymSpacePointwisePadded::SymSpacePointwisePadded(int im_m, int im_n, int S_m, int S_n, 
						NormalizationType norm)
: SymSpace2D(im_m, im_n, S_m<0 ? (2*im_m) : S_m, S_n<0 ? (2*im_n) : S_n, 1, norm, 1., 1.)
{
	m_pPaddedF = new SymMatrix2DReal(2*im_m, 2*im_n);	
	m_pTmpFFT = new SymMatrix2DComplex(2*im_m, 2*im_n);
	m_pTmpConvFFT = new SymMatrix2DComplex(2*im_m, 2*im_n);
	m_pTmpFullAnswer = new SymMatrix2DReal(2*im_m, 2*im_n);		
}

void SymSpacePointwisePadded::CalculateSymmetry(SymMatrix & f, double * L2f)
{
	if (L2f!=NULL)
		*L2f = f.SumElements(2);
	
	SymMatrix2DReal & Sreal = (SymMatrix2DReal &)*m_pSymmetryValues;
	SymMatrix2DReal & fReal = (SymMatrix2DReal &)f;
	for (int i=0; i<m_pPaddedF->m; i++)
		for (int j=0; j<m_pPaddedF->n; j++)
			(*m_pPaddedF)(i,j) = 0;
	
	for (int i=0; i<fReal.m; i++)
		for (int j=0; j<fReal.n; j++)
			(*m_pPaddedF)(i,j) = fReal(i,j);
	
	m_FFTManager.FFT2D(*m_pPaddedF, *m_pTmpFFT);	
	//m_pTmpFFT->Conj(*m_pTmpFFTConj);	
	//m_FFTManager.FFT2D(*m_pPaddedF1, *m_pTmpFFT);
	m_pTmpFFT->PointwiseMultiply(*m_pTmpFFT, *m_pTmpConvFFT);
	m_FFTManager.IFFT2D(*m_pTmpConvFFT, *m_pTmpFFT);
	
	if (m_pTmpFFT->m==Sreal.m && m_pTmpFFT->n==Sreal.n)
	{
		Sreal = *m_pTmpFFT;
	}
	else
	{
		*m_pTmpFullAnswer = *m_pTmpFFT;
		CropCenter(*m_pTmpFullAnswer);
	}	
}

void SymSpacePointwisePadded::TestPointwise()
{
	SymMatrix2DReal f(3, 3);
    int baseInd = 0;
   	f(baseInd+0, baseInd+0) = 12;	f(baseInd+0, baseInd+1) = 12;	f(baseInd+0, baseInd+2) = 1;
	f(baseInd+1, baseInd+0) = 5;    f(baseInd+1, baseInd+1) = 1;	f(baseInd+1, baseInd+2) = 3;
	f(baseInd+2, baseInd+0) = 1;	f(baseInd+2, baseInd+1) = 4;	f(baseInd+2, baseInd+2) = 2;
	
	SymSpacePointwisePadded S(f.m, f.n, -1, -1, SymSpace::NORM_NONE);
	
	SymMatrix2D & Sf = (SymMatrix2D &)S.CalculateSymmetryAndProcess(f);
	std::cout<<"f="<<std::endl;
	f.PrintMatrix();
	std::cout<<"S(f)="<<std::endl;
	Sf.PrintMatrix(0.0001);
	
}

/****
 * ROTATIONAL AROUND SINGLE POINT
 */
SymSpaceRotationalSinglePnt2D::SymSpaceRotationalSinglePnt2D(
									int im_m, int im_n, int numAngles, int numRadii, 
									DeformationModel * deformationModel,
									NormalizationType norm, bool cropToDisk, int centerOfMass,
									double ignoreMinAngle, double ignoreMaxAngle)
: SymSpaceSinglePointReflection2D(im_m, im_n, numAngles, numRadii, 
								deformationModel, 1, norm, 1., cropToDisk, centerOfMass)
{
	assert(m_pDeformationModel!=NULL);
	m_iVisualizationAlphaScaleFactor = 1;
	m_pTmpConjFFT = new SymMatrix2DComplex(m_pTmpFFT->m, m_pTmpFFT->n);
	
	if (ignoreMinAngle<0)
		ignoreMinAngle+= 360;
	m_iIgnoreMin = floor(ignoreMinAngle * numAngles / 360);
	m_iIgnoreMax = ceil(ignoreMaxAngle * numAngles / 360);
	m_fSetToValue = -1;	
}


void SymSpaceRotationalSinglePnt2D::CalculateSymmetry(SymMatrix & f, double * L2f)
{
	SymMatrix2DReal & fReal = (SymMatrix2DReal&)f;
	SymMatrix2DReal & Sreal = (SymMatrix2DReal &)*m_pSymmetryValues;
	
	if (!m_bCropToDisk)
	{
		SetCenterOfMass();
//		double cx, cy;
//		m_pDeformationModel->GetCenterOfMass(&cx, &cy);
//		m_Center.X(cx);
//		m_Center.Y(cy);
	}
	
	ConvertToPolar(fReal, *m_FPolar, m_fScaleR, (double)m_Center.X(), (double)m_Center.Y());
	
	if (L2f!=NULL)
		*L2f = m_FPolar->SumElements(2);
	
	FillColSymmetryPerAngleAutocorr(Sreal, 0, *m_FPolar, m_pTmpFFT, m_pTmpConvFFT, m_pTmpConjFFT);	//remove cx, cy, r	

//	if (m_fSetToValue==-1)
	{
		m_fSetToValue = Sreal.Mean();
	}
	
	if (m_iIgnoreMin < m_iIgnoreMax)
	{
		for (int i=m_iIgnoreMin; i<m_iIgnoreMax; i++)
			Sreal(i, 0) = m_fSetToValue;
	}
	else if (m_iIgnoreMin > m_iIgnoreMax)
	{
		for (int i=m_iIgnoreMin; i<Sreal.m; i++)
			Sreal(i, 0) = m_fSetToValue;
		for (int i=0; i<m_iIgnoreMax; i++)
			Sreal(i, 0) = m_fSetToValue;
	}
}

void SymSpaceRotationalSinglePnt2D::FillColSymmetryPerAngleAutocorr(
									SymMatrix2DReal & fillMe, int col, 
									SymMatrix2DReal & f_polar, SymMatrix2DComplex * tmpFFT, 
									SymMatrix2DComplex * tmpConvFFT, SymMatrix2DComplex * tmpConjFFT)
{
	//	Sf = sum(ifft(fft(frep).*fft(frep)));
	
	m_FFTManager.FFT1DCol(f_polar, *tmpFFT);
	tmpFFT->Conj(*tmpConjFFT);
	tmpFFT->PointwiseMultiply(*tmpConjFFT, *tmpConvFFT);
	m_FFTManager.IFFT1DCol(*tmpConvFFT, *tmpFFT);
	
	for (int i=0; i<fillMe.m; i++)
		fillMe(i, col) = 0;
	
	//	std::cout<<"fillme= "<<fillMe.m<<" x "<<fillMe.n<<std::endl;
	//	std::cout<<"fillme= "<<tmpFFT->m<<" x "<<tmpFFT->n<<std::endl;
	for (int i=0; i<fillMe.m; i++)
		for (int j=0; j<tmpFFT->n; j++)
			fillMe(i, col) += (*tmpFFT)(i, j).dat[0];
}

