#include "SymSpaceReflective.h"
#include "DeformCurve.h"
#include "DrawingRoutines.h"
#include <math.h>
#include "Deformations.h"

/***************** Single point planes code ******************/
SymSpaceSinglePointReflection2D::SymSpaceSinglePointReflection2D(int img_m, int img_n, int numAngles, 
																 int numR, DeformationModel * deformationModel, 
																 double p, NormalizationType norm, 
																 double constNormScalar, bool cropToDisk, 
																 int COMflags)
: SymSpace2D(img_m, img_n, numAngles, 1, p, norm, constNormScalar, 1, cropToDisk)
{
	m_bCenterInitialized = false;
	m_iCenterOfMassFlags = COMflags;
	m_iVisualizationAlphaScaleFactor = 2;
	m_pDeformationModel = deformationModel;
	m_Image = new QImage(img_m, img_n, QImage::Format_RGB32);
	assert(m_pDeformationModel!=NULL);
	SetCenterOfMass();
//	double cx, cy;
//	m_pDeformationModel->GetCenterOfMass(&cx, &cy);
//	m_Center.X(cx);
//	m_Center.Y(cy);	
	m_fNumR = numR;
	if (m_fNumR==-1)
	{
		m_fNumR = qMax(img_m, img_n);
		m_fScaleR = 1;
	}
	else
	{
		m_fScaleR = (double)qMax(img_m, img_n) / (double)numR;
	}
	m_FPolar = new SymMatrix2DReal(numAngles, m_fNumR);
	
	m_pTmpFFT = new SymMatrix2DComplex(numAngles, m_fNumR);
	m_pTmpConvFFT = new SymMatrix2DComplex(numAngles, m_fNumR);
	
}

void SymSpaceSinglePointReflection2D::CalculateSymmetry(SymMatrix & f, double * L2f)
{	
	SymMatrix2DReal & fReal = (SymMatrix2DReal&)f;
	SymMatrix2DReal & Sreal = (SymMatrix2DReal &)*m_pSymmetryValues;
	if (!m_bCropToDisk)
	{
		SetCenterOfMass();
//		double cx, cy;
//		m_pDeformationModel->GetCenterOfMass(&cx, &cy);
//		m_Center.X(cx);
//		m_Center.Y(cy);
	}	
	ConvertToPolar(fReal, *m_FPolar, m_fScaleR, (double)m_Center.X(), (double)m_Center.Y());

	if (L2f!=NULL)
		*L2f = m_FPolar->SumElements(2);
		
	FillColSymmetryPerAngle(Sreal, 0, *m_FPolar, m_pTmpFFT, m_pTmpConvFFT);	//remove cx, cy, r
}

QImage SymSpaceSinglePointReflection2D::GetVisualizationOfSymSpace(SymMatrix2DReal & Sf, QImage * img)
{	
	static int tempID=0;	
//	std::cout<<"Getting Vis"<<std::endl;
	assert(img!=NULL);
//	for (int i=0; i<img->width(); i++)
//		for (int j=0; j<img->height(); j++)
//			m_Image->setPixel(i,j, img->pixel(i,j));
	*m_Image = *img;
	double pixelX = 1. / img->width();
	double pixelY = 1. / img->height();
	double cx = m_Center.X();
	double cy = m_Center.Y();

//	std::cout<<"cx,y=["<<cx<<", "<<cy<<"] pixel="<<pixel<<std::endl;	
	double rMax = qMin(qMin(qAbs(img->width()-cx), qAbs(img->height()-cy)), qMin(cx, cy))-1;
	double Smax = Sf.GetMax(false);
	double Smin = Sf.GetMin(false);
	rMax /= 2;
//	double Smin = 0;
	double maxR = -1;
//	QRgb white = qRgb(255, 255, 255);
//	m_Image->setPixel(m_Center.X(), m_Center.Y(), white);
	DrawingRoutines::DrawCircle(*m_Image, cx*pixelX, cy*pixelY, pixelX, pixelX, qRgb(255, 255, 0), qRgb(0,0,255));
//	std::cout<<"Smin="<<Sf.GetMin(true)<<", Smax="<<Smax<<std::endl;
	for (int i=0; i<Sf.m*m_iVisualizationAlphaScaleFactor; i++)
	{
		double r = 0;
		if (i<Sf.m)
			r = (Sf(i, 0)-Smin)/(Smax-Smin) * rMax;
		else
			r = (Sf(i-Sf.m, 0)-Smin)/(Smax-Smin) * rMax;
			
		maxR = qMax(r, maxR);
		double theta = ((double)i*2*3.14159265/m_iVisualizationAlphaScaleFactor) / (double)Sf.m;
		
		int width=1;
		int px=m_Center.X() + r * cos(theta);
		int py=m_Center.Y() + r * sin(theta);
		for (int x=qMax(0, px-width); x<=qMin(m_Image->width()-1, px+width); x++)
			for (int y=qMax(0, py-width); y<=qMin(m_Image->height()-1, py+width); y++)
				m_Image->setPixel(x, y, qRgb(m_iVisR, m_iVisG, m_iVisB));
		
//		if (i%20==0)
//		{
//			std::cout<<"theta="<<theta<<" r= "<<Sf(i,0)<<" imgR="<<r<<std::endl;
//			m_Image->save("MyVis"+QString::number(tempID++)+".png");
//		}
			
	}
	

//	m_Image->save("MyVis"+QString::number(tempID++)+".png");
	return *m_Image;
}

void SymSpaceSinglePointReflection2D::ConvertToPolar(SymMatrix2DReal & src_f, SymMatrix2DReal & trg_f_polar, double scaleR, double cx, double cy)
{
	int numAngles = trg_f_polar.m;
	int numRadii = trg_f_polar.n;
//	std::cout<<"numRadii="<<numRadii<<" scale="<<scaleR<<" finalR="<<(numRadii-1) * scaleR<<std::endl;
	for (int t=0; t<numAngles; t++)
	{
		double theta = ((double)t * 2 * 3.14159265) / numAngles;
		double cosTheta = cos(theta);
		double sinTheta = sin(theta);
		for (int r=0; r<numRadii; r++)
		{
			//double R = r+1;
			double R = r;
			double sampleX = scaleR * R*cosTheta + cx;
			double sampleY = scaleR * R*sinTheta + cy;
            trg_f_polar(t, r) = src_f.SampleBilinear(sampleX, sampleY);
		}
	}
}

void SymSpaceSinglePointReflection2D::FillColSymmetryPerAngle(SymMatrix2DReal & fillMe, int col, SymMatrix2DReal & f_polar, SymMatrix2DComplex * tmpFFT, SymMatrix2DComplex * tmpConvFFT)
{	
//	Sf = sum(ifft(fft(frep).*fft(frep)));
	
	m_FFTManager.FFT1DCol(f_polar, *tmpFFT);
	tmpFFT->PointwiseMultiply(*tmpFFT, *tmpConvFFT);
	m_FFTManager.IFFT1DCol(*tmpConvFFT, *tmpFFT);
	
	for (int i=0; i<fillMe.m; i++)
		fillMe(i, col) = 0;
	
//	std::cout<<"fillme= "<<fillMe.m<<" x "<<fillMe.n<<std::endl;
//	std::cout<<"fillme= "<<tmpFFT->m<<" x "<<tmpFFT->n<<std::endl;
	for (int i=0; i<fillMe.m; i++)
	{
		for (int j=0; j<tmpFFT->n; j++)
		{
			fillMe(i, col) += (*tmpFFT)(i, j).dat[0];
		}
	}
}

void SymSpaceSinglePointReflection2D::SetCenterOfMass()
{
	if (!m_bCenterInitialized || ((m_iCenterOfMassFlags & COM_UPDATE)!=0))
	{
		m_bCenterInitialized = true;
		double cx, cy;
				
		if ((m_iCenterOfMassFlags & COM_INTENSITY_MASS)!=0)
		{
			m_pDeformationModel->GetCenterOfMass(&cx, &cy);
		}
		else if ((m_iCenterOfMassFlags & COM_SYMMETRY_MASS)!=0)
		{
			GetSymmetryCenterOfMass(&cx, &cy);
		}
		else	//COM_AREA_MASS
		{
			cx = image_m/2;
			cy = image_n/2;
		}

		m_Center.X(cx);
		m_Center.Y(cy);
	}
}

void SymSpaceSinglePointReflection2D::GetSymmetryCenterOfMass(double * cx, double *cy)
{
	//PRST
	SymMatrix2DReal & f = (SymMatrix2DReal &) m_pDeformationModel->DeformAtCurrentHierarchyLevel(m_pDeformationModel->GetParameterVector());
	std::cout<<"[WARNING] Performing PRST. Might take a while. Img="<<f.m<<"x"<<f.n<<std::endl;	
	f.ToImage().save("PRSTVis_f.png");	
//	SymSpacePRST2D::PRST_CALC_SAMPLE_N4
	SymSpacePRST2D S(f.m, f.n, 144, -1, SymSpacePRST2D::PRST_CALC_SAMPLE_ORIENTATIONS, 1, SymSpace::NORM_IMAGE_SPACE);	
	SymMatrix2DReal & Sf = (SymMatrix2DReal &)S.CalculateSymmetryAndProcess(f);
	Sf.ToImage().save("PRSTVis_S.png");
	QImage img(f.m, f.n, QImage::Format_RGB32);
	QImage result = S.GetVisualizationOfSymSpace(Sf, &img);
	result.save("PRSTVis.png");
	double maxVal = -1;
	for (int i=0; i<result.width(); i++)
	{
		for (int j=0; j<result.height(); j++)
		{
			if (maxVal < qGray(result.pixel(i,j)) )
			{
				maxVal = qGray(result.pixel(i,j));
				*cx = i;
				*cy = j;
			}
		}
	}

	result.setPixel((int)*cx, (int)*cy, qRgb(0, 0, 255));
	
	result.save("PRSTVis_pnt.png");
	
}

void SymSpaceSinglePointReflection2D::TestPointReflection()
{
//	double ovalParams[5];
//	ovalParams[0] = 50;		//rx
//	ovalParams[1] = 100;	//ry
//	ovalParams[2] = 0;		//cx
//	ovalParams[3] = 0;	//cy
//	ovalParams[4] = 20;		//# lines
//	DeformCurve curve("oval", ovalParams, 1, -1, -1, -1);	//last one - std_dev
//	std::cout<<"curvedims=["<<curve.m<<", "<<curve.n<<"]"<<std::endl;
//	SymSpaceSinglePointReflection2D S(curve.m, curve.n, 720, 800, &curve, 1, SymSpace::NORM_NONE, 1);
//	
//	SymMatrix2DReal & f = (SymMatrix2DReal &) curve.DeformAtCurrentHierarchyLevel(curve.GetParameterVector());
//	QImage fImg = curve.GetDeformedImage();
//	f.ToImage().save("curve_f.png");
//	fImg.save("curve_img.png");
//	SymMatrix2DReal & Sf = (SymMatrix2DReal &)S.CalculateSymmetryAndProcess(f);
//	Sf.ToImage().save("curve_S.png");
//	S.GetVisualizationOfSymSpace(Sf, &fImg).save("curve_S_img.png");
	
}


/***************** PRST code ******************/
SymSpacePRST2D::SymSpacePRST2D(int img_m, int img_n, int numAngles, int numR, 
							   PRSTCalculationMethod method, double p, NormalizationType norm, 
							   double constNormScalar, bool cropToDisk)
: SymSpace2D(img_m, img_n, numAngles, numR<=0 ? qMax(img_m, img_n) : numR, 
			 p, norm, constNormScalar, 1, cropToDisk)
{
	m_bExtraVerbous = false;
	m_Image = new QImage(img_m, img_n, QImage::Format_RGB32);
	m_VisualizationMatrix = new SymMatrix2DReal(img_m, img_n);	
	m_fCenterX = (double)img_m/2.;
	m_fCenterY = (double)img_n/2.;	
//	m_fCenterX = 0;
//	m_fCenterY = 0;
	m_iNumRadii = numR<=0 ? qMax(img_m, img_n) : numR;
	m_iNumAngles = numAngles;
	m_fScaleR = (double)qMax(img_m, img_n) / (double)m_iNumRadii;
	
	//NOTE: currently it's easy to compose number of 2's. 
	// Ideally we want minimum number combined of 2^a * 3^b * 5^c > maxDiagonal
	m_iMaxDiagonal = pow(2, ceil(log(sqrt(img_m*img_m + img_n*img_n)+2)/log(2)));
	//std::cout<<m_iMaxDiagonal<<std::endl;
	m_FResampled = new SymMatrix2DReal(2*m_iMaxDiagonal, 2*m_iMaxDiagonal);
	m_pTmpFFT = new SymMatrix2DComplex(2*m_iMaxDiagonal, 2*m_iMaxDiagonal);
	m_pTmpConvFFT = new SymMatrix2DComplex(2*m_iMaxDiagonal, 2*m_iMaxDiagonal);	
	m_CalcMethod = method;
	if (m_CalcMethod!=PRST_CALC_SAMPLE_N4
		&& m_CalcMethod!=PRST_CALC_SAMPLE_ORIENTATIONS)
	{
		std::cout<<"SymSpacePRST2D::SymSpacePRST2D Error: Only brute force PRST calculation method is implemented"<<std::endl;
		exit(0);
	}
}

void SymSpacePRST2D::PRSTBruteForce(SymMatrix2DReal & f)
{
	for (int r=0; r<m_iNumRadii; r++)
	{
		if (m_bExtraVerbous && r%10==0)
			std::cout<<"r="<<r<<" of "<<m_iNumRadii<<std::endl;
		for (int t=0; t<m_iNumAngles; t++)
		{
			((SymMatrix2DReal&)*m_pSymmetryValues)(t, r)=0;
			double radius = (double)r * m_fScaleR;
			double theta = 2*3.14159265*(double)t/(double)m_iNumAngles;
			double x, y, dx, dy, nx, ny;
			ConvertToLine(radius, theta, &x, &y, &dx, &dy, &nx, &ny);
			for (int i=0; i<f.m; i++)
			{
				for (int j=0; j<f.n; j++)
				{
					double newDX = i - x;
					double newDY = j - y;
					double t_inF = newDX * dx + newDY * dy;
					double tau_inF = newDX * nx + newDY * ny;
					double sampleX = x + t_inF*dx - tau_inF*nx;
					double sampleY = y + t_inF*dy - tau_inF*ny;
					
					double val2 = f.SampleBilinear(sampleX, sampleY);
					double val1 = f.SampleBilinear(i, j);
					((SymMatrix2DReal&)*m_pSymmetryValues)(t, r) += val1*val2;
				}
			}
		}
	}
}

void SymSpacePRST2D::ConvertToLine(double radius, double theta, double * x, double * y, double * dx, double * dy, double * nx, double * ny)
{
	*x = m_fCenterX + radius * cos(theta);
	*y = m_fCenterY + radius * sin(theta);	
	*dx = -sin(theta);
	*dy = cos(theta);
	*nx = cos(theta);
	*ny = sin(theta);
}

void SymSpacePRST2D::PRSTSampleOrientations(SymMatrix2DReal & f)
{
	for (int t=0; t<m_iNumAngles; t++)
		for (int r=0; r<m_iNumRadii; r++)
			((SymMatrix2DReal&)(*m_pSymmetryValues))(t, r) = 0;
	
	for (int t=0; t<m_iNumAngles/2; t++)
	{
		if (m_bExtraVerbous && (t%10==0))
			std::cout<<"Orientation "<<t<<" of "<<m_iNumAngles/2<<std::endl;
			
		double theta = (2*3.14159265)*(double)t/(double)m_iNumAngles;
		double dx, dy;
		double nx, ny;
		double ox, oy;
		ConvertToLine(0, theta, &ox, &oy, &dx, &dy, &nx, &ny);

		if (m_bExtraVerbous)
			m_TimeProfiler.startedProcess("Resampling");
			//TODO: rotate by theta instead of resampling
		ResampleAt(ox, oy, dx, dy, nx, ny, *m_FResampled, f);
		if (m_bExtraVerbous)
			m_TimeProfiler.finishedProcess("Resampling", false);
		
		if (m_bExtraVerbous)
			m_TimeProfiler.startedProcess("FFT convolution");		
		m_FFTManager.FFT1DRow(*m_FResampled, *m_pTmpFFT);
		m_pTmpFFT->PointwiseMultiply(*m_pTmpFFT, *m_pTmpConvFFT);
		m_FFTManager.IFFT1DRow(*m_pTmpConvFFT, *m_pTmpFFT);
		if (m_bExtraVerbous)
			m_TimeProfiler.finishedProcess("FFT convolution", false);		

		if (m_bExtraVerbous)
			m_TimeProfiler.startedProcess("Copying convolved");
		for (int tau=0; tau<m_FResampled->m; tau++)
		{
			int r=((int)(((double)tau)/m_fScaleR)-m_FResampled->m/2)/2;
			bool invR = false;
			
			if (tau < m_FResampled->m/2)
			{
				invR = true;
				if (r<0)
					r = -r;
			}
			
			int newT = invR ? t+m_iNumAngles/2 : t;
			
			if (r < ((SymMatrix2DReal*)m_pSymmetryValues)->n)	//TOOD: weird bug for the 1st row, t>pi? have to multiply by 2
			{
				for (int j=0; j<m_FResampled->n; j++)
					((SymMatrix2DReal&)(*m_pSymmetryValues))(newT, r) += (r==0&&newT>=m_iNumAngles/2) ? 2*(*m_pTmpFFT)(j, tau).dat[0] : (*m_pTmpFFT)(j, tau).dat[0];
			}
		}
		if (m_bExtraVerbous)
			m_TimeProfiler.finishedProcess("Copying convolved", false);
	}
}

void SymSpacePRST2D::ResampleAt(double ox, double oy, double dx, double dy, 
								double nx, double ny, SymMatrix2DReal & trg, SymMatrix2DReal & src)
{
	//NOTE: it is much faster to rotate by theta
	double lox = ((double)trg.m)/4.;
	double loy = ((double)trg.n)/4.;
	
	for (int i=0; i<trg.m; i++)
	{
		for (int j=0; j<trg.n; j++)	//fill resampled function
		{	
			double x = ox + ((double)i-lox) * dx + ((double)j-loy) * nx;
			double y = oy + ((double)i-lox) * dy + ((double)j-loy) * ny;
			trg(i,j) =src.SampleBilinear(x, y);
		}
	}	
}

void SymSpacePRST2D::CalculateSymmetry(SymMatrix & f, double * L2f)
{	
	//TODO: is it accurate?
	if (L2f!=NULL)
		*L2f = f.SumElements(2);
	
	if (m_bExtraVerbous)
		m_TimeProfiler.startedProcess("Calculating Symmetry");
	
	if (m_CalcMethod==PRST_CALC_SAMPLE_N4)
		PRSTBruteForce((SymMatrix2DReal &) f);
	else if (m_CalcMethod==PRST_CALC_SAMPLE_ORIENTATIONS)
		PRSTSampleOrientations((SymMatrix2DReal&)f);
	
	if (m_bExtraVerbous)
	{
		m_TimeProfiler.finishedProcess("Calculating Symmetry", false);	
		std::cout<<m_TimeProfiler.profile().toStdString().c_str()<<std::endl;
	}
}

QImage SymSpacePRST2D::GetVisualizationOfSymSpace(SymMatrix2DReal & Sf, QImage * img)
{
	assert(img!=NULL);
	*m_Image = *img;	
	if (m_CalcMethod==PRST_CALC_SAMPLE_N4 || m_CalcMethod==PRST_CALC_SAMPLE_ORIENTATIONS)
	{
		double globalMaxVal = ((SymMatrix2DReal*)m_pSymmetryValues)->GetMax(true);
		for (int i=0; i<m_Image->width(); i++)
		{
			for (int j=0; j<m_Image->height(); j++)
			{
				double gammaMax = 0;
				for (int g=0; g<m_iNumAngles; g++)
				{
					double gamma = (double)g * 3.14159265 / (double)(m_iNumAngles);
					double r, theta;
					ConvertXYGammaToRTheta(i,  j,  gamma, &r, &theta);
					double currVal = Sf.SampleBilinear(theta*m_iNumAngles/(2*3.14159265), r / m_fScaleR);
					gammaMax += pow(currVal/globalMaxVal, 4);
					//gammaMax = qMax(currVal/globalMaxVal, gammaMax);
				}
				(*m_VisualizationMatrix)(i,j) = gammaMax;
			}
		}
	}
	double maxVal = m_VisualizationMatrix->GetMax(true);
	if (maxVal>0)
		m_VisualizationMatrix->DivideBy(maxVal);
	for (int i=0; i<m_Image->width(); i++)
	{
		for (int j=0; j<m_Image->height(); j++)
		{
			int c1 = qMax(qRed(m_Image->pixel(i,j)), qBlue(m_Image->pixel(i,j)));
			c1 = qMax(c1, qGreen(m_Image->pixel(i,j)));
			int c2 = (*m_VisualizationMatrix)(i,j)*255;
			if (c1<128)
				m_Image->setPixel(i,j, qRgb(c2,c2,c2));
			
		}
	}
	return *m_Image;
}

void SymSpacePRST2D::ConvertXYGammaToRTheta(double x, double y, double gamma, double * r, double * theta)
{	
	x = x-m_fCenterX;
	y = y-m_fCenterY;
	*r = x * cos(gamma) + y * sin(gamma);
	*theta = gamma;
	if (*r<0)
	{
		*r = -*r;
		*theta = *theta + 3.14159265;
	}
}

void SymSpacePRST2D::TestPRSTReflections()
{
//	double ovalParams[5];
//	ovalParams[0] = 20;		//rx
//	ovalParams[1] = 50;	//ry
//	ovalParams[2] = 0;		//cx
//	ovalParams[3] = 0;	//cy
//	ovalParams[4] = 20;		//# lines
//	//DeformCurve curve("oval", ovalParams, 1, -1, -1, -1);	//last one - std_dev
//	DeformCurve curve("../../../data/curves/lizard_pts_fixed.closed", 1, -1, 200);	
//	std::cout<<"curvedims=["<<curve.m<<", "<<curve.n<<"]"<<std::endl;
//	//SymSpacePRST2D S(curve.m, curve.n, 144, -1, PRST_CALC_SAMPLE_ORIENTATIONS, 1, SymSpace::NORM_IMAGE_SPACE);
//	SymSpacePRST2D S(curve.m, curve.n, 144, -1, PRST_CALC_SAMPLE_N4, 1, SymSpace::NORM_IMAGE_SPACE);
//	S.m_bExtraVerbous = true;
//	SymMatrix2DReal & f = (SymMatrix2DReal &) curve.DeformAtCurrentHierarchyLevel(curve.GetParameterVector());
//	QImage fImg = curve.GetDeformedImage();
//	f.ToImage().save("curve_f.png");
//	fImg.save("curve_img.png");
//	SymMatrix2DReal & Sf = (SymMatrix2DReal &)S.CalculateSymmetryAndProcess(f);
//	Sf.ToImage().save("curve_S.png");
//	S.GetVisualizationOfSymSpace(Sf, &fImg).save("curve_S_img.png");	
}







